// CLIP_IO.cpp package
//CREATED ON January 5, 1997, FOR V3.0 BY J.BIRD
//MOD ON October 29 1997, FOR V3.02 BY J.Bird
#include <windows.h>
#include <alloc.h>
#include <mem.h>
#include "pcshare.h"
#include "file_han.h"

unsigned char *clipbrd_ptr, *clip_wrt_ptr ;
BOOL writing_clipboard, clipboard_open ;
unsigned clipboard_size, clip_count ;

const unsigned max_clipboard_size = 65535 ;

char open_clipboard_device ( BOOL write_flag )
{
	unsigned long i ;
	HANDLE clipboard_handle ;
	unsigned char *clip_ptr, *temp_ptr ;

	if ( clipboard_open )
		return FALSE ;

	// allocate temporary buffer for the clipboard data
	clipbrd_ptr = (unsigned char*)malloc ( max_clipboard_size ) ;
	if ( clipbrd_ptr == NULL )
		return FALSE ;

	clipboard_size = 0 ;
	clip_count = 0 ;
	clip_wrt_ptr = clipbrd_ptr ;

	// set flag to determine txfrs allowed to/from clipboard
	if ( write_flag )
		writing_clipboard = TRUE ;
	else
	{
		writing_clipboard = FALSE ;

		// get the clipboard data now if reading
		if ( !OpenClipboard ( Main_hWnd ) )
		{
			free ( clipbrd_ptr ) ;
				return FALSE ;
		}
		clipboard_handle = GetClipboardData ( CF_TEXT ) ;

		// if valid data is available
		if ( clipboard_handle != NULL )
		{
			// get ptr to data
			clip_ptr = (unsigned char*)GlobalLock ( clipboard_handle ) ;
			temp_ptr = clip_ptr ;
			// look for terminating NULL
			for (i=0 ; i<=max_clipboard_size ; i++ )
			{
				if ( *temp_ptr == 0 )
				{
					clipboard_size = i ;
					i=65536 ;
				}
				temp_ptr++ ;
			}
			// copy the required amount into our buffer
			// and release the clipboard
			memcpy ( clipbrd_ptr, clip_ptr, clipboard_size ) ;
			GlobalUnlock ( clipboard_handle ) ;
		}
		CloseClipboard() ;
	}
	clipboard_open = TRUE ;
	return TRUE ;

}

void close_clipboard_device ( void )
{
	HANDLE clipboard_handle ;
	unsigned char *clip_ptr ;

	// if clipboard being written to, copy contents of
	// clipboard buffer to clipboard
	if ( writing_clipboard )
	{
		if ( OpenClipboard ( Main_hWnd ) )
		{
			EmptyClipboard() ;
			// add terminating NULL
			clipboard_size++ ;
			*clip_wrt_ptr = 0 ;
			clipboard_handle = GlobalAlloc ( GHND, clipboard_size ) ;
			clip_ptr = (unsigned char*)GlobalLock ( clipboard_handle ) ;
			memcpy ( clip_ptr, clipbrd_ptr, clipboard_size ) ;
			GlobalUnlock ( clipboard_handle ) ;
			SetClipboardData ( CF_OEMTEXT, clipboard_handle ) ;
			CloseClipboard() ;
		}
	}

	clipboard_open = FALSE ;
	// free the clipboard buffer
	free ( clipbrd_ptr ) ;
}

char read_clipboard_device ( unsigned char *data_buffer, unsigned short *byte_count, int mode_flag )
{
	unsigned short local_count ;
	char cr_read ;
	unsigned char *bfr_ptr ;

	// if clipboard in write mode or size of data has been reached
	// return with fail
	if ( writing_clipboard || clip_count == clipboard_size )
	{
		*byte_count=0 ;
		return FALSE ;
	}

	cr_read = FALSE ;
	bfr_ptr = data_buffer ;
	for ( local_count=0 ; local_count<*byte_count ; local_count++ )
	{
		// do no more copying if clipboard size has been reached
		// but do not return an error as valid data will have been copied
		if ( clip_count == clipboard_size )
			*byte_count = local_count ;
		else
		{
			// copy data from clipboard buffer to o/p buffer
			*bfr_ptr = *clip_wrt_ptr ;

			if ( *bfr_ptr == 0xD )
				cr_read = TRUE ;

			// if doing ASCII txfrs, stop when a CR/LF is encountered
			if ( mode_flag == TEXT && *bfr_ptr == 0xA && cr_read )
				*byte_count = local_count+1 ;
			bfr_ptr++ ;
			clip_wrt_ptr++ ;
			clip_count++ ;
		}
	}
	return TRUE ;
}

char write_clipboard_device ( unsigned char *data_buffer, unsigned short *byte_count, int mode_flag )
{
	unsigned short local_count ;
	char cr_written ;
	unsigned char *bfr_ptr ;

	// if clipboard in read mode or max size reached, stop here
	if ( !writing_clipboard || clipboard_size == max_clipboard_size )
	{
		*byte_count=0 ;
		return FALSE ;
	}

	bfr_ptr = data_buffer ;
	cr_written = FALSE ;

	// copy data from buffer to clipboard buffer
	for ( local_count=0 ; local_count<*byte_count ; local_count++ )
	{
		if ( clipboard_size == max_clipboard_size )
			*byte_count=local_count ;
		else
		{
			*clip_wrt_ptr=*bfr_ptr ;

			if ( *bfr_ptr == 0xD )
				cr_written = TRUE ;
			// if doing ASCII txfrs, stop when a CR/LF is encountered
			if ( mode_flag == TEXT && *bfr_ptr == 0xA && cr_written )
				*byte_count = local_count+1 ;

			bfr_ptr++ ;
			clip_wrt_ptr++ ;
			clipboard_size++ ;
		}
	}
	return TRUE ;
}

// determine 'eof' for clipboard
char clipboard_device_eof ( void )
{
	if ( !writing_clipboard )
	{
		if ( clip_count < clipboard_size )
			return FALSE ;
		else
			return TRUE ;
	}
	return TRUE ;
}
