// common emulator functions
#include "globals.h"
#include "commemu.h"

// class constructor
common_emulator::common_emulator()
{
}

// all provided functions can be overwridden by a derived emulator class
int common_emulator::process_emulation_sequence ( char textcr )
{
	// call the standard function processor
	return process_comm_emulator_sequence ( textcr ) ;
}

int common_emulator::emulation_detect ( char textcr )
{
	// check for a common emulator sequence
	return comm_emulator_detect ( textcr ) ;
}

int common_emulator::emulation_in_progress ( void )
{
	// no common emulator functions are more than 1 character
	return FALSE ;
}

int common_emulator::emulation_sequence_pending ( void )
{
	// no common emulator functions are more than 1 character
	return NO_ACTION ;
}

int common_emulator::find_next_tab ( int current_x )
{
	const int def_tab_size = 8 ;

	// default tabs to 8 characters on from closist previous posn
	return (current_x / def_tab_size)*def_tab_size+def_tab_size ;
}

void common_emulator::modify_tab ( int column, char set_flag )
{
	// tab positioning not supported
}

int common_emulator::comm_emulator_detect ( char textcr )
{
	// check for standard special function characters
	if ( textcr == Bell || textcr == Backspace || textcr == Home )
		return TRUE ;
	if ( textcr == Carriage_Return || textcr == Line_Feed || textcr == Clear || textcr == Tab  )
		return TRUE ;

	 return FALSE ;
}

int common_emulator::process_comm_emulator_sequence ( char textcr )
{
	// common emulator sequences
	switch ( textcr )
	{
		case Carriage_Return :
			return Carriage_Return ;
		case Line_Feed :
			return Line_Feed ;
		case Bell :
			return Bell ;
		case Backspace :
			return Backspace ;
		case Home :
			return Home ;
		case Clear :
			return Clear ;
		case Tab :
			return Tab ;
	}
	return NO_ACTION ;
}