// CONFIG.c package
//CREATED ON September 22, 1995, FOR V2.01 BY J.BIRD
//MOD ON February 21, 1996, FOR V2.04 BY J.Bird
//MOD ON April 26 1996, FOR V2.1 BY J.Bird
//MOD ON May 30 1996, FOR V3.0 BY J.Bird
//MOD ON May 01 1997, FOR V3.01 BY J.Bird
//MOD ON March 08, 1998, FOR V4.0 BY J.Bird
//MOD ON June 21, 1998, FOR V3.11 BY J.Bird
//MOD ON May 20, 2000, FOR V4.1 BY J.Bird

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pcshare.h"
#include "config.h"
#include <commdlg.h>

BOOL CALLBACK _export SystemConfig (HWND, UINT, WPARAM, LPARAM);

void open_system_config ( void ) ;
void display_key_from_keymap ( HWND hDlg, struct keymap_type *keymap, int entry ) ;
void save_key_to_keymap ( HWND hDlg, struct keymap_type *keymap, int entry ) ;

FILE *config_file ;

struct config_info_type system_config ;
int valid_config ;
const char *RegCfgPath = "Config Path" ;

// user mappable keys
const char *keystrs[KEY_COUNT] =
							  { "F1",
								 "F2",
								 "F3",
								 "F4",
								 "F5",
								 "F6",
								 "F7",
								 "F8",
								 "F9",
								 "F10",
								 "F11",
								 "F12",
								 "Break",
								 "Insert",
								 "Home",
								 "Page Up",
								 "Page Down",
								 "End",
								 "Delete",
								 "BackSpace",
								 "Up Arrow",
								 "Down Arrow",
								 "Left Arrow",
								 "Right Arrow",
								 "Tab" } ;

// lookup table for virtual keyboard to PCshare udf keys
unsigned char vtkey_to_udfkey[256] ;

// matrix to map udf keys into virtual keycodes
const unsigned char udf_matrix[KEY_COUNT] = { VK_F1,
															 VK_F2,
															 VK_F3,
															 VK_F4,
															 VK_F5,
															 VK_F6,
															 VK_F7,
															 VK_F8,
															 VK_F9,
															 VK_F10,
															 VK_F11,
															 VK_F12,
															 VK_PAUSE,
															 VK_INSERT,
															 VK_HOME,
															 VK_PRIOR,
															 VK_NEXT,
															 VK_END,
															 VK_DELETE,
															 VK_BACK,
															 VK_UP,
															 VK_DOWN,
															 VK_LEFT,
															 VK_RIGHT,
															 VK_TAB } ;

BOOL CALLBACK _export SystemConfig (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
	char hex_num[10] ;
	BOOL valid ;

	switch ( Message )
	{
		case WM_INITDIALOG :

			if ( config_file == NULL )
			{
				MessageBox ( hDlg, "Config File Not Found", "PCShare", MB_OK | MB_ICONEXCLAMATION ) ;
				system_config.fore_colour = RGB ( 0,0,0 ) ;
				system_config.back_colour = RGB ( 255, 255, 255 ) ;
				system_config.base_addr = 0x310 ;
				system_config.irq = 7 ;
				system_config.dma = 1 ;
				system_config.hard_drives = 1 ;
				system_config.sector_size = 256 ;
				system_config.emulation = NO_EMULATION ;
				strcpy ( system_config.user_dir, "c:\\pc_share\\" ) ;
        // v4.0 - cmd prompt & delete key
				strcpy ( system_config.cmd_prompt, "OS9:" ) ;
        system_config.del_key = 0x8 ;

				// default font
				system_config.font.lfHeight         =  14;
				system_config.font.lfWidth          =  9;
				system_config.font.lfEscapement     =  0;
				system_config.font.lfOrientation    =  0;
				system_config.font.lfWeight         =  FW_NORMAL;
				system_config.font.lfItalic         =  FALSE;
				system_config.font.lfUnderline      =  FALSE;
				system_config.font.lfStrikeOut      =  FALSE;
				system_config.font.lfCharSet        =  ANSI_CHARSET;
				system_config.font.lfOutPrecision   =  OUT_DEFAULT_PRECIS;
				system_config.font.lfClipPrecision  =  CLIP_DEFAULT_PRECIS;
				system_config.font.lfQuality        =  DEFAULT_QUALITY;
				system_config.font.lfPitchAndFamily =  FIXED_PITCH | FF_DONTCARE;
				strcpy((char *)system_config.font.lfFaceName, "System") ;
			}
			else
			{
				fread ( &system_config, sizeof ( system_config ), 1, config_file ) ;
				fclose ( config_file ) ;
			}

			sprintf ( hex_num, "%x", system_config.base_addr ) ;
			SetDlgItemText ( hDlg, CONFIG_IOPORT, hex_num ) ;
			SetDlgItemInt ( hDlg, PORT_IRQ, system_config.irq, FALSE ) ;
			SetDlgItemInt ( hDlg, CONFIG_HDS, system_config.hard_drives, FALSE ) ;
			SetDlgItemInt ( hDlg, CONFIG_SECTOR, system_config.sector_size, FALSE ) ;
			SetDlgItemText ( hDlg, CONFIG_PATH, system_config.user_dir ) ;
      // v4.0 - cmd prompt & delete key
			SetDlgItemText ( hDlg, CONFIG_CMDPROMPT, system_config.cmd_prompt ) ;
			SetDlgItemInt ( hDlg, CONFIG_DEL, (int)system_config.del_key, FALSE ) ;

			switch ( system_config.emulation )
			{
				case NO_EMULATION :

					CheckDlgButton ( hDlg, EMUL_NONE, TRUE ) ;
					break ;

				case VT100_EMUL :

					CheckDlgButton ( hDlg, EMUL_VT100, TRUE ) ;
					break ;

				case GO51_EMUL :

					CheckDlgButton ( hDlg, EMUL_GO51, TRUE ) ;
					break ;
			}

			valid_config = TRUE ;
			return 1 ;

		case WM_COMMAND :

			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case IDOK :

					GetDlgItemText ( hDlg, CONFIG_IOPORT, hex_num, 5 ) ;
					sscanf ( hex_num, "%x", &system_config.base_addr ) ;
					system_config.irq = GetDlgItemInt ( hDlg, PORT_IRQ, &valid, FALSE ) ;
					system_config.hard_drives = GetDlgItemInt ( hDlg, CONFIG_HDS, &valid, FALSE ) ;
					system_config.sector_size = GetDlgItemInt ( hDlg, CONFIG_SECTOR, &valid, FALSE ) ;
					GetDlgItemText ( hDlg, CONFIG_PATH, system_config.user_dir, 190 ) ;
          // v3.11 bug fix - ensure path ends in a backslash
					if ( system_config.user_dir[strlen(system_config.user_dir)] != '\\' )
						strcat ( system_config.user_dir, "\\" ) ;
          // v4.0 - cmd prompt & delete key
					GetDlgItemText ( hDlg, CONFIG_CMDPROMPT, system_config.cmd_prompt, 79 ) ;
          system_config.del_key = GetDlgItemInt ( hDlg, CONFIG_DEL, &valid, FALSE ) ;
					if ( IsDlgButtonChecked ( hDlg, EMUL_NONE ) )
						system_config.emulation = NO_EMULATION ;
					if ( IsDlgButtonChecked ( hDlg, EMUL_VT100 ) )
						system_config.emulation = VT100_EMUL ;
					if ( IsDlgButtonChecked ( hDlg, EMUL_GO51 ) )
						system_config.emulation = GO51_EMUL ;
					save_config_data() ;
					EndDialog ( hDlg, 0 ) ;
					return 0 ;

				case IDCANCEL :

					EndDialog ( hDlg, 0 ) ;
					return 0 ;
			}
			break ;
	}
	return 0 ;
}

// display system config data
void display_config_data ( void )
{
	// open config file and display system configuration
	open_system_config() ;
	DialogBox ( hInst, MAKEINTRESOURCE ( SYSTEM_CONFIG ), Main_hWnd, (DLGPROC)SystemConfig ) ;
}

void get_config_data ( void )
{
	int i ;

	// open config file and read in. If file cannot be accessed, call the options dialog box
	// to load in system defaults
	open_system_config() ;
	if ( config_file == NULL )
		DialogBox ( hInst, MAKEINTRESOURCE ( SYSTEM_CONFIG ), Main_hWnd, (DLGPROC)SystemConfig ) ;
	else
	{
		fread ( &system_config, sizeof ( system_config ), 1, config_file ) ;
		fclose ( config_file ) ;
	}

	// fill in the keyboard lookup table
	for ( i=0 ; i< 256 ; i++ )
		vtkey_to_udfkey[i] = 0xff ;
	for ( i=0 ; i<KEY_COUNT ; i++ )
		vtkey_to_udfkey[udf_matrix[i]] = i ;
}

void open_system_config ( void )
{
	char cfg_path[200] ;
	char *cfg_ptr ;
  DWORD Spare  ;
  HKEY hKey ;
  DWORD Size ;

	// v4.1 - get path to config file from registry
  cfg_path[0] = 0 ;
  if ( RegCreateKeyEx ( HKEY_LOCAL_MACHINE,
                        REGPATH,
                        0,
                        NULL,
                        REG_OPTION_NON_VOLATILE,
                        KEY_QUERY_VALUE | KEY_SET_VALUE,
                        NULL,
                        &hKey,
                        &Spare ) == ERROR_SUCCESS )
  {
    Size = sizeof(cfg_path) ;
    RegQueryValueEx ( hKey,
                      (LPCTSTR)RegCfgPath,
                       NULL,
                       &Spare,
                       (LPBYTE)cfg_path,
                       &Size );
    RegCloseKey ( hKey ) ;
  }
  strcat ( cfg_path, "\\pcshare4.cfg" ) ;
  config_file = fopen ( cfg_path, "rb" ) ;
}

// save current config to disk
void save_config_data ( void )
{
	char filename[250] ;
  DWORD Spare  ;
  HKEY hKey ;
  DWORD Size ;

  // v4.1 - hold 'path' info in registry
  if ( RegCreateKeyEx ( HKEY_LOCAL_MACHINE,
                        REGPATH,
                        0,
                        NULL,
                        REG_OPTION_NON_VOLATILE,
                        KEY_QUERY_VALUE | KEY_SET_VALUE,
                        NULL,
                        &hKey,
                        &Spare ) == ERROR_SUCCESS )
  {
    RegSetValueEx ( hKey,
                    RegCfgPath,
                    0,
                    REG_SZ,
                    (CONST BYTE*)system_config.user_dir,
                    strlen(system_config.user_dir)+1 ) ;
    RegCloseKey ( hKey ) ;
  }

	strcpy ( filename, system_config.user_dir ) ;
	strcat ( filename, "pcshare4.cfg" ) ;
	config_file = fopen ( filename, "wb" ) ;
	// v3.11 bug fix - validate file handle
	if ( config_file == NULL )
		MessageBox ( Main_hWnd, "Error writing config file", "PCShare", MB_OK | MB_ICONEXCLAMATION ) ;
	else
	{
		fwrite ( &system_config, sizeof ( system_config ), 1, config_file ) ;
		fclose ( config_file ) ;
	}
}

// get new font info and font colour
BOOL get_new_font ( HWND hWnd, LOGFONT *lf, COLORREF *text_colour )
{
	CHOOSEFONT cf ;

	cf.lStructSize = sizeof ( cf ) ;
	cf.hwndOwner = hWnd ;
	cf.lpLogFont = lf ;
	cf.rgbColors = *text_colour ;
	cf.Flags = CF_EFFECTS | CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT | CF_FIXEDPITCHONLY ;
	cf.nFontType = SCREEN_FONTTYPE ;

	if ( ChooseFont ( &cf ) )
   {
		*text_colour = cf.rgbColors ;
      return TRUE ;
   }
   else
   	return FALSE ;
}

// get new colour
BOOL get_new_colour ( HWND hWnd, COLORREF *colour )
{
	CHOOSECOLOR new_clut ;
	COLORREF crTemp[16] = {0} ;  // Important, sometimes unused, array

	new_clut.lStructSize = sizeof( CHOOSECOLOR );
	new_clut.hwndOwner = hWnd ;
	new_clut.hInstance = hInst ;
	new_clut.rgbResult = *colour ;

/*
lpCustColors must be set to a valid array of 16 COLORREF's, even if it
is not used.  If it isn't you will probably fail with a GP fault in
COMMDLG.DLL.
*/
	new_clut.lpCustColors = crTemp;
	new_clut.Flags = CC_RGBINIT;
	new_clut.lCustData = 0L;
	new_clut.lpfnHook = NULL;
	new_clut.lpTemplateName = NULL;

	if ( ChooseColor( &new_clut ) )
   {
		*colour = new_clut.rgbResult ;
      return TRUE ;
   }
   else
   	return FALSE ;
}

// keyboard map dialog
BOOL CALLBACK _export DefineKeyMap (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
	int i ;
	static int entry ;
	static struct keymap_type keymap ;
   static struct keymap_type *inkeys ;

	switch ( Message )
	{
		case WM_INITDIALOG :

			inkeys = (struct keymap_type*)lParam ;
      	memcpy ( &keymap, inkeys, sizeof ( keymap ) ) ;
			// clear out listbox
			SendDlgItemMessage ( hDlg, KEYS_LIST, LB_RESETCONTENT, 0, 0 ) ;
			// load listbox with user definable keys
			for (i=0 ; i<KEY_COUNT ; i++ )
				SendDlgItemMessage ( hDlg, KEYS_LIST, LB_ADDSTRING, 0, (LPARAM) ((LPSTR) keystrs[i] ) ) ;
			// set initial entries
			SendDlgItemMessage ( hDlg, KEYS_LIST, LB_SETCURSEL, 0, 0 ) ;
			entry=0 ;
			display_key_from_keymap ( hDlg, &keymap, entry ) ;
			return 1 ;

		case WM_COMMAND :

			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case IDOK :

					// save current entry to keymap
					save_key_to_keymap ( hDlg, &keymap, entry ) ;
					memcpy ( inkeys, &keymap, sizeof ( keymap ) ) ;
					EndDialog ( hDlg, TRUE ) ;
					return 0 ;

				case IDCANCEL :

					// return
					EndDialog ( hDlg, FALSE ) ;
					return 0 ;

				case KEYS_LIST :

					// if new entry selected or just selected OK button
					if ( GET_WM_COMMAND_CMD ( wParam, lParam ) == LBN_SELCHANGE )
					{
						// save current entry to keymap
						save_key_to_keymap ( hDlg, &keymap, entry ) ;
						// get new entry
						entry = SendDlgItemMessage ( hDlg, KEYS_LIST, LB_GETCURSEL, 0, 0 ) ;
						// display entries from keymap
						display_key_from_keymap ( hDlg, &keymap, entry ) ;
					}
					break ;
			}
	}
	return 0 ;
}

// loads the requested keymap entry into the dialog box
void display_key_from_keymap ( HWND hDlg, struct keymap_type *keymap, int entry )
{
	SetDlgItemInt ( hDlg, NORMAL_CODE_1, keymap->keys[NORMAL_CHAR][entry][0], FALSE ) ;
	SetDlgItemInt ( hDlg, NORMAL_CODE_2, keymap->keys[NORMAL_CHAR][entry][1], FALSE ) ;
	SetDlgItemInt ( hDlg, NORMAL_CODE_3, keymap->keys[NORMAL_CHAR][entry][2], FALSE ) ;
	SetDlgItemInt ( hDlg, NORMAL_CODE_4, keymap->keys[NORMAL_CHAR][entry][3], FALSE ) ;
	SetDlgItemInt ( hDlg, SHIFT_CODE_1, keymap->keys[SHIFTED][entry][0], FALSE ) ;
	SetDlgItemInt ( hDlg, SHIFT_CODE_2, keymap->keys[SHIFTED][entry][1], FALSE ) ;
	SetDlgItemInt ( hDlg, SHIFT_CODE_3, keymap->keys[SHIFTED][entry][2], FALSE ) ;
	SetDlgItemInt ( hDlg, SHIFT_CODE_4, keymap->keys[SHIFTED][entry][3], FALSE ) ;
	SetDlgItemInt ( hDlg, CTRL_CODE_1, keymap->keys[CONTROL][entry][0], FALSE ) ;
	SetDlgItemInt ( hDlg, CTRL_CODE_2, keymap->keys[CONTROL][entry][1], FALSE ) ;
	SetDlgItemInt ( hDlg, CTRL_CODE_3, keymap->keys[CONTROL][entry][2], FALSE ) ;
	SetDlgItemInt ( hDlg, CTRL_CODE_4, keymap->keys[CONTROL][entry][3], FALSE ) ;
}

// copies the keymap information from the dialog box into a keymap entry
void save_key_to_keymap ( HWND hDlg, struct keymap_type *keymap, int entry )
{
	BOOL ferr ;

	keymap->keys[NORMAL_CHAR][entry][0] =
		GetDlgItemInt ( hDlg, NORMAL_CODE_1, &ferr, FALSE ) ;
	keymap->keys[NORMAL_CHAR][entry][1] =
		GetDlgItemInt ( hDlg, NORMAL_CODE_2, &ferr, FALSE ) ;
	keymap->keys[NORMAL_CHAR][entry][2] =
		GetDlgItemInt ( hDlg, NORMAL_CODE_3, &ferr, FALSE ) ;
	keymap->keys[NORMAL_CHAR][entry][3] =
		GetDlgItemInt ( hDlg, NORMAL_CODE_4, &ferr, FALSE ) ;
	keymap->keys[SHIFTED][entry][0] =
		GetDlgItemInt ( hDlg, SHIFT_CODE_1, &ferr, FALSE ) ;
	keymap->keys[SHIFTED][entry][1] =
		GetDlgItemInt ( hDlg, SHIFT_CODE_2, &ferr, FALSE ) ;
	keymap->keys[SHIFTED][entry][2] =
		GetDlgItemInt ( hDlg, SHIFT_CODE_3, &ferr, FALSE ) ;
	keymap->keys[SHIFTED][entry][3] =
		GetDlgItemInt ( hDlg, SHIFT_CODE_4, &ferr, FALSE ) ;
	keymap->keys[CONTROL][entry][0] =
		GetDlgItemInt ( hDlg, CTRL_CODE_1, &ferr, FALSE ) ;
	keymap->keys[CONTROL][entry][1] =
		GetDlgItemInt ( hDlg, CTRL_CODE_2, &ferr, FALSE ) ;
	keymap->keys[CONTROL][entry][2] =
		GetDlgItemInt ( hDlg, CTRL_CODE_3, &ferr, FALSE ) ;
	keymap->keys[CONTROL][entry][3] =
		GetDlgItemInt ( hDlg, CTRL_CODE_4, &ferr, FALSE ) ;
}

