// FILE_HANDLER.c package
//CREATED ON October 8, 1995, FOR V2.03 BY J.BIRD
//MOD ON March 24 1996, FOR V2.1 BY J.Bird
//MOD ON October 04 1996, FOR V2.2 BY J.Bird
//MOD ON November 24 1996, FOR V3.0 BY J.Bird
//MOD ON October 29 1997, FOR V3.02 BY J.Bird
//MOD ON November 21 1997, FOR V3.1 BY J.Bird
#include <windows.h>
#include <stdio.h>
#include <dir.h>
#include <io.h>
#include <dos.h>
#include <string.h>
#include <ctype.h>
#include <alloc.h>
#include <dirent.h>
#include "pcshare.h"
#include "file_han.h"
#include "rs232.h"
#include "clip_io.h"

#define TOTAL_FILES 10

// V2.2 - fix minor bugs and add support for serial devices (COMx)
// V3.0 - add support for windows clipboard, change RS232 i/f to Win API
// V3.1 - read directory protocol change for win95
#define UNUSED 0
#define FILE_IO 1
#define SERIAL_IO 2
#define CLIPBRD_IO 3
#define DIR_IO 4

FILE *access_table[TOTAL_FILES] ;

struct dir_table_type { DIR *entry ;
								unsigned short entry_no ; } ;

struct dir_table_type dir_table[TOTAL_FILES] ;

int in_use[TOTAL_FILES] = { UNUSED } ;
// serial configs
serial_232_t com_config[4] ;

// open/create filename specified and assign it an FCB no.
char open_file ( char *filename, char *access_string, char *fcb )
{
	int i ;
	char local_fcb ;
	int fcb_found = FALSE ;
	int port_id ;
	BOOL success_flag ;

	// look for free FCB in local table
	for ( i=0;i<TOTAL_FILES;i++)
		if ( !in_use[i] && !fcb_found )
		{
			local_fcb = i ;
			*fcb = local_fcb ;
			fcb_found = TRUE ;
		}

	if ( !fcb_found )
		return FALSE ;

	// V2.2 additional - check for a COM serial device
	// and open accordingly
	if ( ( strncmpi ( filename, "COM", 3) == 0 ) && ( filename[4] == ':' ) )
	{
		// identify com port from filename
		port_id = filename[3]-'1' ;
		if ( init_rs232 ( &com_config[port_id] ) )
		{
			in_use[local_fcb] = SERIAL_IO ;
			return TRUE ;
		}
		else
			return FALSE ;
	}

	// V3.0 additional - check for clipboard
	if ( strncmpi ( filename, "CLIP:", 5) == 0 )
	{
		// open clipboard device for read or writing
		if ( access_string[1] == '+' || access_string[0] == 'w' )
			success_flag = open_clipboard_device ( TRUE ) ;
		else
			success_flag = open_clipboard_device ( FALSE ) ;
		// mark fcb as used by clipboard
		if ( success_flag )
			in_use[local_fcb] = CLIPBRD_IO ;
		return success_flag ;
	}

	// V3.1 additional - check for directory
	if ( strncmpi ( access_string, "rdir", 4 ) == 0 )
	{
		// open directory, and validate result
		dir_table[local_fcb].entry = opendir ( filename ) ;
		if ( dir_table[local_fcb].entry == NULL )
			return FALSE ;
		else
		{
			in_use[local_fcb] = DIR_IO ;
			dir_table[local_fcb].entry_no = 0 ;
			return TRUE ;
		}
	}

	// perform file operation
	if ( ( access_table[local_fcb] = fopen ( filename, access_string ) ) == NULL )
		return FALSE ;
	else
	{
		in_use[local_fcb] = FILE_IO ;
		return TRUE ;
	}
}

char change_dir ( char *dirname )
{
	char *path_ptr ;
	int drivenum ;

	// check for colon in drive separator
	path_ptr = strchr ( dirname, ':' ) ;

	// if drive specified, do drive change first
	if ( path_ptr != NULL )
	{
		// calculate drive number from letter & change drive
		drivenum = toupper ( dirname[0] ) - 65 ;
		setdisk ( drivenum ) ;
		if ( getdisk() != drivenum )
			return FALSE ;

		// increment pointer to point to remainder of pathlist
		path_ptr++ ;
	}
	else
		path_ptr = dirname ;

	// check there is a pathname to process
	if ( strlen ( path_ptr ) != 0 )
		if ( chdir ( dirname ) == 0 )
			return TRUE ;
		else
			return FALSE ;

	return TRUE ;
}

char close_file ( char fcb )
{
	switch ( in_use[fcb] )
	{
		case FILE_IO :
			fclose ( access_table[fcb] ) ;
			break ;

		case SERIAL_IO :
			// V2.2 - close COM port
			de_init_rs232() ;
			break ;

		case CLIPBRD_IO :

			// V3.0, clipboard handling
			// if clipboard bfr has just been written
			// txfr it to the windows clipboard
			close_clipboard_device() ;
			break ;

		case DIR_IO :

			// V3.1 - close directory
			closedir( dir_table[fcb].entry ) ;
			break ;
	}

	in_use[fcb] = UNUSED ;
	return TRUE ;
}

char delete_file ( char *filename )
{
	if ( remove ( filename ) == 0 )
		return TRUE ;
	else
		return FALSE ;
}

char make_dir ( char *dirname )
{
	if ( mkdir ( dirname ) == 0 )
		return TRUE ;
	else
		return FALSE ;
}

// read block from open file as text or binary
char read_file ( char fcb, unsigned char *data_buffer, unsigned short *byte_count, int mode_flag )
{
unsigned short local_count ;
char success, cr_read ;
unsigned char *bfr_ptr ;

	switch ( in_use[fcb] )
	{
		case FILE_IO :

			// in binary mode, use fread to get block of data
			if ( mode_flag == BINARY )
			{
				local_count = fread ( data_buffer, 1, *byte_count, access_table[fcb] ) ;

				// if incorrect byte count read, mark block as failed & return
				if ( local_count != *byte_count )
					success = FALSE ;
				else
					success = TRUE ;

				*byte_count = local_count ;
			}
			else
			{
				if ( ( fgets ( data_buffer, *byte_count, access_table[fcb] ) ) == NULL )
				{
					// if fgets failed then must be error or at EOF, byte count will be 0
					// won't return EOF unless it's already there
					success = FALSE ;
					*byte_count = 0 ;
				}
				else
				{
					// update byte count on success
					success = TRUE ;
					*byte_count = strlen ( data_buffer ) ;
				}
			}
			return success ;

		case SERIAL_IO :

			bfr_ptr = data_buffer ;
			cr_read = FALSE ;
			for ( local_count=0 ; local_count<*byte_count ; local_count++ )
			{
				// get char from 232 port
				*bfr_ptr = read_rs232() ;

				if ( *bfr_ptr == 0xD )
					cr_read = TRUE ;

				// if doing ASCII txfrs, stop when a CR/LF is encountered
				if ( mode_flag == TEXT && *bfr_ptr == 0xA && cr_read )
					*byte_count = local_count+1 ;
				bfr_ptr++ ;
			}

			return TRUE ;

		case CLIPBRD_IO :

			// V3.0 - clipboard access
			return read_clipboard_device ( data_buffer, byte_count, mode_flag ) ;

		default :
			return FALSE ;
	}
}

char write_file ( char fcb, unsigned char *data_buffer, unsigned short *byte_count, int mode_flag )
{
unsigned short local_count ;
char success, cr_written ;
unsigned char *bfr_ptr ;

	switch ( in_use[fcb] )
	{
		case FILE_IO :

			// in binary mode, use fwrite to write data
			if ( mode_flag == BINARY )
			{
				local_count = fwrite ( data_buffer, 1, *byte_count, access_table[fcb] ) ;

				// if incorrect byte count read, mark block as failed & return
				if ( local_count != *byte_count )
					success = FALSE ;
				else
					success = TRUE ;

				*byte_count = local_count ;
			}
			else
			{
				if ( ( fputs ( data_buffer, access_table[fcb] ) ) == EOF )
				{
					success = FALSE ;
					*byte_count = 0 ;
				}
				else
				{
					// update byte count on success
					success = TRUE ;
					*byte_count = strlen ( data_buffer ) ;
				}
			}
			return success ;

		case SERIAL_IO :

			bfr_ptr = data_buffer ;
			cr_written = FALSE ;

			for ( local_count=0 ; local_count<*byte_count ; local_count++ )
			{
				write_rs232( *bfr_ptr ) ;
				if ( *bfr_ptr == 0xD )
					cr_written = TRUE ;

				// if doing ASCII txfrs, stop when a CR/LF is encountered
				if ( mode_flag == TEXT && *bfr_ptr == 0xD && cr_written )
					*byte_count = local_count+1 ;
				bfr_ptr++ ;
			}
			return TRUE ;

		case CLIPBRD_IO :

			// V3.0 - add clipboard access
			return write_clipboard_device ( data_buffer, byte_count, mode_flag ) ;

		default :
			return FALSE ;
	}
}

char seek_file ( char fcb, long file_ptr )
{
	switch ( in_use[fcb] )
	{
		case FILE_IO :

			if ( fseek ( access_table[fcb], file_ptr, SEEK_SET ) == 0 )
				return TRUE ;
			else
				return FALSE ;

		default :
			return FALSE ;
	}
}

char get_file_size ( char fcb, long *file_size )
{
	int f_handle ;
	long l_file_size ;

	switch ( in_use[fcb] )
	{

		case FILE_IO :

		// get the file handle associated with this file
		f_handle = fileno ( access_table[fcb] ) ;

		l_file_size = filelength ( f_handle ) ;
		if ( l_file_size == -1L )
			return FALSE ;
		else
		{
			*file_size = l_file_size ;
			return TRUE ;
		}

		case SERIAL_IO :

			// for serial io return chars in buffer
			*file_size = rs232_ready() ;
			return TRUE ;

		case CLIPBRD_IO :

			// V3.0 - return size of data on clipboard or quantity written so far
			*file_size = clipboard_size ;
			return TRUE ;

		default :
			return FALSE ;
	}
}

char get_file_posn ( char fcb, long *file_posn )
{
	long l_file_posn ;

	switch ( in_use[fcb] )
	{
		case FILE_IO :

			l_file_posn = ftell ( access_table[fcb] ) ;
			if ( l_file_posn != -1L )
			{
				*file_posn = l_file_posn ;
				return TRUE ;
			}
			else
				return FALSE ;

		case CLIPBRD_IO :

			// V3.0 - return clipboard size so far
			*file_posn = clipboard_size ;
			return TRUE ;

		default :
			return FALSE ;
	}
}

char end_of_file ( char fcb )
{
	switch ( in_use[fcb] )
	{
		case FILE_IO :

			if ( feof ( access_table[fcb] ) == 0 )
				return FALSE ;
			else
				return TRUE ;

		case SERIAL_IO :

			// V2.2 - serial devices never return EOF
			return FALSE ;

		case CLIPBRD_IO :

			// V3.0 - return EOF dependent on whether all
			// the data on the clipboard has been read yet
			return clipboard_device_eof() ;

		default :
			return TRUE ;
	}
}

char set_file_size ( char fcb, long file_size )
{
	int f_handle ;

	if ( in_use[fcb] != FILE_IO )
		return FALSE ;

	f_handle = fileno ( access_table[fcb] ) ;
	if ( chsize ( f_handle, file_size ) == 0 )
		return TRUE ;
	else
		return FALSE ;
}

// new in v2.1
// protocol change for V3.1
char read_dir_entry ( char fcb, unsigned short entry_no, char *entryname )
{
	struct dirent *entryinfo ;
	unsigned short i ;

	// if entry no not next in sequence, rewind and move
	// to required entry
	if ( entry_no != dir_table[fcb].entry_no )
	{
		dir_table[fcb].entry_no = 0 ;
		rewinddir ( dir_table[fcb].entry ) ;
		for ( i=0 ; i<entry_no ; i++ )
		{
			entryinfo = readdir ( dir_table[fcb].entry ) ;
			dir_table[fcb].entry_no++ ;
		}
	}

	// read the entry & update current entry no
	entryinfo = readdir ( dir_table[fcb].entry ) ;
	dir_table[fcb].entry_no++ ;

	// validate response and return
	if ( entryinfo == NULL )
		return FALSE ;
	else
	{
		strcpy ( entryname, entryinfo->d_name ) ;
		return TRUE ;
	}
}

char delete_dir ( char *dirname )
{
	return !(rmdir ( dirname ) ) ;
}

char get_current_dir ( char *dirname )
{
	if (getcwd ( dirname, 200 ) == NULL )
		return FALSE ;
	else
		return TRUE ;
}

// new in V2.2
char device_ready ( char fcb )
{
	// has no meaning for non-serial io
	if ( in_use[fcb] != SERIAL_IO )
		return FALSE ;
	else
	{
		if ( rs232_ready() == 0 )
			return FALSE ;
		else
			return TRUE ;
	}
}

void set_com_config ( int port_id, int baud_id, int data_bits, int parity_id,
	int stop_bits )
{
	com_config[port_id-1].comport = port_id ;
	com_config[port_id-1].irq = 0 ;
	com_config[port_id-1].baud = baud_id ;
	com_config[port_id-1].bits = data_bits ;
	com_config[port_id-1].parity = parity_id ;
	com_config[port_id-1].stops = stop_bits ;
}


