/*
go51 emulation package
*/
//CREATED ON September 22, 1995, FOR V2.01 BY J.BIRD
//MOD ON October 29 1996, FOR V2.03 BY J.Bird
//MOD ON February 16 1996, FOR V2.1 BY J.Bird
//MOD ON October 26 1996, FOR V3.0 BY J.Bird
#include "globals.h"
#include "commemu.h"
#include "go51_emu.h"
/*
Types/Data
*/
#define NORMAL 0
#define ESC_DETECT 1
#define CURSOR_SEQ_X 2
#define CURSOR_SEQ_Y 3

#define Escape 27

go51_emulator::go51_emulator()
{
	go51_level = NORMAL ;
}

// override the common emulator detection code
// to check for GO51 specifics as well
int go51_emulator::emulation_detect ( char textcr )
{
	// check for a common emulator sequence or escape character
	if ( comm_emulator_detect ( textcr ) || textcr == Escape )
		 return TRUE ;

	 return FALSE ;
}

int go51_emulator::emulation_in_progress ( void )
{
	 if ( go51_level == NORMAL )
		 return FALSE ;
	 else
		  return TRUE ;
}

int go51_emulator::process_emulation_sequence ( char textcr )
{
	switch ( go51_level )
	{
		case NORMAL :
		{
			// check for escape or common emulation sequence
			switch ( textcr )
			{
				case Escape :
					go51_level = ESC_DETECT ;
					return NO_ACTION ;
				default :
					return process_comm_emulator_sequence ( textcr ) ;
			}
		}

		case ESC_DETECT :
		{
			switch ( textcr )
			{
				case 'A' :
					go51_level = CURSOR_SEQ_X ;
					return NO_ACTION ;

				case 'B' :
					go51_level = NORMAL ;
					return CLEAR_EOL ;

				case 'C' :
					go51_level = NORMAL ;
					new_x = 1 ;
					return CURSOR_RIGHT ;

				case 'D' :
					go51_level = NORMAL ;
					new_y = 1 ;
					return CURSOR_UP ;

				case 'E' :
					go51_level = NORMAL ;
					new_y = 1 ;
					return CURSOR_DOWN ;

				case 'F' :
					go51_level = NORMAL ;
					return REVERSE_ON ;

				case 'G' :
					go51_level = NORMAL ;
					return REVERSE_OFF ;

				case 'H' :
					go51_level = NORMAL ;
					return UNDERLINE_ON ;

				case 'I' :
					go51_level = NORMAL ;
					return UNDERLINE_OFF ;

				case 'J' :
					go51_level = NORMAL ;
					return CLEAR_EOS ;

				case 'K' :
					go51_level = NORMAL ;
					return SCROLL_DOWN ;

				case 'L' :
					go51_level = NORMAL ;
					return SCROLL_UP ;

				default :
					go51_level = NORMAL ;
					return NO_ACTION ;
			}
		}

		case CURSOR_SEQ_X :
			new_x = textcr ;
			go51_level = CURSOR_SEQ_Y ;
			return NO_ACTION ;

		case CURSOR_SEQ_Y :
			new_y = textcr ;
			go51_level = NORMAL ;
			return SET_CURSOR ;
	}
	return NO_ACTION ;
}

