/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PACKAGE GRAPHICS
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON May 06, 1998 FOR V3.11 BY J.Bird
MOD ON June 21, 1998 FOR V4.00 BY J.Bird

Package Video_Ctrl.Graphics is

DESCRIPTION

This object is responsible for the execution of requested graphics commands on
the PC. It's primary function is to evaluate the graphics parameters passed to
each routine (performing reduction, calculation of default parameters etc),
and then call a language function or procedure to perform the physical
graphics operation.

Requests to utilise current/default X & Y co-ordinates and colours as defined
by the Data Interface Description Document, ref 106-4-12, Issue 1, February
1995 are converted from the default specification to 'real' data by this
object.

REQUIRED INTERFACE
 None
*/
#include <windows.h>
#include <alloc.h>
#include <math.h>
#include "pcshare.h"
#include "graphics.h"
/*
Types/Data

-- current settings
CURRENT_X      : integer := 0 ;
CURRENT_Y      : integer := 0 ;
CURRENT_COLOUR : integer := 0 ;

CURRENT_MODE   : integer := 0 ;

-- Valid modes, & X, Y ranges, reduction/multiply factors
-- identifies a type for specifying data ranges
type RANGE_TYPE is
 MIN : integer ;
 MAX : integer ;
end ;

-- 1 of these per mode giving a valid X/Y & colour range & any resolution
factor to apply to incoming co-ords.
type MODE_DESCRIPTOR_TYPE is
 X_RANGE : RANGE_TYPE ;
 Y_RANGE : RANGE_TYPE ;
 COLOUR_RANGE : RANGE_TYPE ;
 X_FACTOR : INTEGER ;
 Y_FACTOR : INTEGER ;
end ;
*/
#define DEFAULT_X 0xFFFF
#define DEFAULT_Y 0xFFFF
#define DEFAULT_COL 0xFF

// v4.0 - pen styles
#define GFX_PEN_SOLID 0
#define GFX_PEN_DASH 1
#define GFX_PEN_DOT 2

// fixed 16 colour palette maps to std VGA colours
#define PALETTE_ENTRIES 256
#define BASE_PALETTE_ENTRIES 16
const COLORREF fpalette[] = 					 { RGB(0,0,0),      // black
															RGB(0,0,255),    // blue
															RGB(0,255,0),    // green
															RGB(0,255,255),  // cyan
															RGB(255,0,0),    // red
															RGB(255,0,255),  //magenta
															RGB(128,64,64),  // brown
															RGB(192,192,192),// l/grey
															RGB(128,128,128),// d/grey
															RGB(0,128,255),  // l/blue
															RGB(0,255,128),  // l/green
															RGB(128,255,255),// l/cyan
															RGB(255,128,128),// l/red
															RGB(255,128,192),// l/magenta
															RGB(255,255,0),  // yellow
															RGB(255,255,255)} ; // white
static int modules_active = 0 ;
int bx ;

/*
Provides Operations
 MODE_CHANGE
 CURSOR_UPDATE
 SET_POINT
 COLOUR_CHANGE
 CLEAR_SCREEN
 DRAW_LINE
 DRAW_CIRCLE
 READ_PIXEL
*/

/*
Internal Operations
 EVALUATE_XY_COORDS
 EVALUATE_COLOUR

END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION EVALUATE_XY_COORDS
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Function EVALUATE_XY_COORDS ( X_IN, Y_IN ) return X_OUT,Y_OUT
*/
unsigned graphics_support::evaluate_x ( unsigned x_in )
/*
DESCRIPTION

This operation performs the processing of the specified co-ordinates by:

-- determining if current values have been requested, and replacing them with
the current values.
-- performing any resolution multiply that is required.
-- ensures the values are in range.
-- updates the default values with the new co-ordinates.

USES OPERATIONS
 None

Types/Data
 None
*/
{
unsigned x_out ;
/*
PSUEDO CODE

 if X_In = Default_Coord_Request
 then

	  X_Out := Current_X ;

 else

	  X_Out := X_In * Graphics_Modes ( Current_Mode ).X_Factor ;

	  if X_Out > Graphics_Modes ( Current_Mode ).X_Range.Max
	  then

			X_Out := Graphics_Modes ( Current_Mode ).X_Range.Max ;

	  end if ;

	  if X_Out < Graphics_Modes ( Current_Mode ).X_Range.Min
	  then

			X_Out := Graphics_Modes ( Current_Mode ).X_Range.Min ;

	  end if ;

 end if ;

 Current_X := X_Out ;

 if Y_In = Default_Coord_Request
 then

	  Y_Out := Current_Y ;

 else

	  Y_Out := Y_In * Graphics_Modes ( Current_Mode ).Y_Factor ;

	  if Y_Out > Graphics_Modes ( Current_Mode ).Y_Range.Max
	  then

			Y_Out := Graphics_Modes ( Current_Mode ).Y_Range.Max ;

	  end if ;

	  if Y_Out < Graphics_Modes ( Current_Mode ).Y_Range.Min
	  then

			Y_Out := Graphics_Modes ( Current_Mode ).Y_Range.Min ;

	  end if ;

 end if ;

 Current_Y := Y_Out ;

 return X_Out, Y_Out

CODE
*/
		if ( x_in == DEFAULT_X )
				return current_x ;
		else
		{
			if ( x_scaling_factor > 0 )
				x_out = x_in * x_scaling_factor ;
			else
				x_out = max_x + x_in * x_scaling_factor ;

				if ( x_out > max_x )
				{
						current_x = max_x ;
						return ( max_x ) ;
				}
				else
				{
					current_x = x_out ;
					return x_out ;
				}
		}
}
/*
END.
*/

unsigned graphics_support::evaluate_y ( unsigned y_in )
{
	int y_out ;

	if ( y_in == DEFAULT_Y )
		return current_y ;
	else
	{
		if ( y_scaling_factor > 0 )
			y_out = y_in * y_scaling_factor ;
		else
			y_out = max_y + y_in * y_scaling_factor ;

		if ( y_out > max_y )
		{
			current_y = max_y ;
			return ( max_y ) ;
		}
		else
		{
			current_y = y_out ;
			return y_out ;
		}
	}
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION EVALUATE_COLOUR
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON June 21, 1998 FOR V4.0 BY J.Bird

Function Evaluate_Colour ( In_Colour ) return Out_Colour is
*/
COLORREF graphics_support::evaluate_colour ( unsigned char in_colour )
/*
DESCRIPTION

This operation performs colour evaluation on the requested colour by:

-- Determining if the current colour has been requested, and replacing it with
the 'real' one.

-- Ensuring the colour is in range for the given mode.

USES OPERATIONS
 None

Types/Data
 None
PSUEDO CODE

 if In_Colour = Default_Colour_Request
 then

	  Out_Colour := Current_Colour ;

 else

	  if In_Colour > Graphics_Modes ( Mode_Id ).Colour_Range.Max
	  or In_Colour < Graphics_Modes ( Mode_Id ).Colour_Range.Min
	  then

				 Out_Colour := Current_Colour ;

	  end if ;

 end if ;

 return Out_Colour ;

CODE
*/
{
	// if requested colour is 'use default' then
	// return the current default colour
	if ( in_colour != DEFAULT_COL )
	  current_col = RGB ( pal_data->palPalEntry[in_colour].peRed,
                        pal_data->palPalEntry[in_colour].peGreen,
                        pal_data->palPalEntry[in_colour].peBlue ) ;

	return current_col ;
}
/*
END.
*/

// new v4.0 - pen creation
HPEN graphics_support::CreateReqdPen ( unsigned char width, unsigned char style, unsigned char col )
{
  switch ( style )
  {
    case GFX_PEN_SOLID :
    default :

	    return CreatePen ( PS_SOLID, width, evaluate_colour ( col ) ) ;

    case GFX_PEN_DASH :

	    return CreatePen ( PS_DASH, width, evaluate_colour ( col ) ) ;

    case GFX_PEN_DOT :

	    return CreatePen ( PS_DOT, width, evaluate_colour ( col ) ) ;
  }
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION MODE_CHANGE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON June 21, 1998 FOR V4.0 BY J.Bird

Procedure Mode_Change ( Mode_Id ) is
*/
graphics_support::graphics_support ( HDC hDC, unsigned x_pixels, unsigned y_pixels, float x_scale, float y_scale )
/*
DESCRIPTION

This operation initiates the requested graphics mode change, and sets the
current mode data.

USES OPERATIONS
 None

Types/Data
 None
PSUEDO CODE

 if  Mode_Id <= Graphics_Modes ( first )
 and Mode_Id >= Graphics_Modes ( last )
 then

		 Perform_Graphics_Mode_Change ( Mode_Id ) ;

		 Current_Mode := Mode_Id ;

 end if ;

CODE
*/
{
	int i ;

	// init object with parameters
	x_scaling_factor = x_scale ;
	y_scaling_factor = y_scale ;
	max_x = x_pixels*fabs(x_scale) ;
	max_y = y_pixels*fabs(y_scale) ;

	// create a bitmap for the required graphics region
	gbitmap = CreateCompatibleBitmap ( hDC, max_x, max_y ) ;

	// create a memory device context for the graphics rtn
	GhDC = CreateCompatibleDC ( hDC ) ;
	oldbmap = SelectObject ( GhDC, gbitmap ) ;
  SetBkMode ( GhDC, TRANSPARENT	 ) ;
  
	// initialise variables
	current_x = 0 ;
	current_y = 0 ;
	current_col = fpalette[0] ;

	// create palette for graphics window
	pal_data = (LOGPALETTE*)malloc ( sizeof ( LOGPALETTE ) + PALETTE_ENTRIES * sizeof (
		PALETTEENTRY ) ) ;
	pal_data->palVersion = 0x300 ;
	pal_data->palNumEntries = PALETTE_ENTRIES ;
	// enter the fixed base palette values
	for ( i=0 ; i<BASE_PALETTE_ENTRIES ; i++ )
	{
		pal_data->palPalEntry[i].peRed = GetRValue ( fpalette[i] ) ;
		pal_data->palPalEntry[i].peBlue = GetBValue ( fpalette[i] ) ;
		pal_data->palPalEntry[i].peGreen = GetGValue ( fpalette[i] ) ;
		pal_data->palPalEntry[i].peFlags = NULL ;
	}
	// initialise all others to black
	for ( i=BASE_PALETTE_ENTRIES ; i<PALETTE_ENTRIES ; i++ )
	{
		pal_data->palPalEntry[i].peRed = 0 ;
		pal_data->palPalEntry[i].peBlue = 0 ;
		pal_data->palPalEntry[i].peGreen = 0 ;
		pal_data->palPalEntry[i].peFlags = NULL ;
	}
	palette = CreatePalette ( pal_data ) ;
	old_palette = SelectPalette ( GhDC, palette, 0 ) ;
	clear_window ( 15 ) ;

	// increment count of modules in use
	modules_active++ ;
}
/*
END.
*/

graphics_support::~graphics_support()
{
	// clean up
	SelectObject ( GhDC, oldbmap ) ;
	SelectPalette ( GhDC, old_palette, 0 ) ;
	DeleteObject ( gbitmap ) ;
	DeleteObject ( palette ) ;
  free ( pal_data ) ;
	DeleteDC ( GhDC ) ;
	modules_active-- ;
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION CURSOR_UPDATE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Cursor_Update ( X, Y ) is
*/
void graphics_support::cursor_update ( unsigned x, unsigned y )
/*
DESCRIPTION

This operation updates the current co-ordinates with those supplied. The
Evaluate_XY_Coords operation is used which will setup the current co-ords. The
return values are discarded.

USES OPERATIONS
 EVALUATE_XY_COORDS

Types/Data
 Null_XY

PSUEDO CODE

 Null_XY := Evaluate_XY_Coords ( X, Y ) ;

CODE
*/
{
	evaluate_x ( x ) ;
	evaluate_y ( y ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION SET_POINT
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Set_Point ( X, Y, Colour ) is
*/
void graphics_support::set_point ( unsigned x, unsigned y, unsigned char colour )
/*
DESCRIPTION

This operation sets the requested X & Y co-ordinates to the requested colour.
The Evaluate operations are utilised to determine the absolute co-ordinates &
colour for the current mode.

USES OPERATIONS
 EVALUATE_XY_COORDS
 EVALUATE_COLOUR

Types/Data
 New_XY
 New_Colour

PSUEDO CODE

 New_XY := Evaluate_XY_Coords ( X, Y ) ;
 New_Colour := Evaluate_Colour ( Colour ) ;

 Perform_Graphics_Set_Point ( New_X, New_Y, New_Colour ) ;

CODE
*/
{
  SetPixel ( GhDC, evaluate_x ( x ), evaluate_y ( y ), evaluate_colour ( colour ) ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION DRAW_LINE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Draw_Line ( X1, Y1, X2, Y2, Colour ) is
*/
void graphics_support::draw_line ( unsigned x1,
                                   unsigned y1,
                                   unsigned x2,
                                   unsigned y2,
                                   unsigned char colour,
                                   unsigned char width,
                                   unsigned char style )
/*
DESCRIPTION

This operation draws a line from the co-ordinates specified in the requested
colour. The Evaluate operations are called to determine the correct values for
the current mode.

USES OPERATIONS
 EVALUATE_XY_COORDS
 EVALUATE_COLOUR

Types/Data
 New_XY1
 New_XY2
 New_Colour

PSUEDO CODE

 New_XY1 := Evaluate_XY_Coords ( X1, Y1 ) ;
 New_XY2 := Evaluate_XY_Coords ( X2, Y2 ) ;
 New_Colour := Evaluate_Colour ( Colour ) ;

 Perform_Draw_Line_Operation ( New_X1, New_Y1, New_X2, New_Y2, New_Colour ) ;

CODE
*/
{
	 HPEN old_pen, new_pen ;

	 new_pen = CreateReqdPen ( width, style, colour ) ;
	 old_pen = SelectObject ( GhDC, new_pen ) ;
	 MoveToEx ( GhDC, evaluate_x ( x1 ), evaluate_y ( y1 ), NULL ) ;
	 LineTo ( GhDC, evaluate_x ( x2 ), evaluate_y ( y2 ) ) ;
	 SelectObject ( GhDC, old_pen ) ;
	 DeleteObject ( new_pen ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION DRAW_CIRCLE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Draw_Circle ( X, Y, Radius, Colour ) is
*/
void graphics_support::draw_circle ( unsigned x,
                                     unsigned y,
                                     unsigned radius,
                                     unsigned char colour,
                                     unsigned char width,
                                     unsigned char style )
/*
DESCRIPTION

This operation draws a circle at the specified centre co-ordinates, radius and
colour. The Evaluate operations are called to determine the correct values for
this mode.

USES OPERATIONS
 EVALUATE_XY_COORDS
 EVALUATE_COLOUR

Types/Data
 New_XY
 New_Colour

PSUEDO CODE

 New_XY := Evaluate_XY_Coords ( X, Y ) ;
 New_Colour := Evaluate_Colour ( Colour ) ;

 Perform_Draw_Circle ( New_X, New_Y, Radius, New_Colour ) ;

CODE
*/
{
	HPEN old_pen, new_pen ;
	unsigned new_x, new_y, sx, sy, ex, ey ;

	new_x = evaluate_x ( x ) ;
	new_y = evaluate_y ( y ) ;
	sx = new_x-radius*fabs(x_scaling_factor) ;
	sy = new_y-radius*fabs(y_scaling_factor) ;
	ex = new_x+radius*fabs(x_scaling_factor) ;
	ey = new_y+radius*fabs(y_scaling_factor) ;

	new_pen = CreateReqdPen ( width, style, colour ) ;

	old_pen = SelectObject ( GhDC, new_pen ) ;
	Arc ( GhDC, sx, sy, ex, ey, sx, sy, sx, sy ) ;
	SelectObject ( GhDC, old_pen ) ;
	DeleteObject ( new_pen ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION READ_PIXEL
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Function Read_Pixel ( X, Y ) return colour is
*/
unsigned char graphics_support::read_pixel ( unsigned x, unsigned y )
/*
DESCRIPTION

This operation determines the colour of the specified co-ordinates. The
Evaluate operation is used to obtain the correct co-ordinates for the current
mode.

USES OPERATIONS
 EVALUATE_XY_COORDS

Types/Data
 New_XY

PSUEDO CODE

 New_XY := Evaluate_XY_Coords ( X, Y ) ;

 return Perform_Read_Pixel ( New_X, New_Y ) ;

CODE
*/
{
	 return GetNearestPaletteIndex ( palette,
		GetPixel ( GhDC, evaluate_x ( x ), evaluate_y ( y ) ) ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION COLOUR_CHANGE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Colour_Change ( Colour ) is
*/
void graphics_support::colour_change ( unsigned char colour )
/*
DESCRIPTION

This operation changes the current colour setting. The Evaluate operation is
called to determine if a valid colour has been requested prior to performing
the change.

USES OPERATIONS
 EVALUATE_COLOUR

Types/Data
 None

PSUEDO CODE

 Current_Colour := Evaluate_Colour ( Colour ) ;

CODE
*/
{
	 evaluate_colour ( colour ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION CLEAR_SCREEN
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 24, 1995, FOR V2.01 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON June 21, 1998 FOR V4.0 BY J.Bird

Procedure Clear_Screen ( Colour ) is
*/
void graphics_support::clear_window ( unsigned char colour )
/*
DESCRIPTION

This operation clears the graphics screen to the requested colour. The
Evaluate operation is utilised to determine the correct colour for the current
mode prior to performing the clear.

USES OPERATIONS
 EVALUATE_COLOUR

Types/Data
 New_Colour

PSUEDO CODE

 New_Colour := Evaluate_Colour ( Colour ) ;

 Perform_Clear_Screen ( Colour ) ;

CODE
*/
{
	HBRUSH fill_brush ;
	RECT rc ;
	COLORREF bkcolour ;
	PALETTEENTRY PalEntry ;

	 if ( colour == DEFAULT_COL )
		bkcolour = current_col ;
	 else
	 {
		GetPaletteEntries ( palette, colour, 1, &PalEntry ) ;
		bkcolour = RGB ( PalEntry.peRed, PalEntry.peGreen, PalEntry.peBlue ) ;
	 }

	 fill_brush = CreateSolidBrush ( bkcolour ) ;
	 rc.top = 0 ;
	 rc.left = 0 ;
	 rc.right = max_x ;
	 rc.bottom = max_y ;
	 FillRect ( GhDC, &rc, fill_brush ) ;
	 DeleteObject ( fill_brush ) ;
}
/*
END.
*/

// copy the graphics memory into the supplied display context
// v3.11 bug fix - palette not restored
void graphics_support::txfr_to_display_window ( HDC hDC, unsigned x_pixels, unsigned y_pixels )
{
	HPALETTE old_palette ;

	old_palette = SelectPalette ( hDC, palette, 0 ) ;
	RealizePalette ( hDC ) ;
	StretchBlt ( hDC, 0, 0, x_pixels, y_pixels, GhDC, 0, 0, max_x, max_y, SRCCOPY ) ;
	SelectPalette ( hDC, old_palette, 0 ) ;
}

// new in v4.0 - define palette entry
void graphics_support::define_palette_entry ( int pal_entry, unsigned char red,
					unsigned char green, unsigned char blue )
{
	// setup structure
	pal_data->palPalEntry[pal_entry].peRed = red ;
	pal_data->palPalEntry[pal_entry].peGreen = green ;
	pal_data->palPalEntry[pal_entry].peBlue = blue ;
	pal_data->palPalEntry[pal_entry].peFlags = NULL ;

  // delete the old palette & create new one into dc
  SelectPalette ( GhDC, old_palette, 0 ) ;
  DeleteObject ( palette ) ;
	palette = CreatePalette ( pal_data ) ;
	old_palette = SelectPalette ( GhDC, palette, 0 ) ;
}


