/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PACKAGE HDISK
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 23, 1995, FOR V2.01 BY J.Bird
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON June 21 1998, FOR V3.11 BY J.Bird

Package Message_Processor.HDisk is

DESCRIPTION

This object provides the functionality for providing the shared hard disk for
the external computer. This is implemented at a sector level, with the caller
passing sectors to be read/written to the hard disk.

The shared hard disk is implemented as a large indexed DOS file, allowing the
Logical Sector Number (LSN) supplied to be directly converted to a file
position within this file.

REQUIRED INTERFACE
 None
*/
#include <windows.h>
#include <stdio.h>
#include <string.h>
#include "globals.h"
#include "config.h"
/*

Types/Data

-- file handle of the DOS file
 HDisk_File_Handle
*/
FILE *hdisk_handle[10] ;
/*
PROVIDES OPERATIONS
 READ_SECTOR
 WRITE_SECTOR
*/

/*
INTERNAL OPERATIONS
 None

END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION READ_SECTOR
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 23, 1995, FOR V2.01 BY J.Bird
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON June 21 1998, FOR V3.11 BY J.Bird

Function Read_Sector ( LSN ) return Sector_Buffer is
*/
char read_sector ( unsigned char drive, unsigned long LSN, char *sector_buffer )
/*
DESCRIPTION

This operation converts the LSN provided into a DOS file position, seeks to
the position in the HDisk file and transfers the next 256 bytes of data into
the sector buffer.

USES OPERATIONS
 None

Types/Data
 File_Posn : long_int ;
 Sector_Index : byte ;
*/
{
unsigned long file_posn ;
/*
PSUEDO CODE

 File_Posn := LSN * Sector_Size ;

 DOS_Seek ( HDisk_File_Handle, File_Posn ) ;

 Sector_Index := 0 ;

 Loop_for_sector_size

		 Sector_Buffer ( Sector_Index ) := DOS_Read_Byte ( HDisk_File_Handle ) ;

		 Sector_Index := Sector_Index + 1 ;

 End loop ;

CODE
*/
	if ( drive >= system_config.hard_drives )
		return FALSE ;

	// v3.11 fix - catch file not actually open
	if ( hdisk_handle[drive] == NULL )
		return FALSE ;

	file_posn = LSN * system_config.sector_size ;

	if ( fseek ( hdisk_handle[drive], file_posn, SEEK_SET ) != 0 )
		return FALSE ;

	if ( fread ( sector_buffer, 1, system_config.sector_size, hdisk_handle[drive] ) != system_config.sector_size )
		if ( feof ( hdisk_handle[drive] ) )
			return TRUE ;
		else
			return FALSE ;
	else
		return TRUE ;

}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION WRITE_SECTOR
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 23, 1995, FOR V2.01 BY J.Bird
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON June 21 1998, FOR V3.11 BY J.Bird

Procedure Write_Sector ( Sector_Buffer, LSN ) is
*/
char write_sector ( unsigned char drive, unsigned long LSN, char *sector_buffer )
/*
DESCRIPTION

This operation will convert the provided LSN into a DOS file position, seek to
the position in the HDisk file and transfer the sector buffer into the file.

USES OPERATIONS
 None

Types/Data
 File_Posn : long_int ;
 Sector_Index : byte ;
*/
{
unsigned long file_posn ;
/*
PSUEDO CODE

 File_Posn := LSN * Sector_Size ;

 Sector_Index := 0 ;

 Loop_for_sector_size

		 DOS_Write_Byte ( HDisk_Handle, Sector_Buffer ( Sector_Index ) ) ;

		 Sector_Index := Sector_Index + 1 ;

 End loop ;

CODE
*/
	if ( drive >= system_config.hard_drives )
		return FALSE ;

	// v3.11 fix - catch file not actually open
	if ( hdisk_handle[drive] == NULL )
		return FALSE ;

	file_posn = LSN * system_config.sector_size ;

	if ( fseek ( hdisk_handle[drive], file_posn, SEEK_SET ) != 0 )
		return FALSE ;

	if ( fwrite ( sector_buffer, 1, system_config.sector_size, hdisk_handle[drive] ) != system_config.sector_size )
		return FALSE ;
	else
		return TRUE ;

}
/*
END.
*/


void open_hdisk ( void )
{
	char drvpath[200] ;
	int drv_no ;

	for ( drv_no=0 ; drv_no<system_config.hard_drives ; drv_no++ )
	{
		sprintf ( drvpath, "%s%s%d%s", system_config.user_dir, "HDISK", drv_no, ".BIN" ) ;
		if (( hdisk_handle[drv_no] = fopen ( drvpath, "r+b" )) == NULL )
			hdisk_handle[drv_no] = fopen ( drvpath, "w+b" ) ;
	}
}


void close_hdisk ( void )
{
	int drv_no ;

	for ( drv_no=0 ; drv_no < system_config.hard_drives ; drv_no++ )
	{
		// v3.11 fix - catch file not actually open
		if ( hdisk_handle[drv_no] != NULL )
			fclose ( hdisk_handle[drv_no] ) ;
	}
}

























