/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PACKAGE IO_PORTS
CREATED ON September 23 1995, FOR V2.01 BY J.Bird
MOD ON February 18, 1997, FOR V3.0 BY J.BIRD

Package Message_Processor.IO_Ports is

DESCRIPTION

This package is responsible for the hardware handling of the input ports to
the PCRSS. Any data correction/synchronization will be handled by the routines
here.

REQUIRED INTERFACE
 None
*/
#include <windows.h>
#include <conio.h>
#include "globals.h"
#include "config.h"
#include "io_ports.h"
/*
Types/Data

-- valid input ports
 type PORT_ID_TYPE is ( VIDEO_PORT,
			 HDISK_PORT ) ;

-- other data is hardware specific
*/
#define PORT_A 0
#define PORT_B 1
#define PORT_C 2
#define CTRL_REG 3

#define INPUT 0
#define OUTPUT 1

int current_dir ;
/*
PROVIDES OPERATIONS
 PORT_STATUS
 READ_DATA
*/

/*
INTERNAL OPERATIONS
 None

END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION PORT_STATUS
CREATED ON September 23 1995, FOR V2.01 BY J.Bird

Function In_Port_Status ( Port_Id ) return boolean is
*/
int in_port_status ( void )
/*
DESCRIPTION

This operation determines if there is data ready at the specified input port,
returning a boolean value to indicate data presence.

USES OPERATIONS
 None

Types/Data
 None

PSUEDO CODE

 if Port_Id = Video_Port
 then

	  Read_video_port_status_information ;

	  if hardware_indicates_data_available
	  then

			return true ;

	  else

			return false ;

 else

	  Read_hdisk_port_status_information ;

	  if hardware_indicates_data_available
	  then

			return true ;

	  else

			return false ;

	  end if ;

 end if ;

CODE
*/
{
	if ( current_dir != INPUT )
		init_in_port() ;

	if ( ( inp ( ( system_config.base_addr + PORT_C ) ) & 2 ) == 2 )
		return TRUE ;
	else
		return FALSE ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION READ_DATA
CREATED ON September 23 1995, FOR V2.01 BY J.Bird

Function READ_DATA ( Port_Id ) return byte is
*/
unsigned char read_data ( void )
/*
DESCRIPTION

This operation waits until data is available on the requested port, and then
reads the data from the port.

Note: if the caller is not certain if data is available at the port, it should
first call the Port_Status operation to determine if Read_Data should be
called.

USES OPERATIONS
 PORT_STATUS

Types/Data
 None
*/
/*
PSUEDO CODE

 Loop while not Port_Status ( Port_Id )

	  null

 End loop ;

 if Port_Id = Video_Port
 then

		return read_video_port_hardware_for_byte ;

 else

		return read_hdisk_port_hardware_for_byte ;

 end if ;

CODE
*/
{
//		while ( in_port_status ( port_id ) == FALSE )
//		{
//		}
	if ( current_dir != INPUT )
		init_in_port() ;

	while ( ! in_port_status() )
	{
	}

	return inp ( system_config.base_addr + PORT_B ) ;
}
/*
END.
*/

void init_in_port ( void )
{
int mode_1b_input = 134 ;

	mode_1b_input = outp ( ( system_config.base_addr + CTRL_REG ), mode_1b_input ) ;
	current_dir = INPUT ;
}

void init_out_port ( void )
{
int mode_1b_output = 132 ;

	mode_1b_output = outp ( system_config.base_addr + CTRL_REG, mode_1b_output ) ;
	current_dir = OUTPUT ;
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION WRITE_DATA
CREATED ON September 23 1995, FOR V2.01 BY J.Bird

Procedure WRITE_DATA ( Port_Id, Data_Byte ) is
*/
void write_data ( unsigned char data_byte )
/*
DESCRIPTION

This operation waits until the output port is ready to receive data, and then
transmits the passed byte.

Note: if the caller is not certain if the output port is ready, it should
first call the Port_Status operation to determine if Write_Data should be
called.

USES OPERATIONS
 PORT_STATUS

Types/Data
 None

PSUEDO CODE

 Loop while not Port_Status ( Port_Id )

	  null

 End loop ;

 if Port_Id = Video_Port
 then

		Store_data_byte_in_video_output_port ;

 else

		Store_data_byte_in_hdisk_output_port ;

 end if ;

CODE
*/
{
	if ( current_dir != OUTPUT )
		init_out_port() ;

	outp ( system_config.base_addr + PORT_B, data_byte ) ;

	while ( !out_port_status())
	{
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION PORT_STATUS
CREATED ON September 23 1995, FOR V2.01 BY J.Bird

Function Out_Port_Status ( Port_Id ) return boolean is
*/
int out_port_status ( void )
/*
DESCRIPTION

This function determines if the output port is ready to receive more data,
returning a boolean value of true if it is.

USES OPERATIONS
 None

Types/Data
 None

PSUEDO CODE

 if Port_Id = Video_Port
 then

	  Read_video_port_status_information ;

	  if hardware_indicates_port_ready
		 then

			return true ;

	  else

			return false ;

 else

	  Read_hdisk_port_status_information ;

	  if hardware_indicates_port_ready
	  then

			return true ;

	  else

			return false ;

	  end if ;

 end if ;

CODE
*/
{
	if ( current_dir != OUTPUT )
		init_out_port() ;

	if ( ( inp ( system_config.base_addr + PORT_C ) & 2 ) == 2 )
		return TRUE ;
	else
		return FALSE ;
}
/*
END.
*/
