/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PACKAGE MESSAGE_PROCESSOR
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON October 8, 1995, FOR V2.03 BY J.BIRD
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON October 09, 1996, FOR V3.0 BY J.Bird
MOD ON May 01, 1997, FOR V3.01 BY J.Bird
MOD ON October 29 1997, FOR V3.02 BY J.Bird
MOD ON November 21 1997, FOR V3.1 BY J.Bird
MOD ON June 21 1998, FOR V4.0 BY J.Bird
MOD ON April 30 2000, FOR V4.1 BY J.Bird

Package MESSAGE_PROCESSOR is

DESCRIPTION

The Message Processor object is responsible for handling and interpreting the
communcation between the external computer. This will involve the following
processes:

-- reading data from the input ports
-- determining the message that is being requested
-- requesting an internal or external object to perform the operation
-- format & return messages through the output ports

The object also comprises 5 internal packages:

 -- IN_PORT for accessing the input ports
 -- OUT_PORT for accessing the output ports
 -- HDISK for performing the simulated hard disk transfer
 -- FILE_TRANSFER for transferring DOS files across the interface
 -- MISC_UTILS for performing any other miscellaneous operations

REQUIRED INTERFACE
 VIDEO_CTRL
 IN_PORT
 OUT_PORT
 HDISK
 FILE_TRANSFER
 MISC_UTILS
*/
#include <windows.h>
#include <string.h>
#include <mem.h>
#include <malloc.h>
#include <errno.h>
#include <dir.h>
#include <stdio.h>
#include "globals.h"
#include "config.h"
#include "hdisk.h"
#include "io_ports.h"
#include "message.h"
#include "file_han.h"
#include "video_ct.h"
#include "misc_uti.h"
#include <except.h>
#include "pcshare.h"
/*

Types/Data

 -- defines the state of an incoming message
 type MESSG_STATE_TYPE is ( NORMAL,
									IDENT_1,
									IDENT_2 ) ;
*/
#define NORMAL 0
#define IDENT_1 1
#define IDENT_2 2

#define WRITE_STATUS(a) write_data ( (void*)&a, 1 ) 
/*

-- Message Data in accordance with Data Interface Description Document, Ref:
106-4-12, Issue 1, February 1995.

-- defines the valid messages
 type MESSG_TYPE is ( GRAPHICS_MODE_CHANGE,
							 GRAPHICS_MOVE_CURSOR,
							 GRAPHICS_COLOUR_CHANGE,
							 GRAPHICS_SET_POINT,
							 GRAPHICS_CLEAR_SCREEN,
							 GRAPHICS_DRAW_LINE,
							 GRAPHICS_DRAW_CIRCLE,
							 GRAPHICS_READ_PIXEL_COLOUR,
							 FILE_WRITE,
							 FILE_READ,
							 PRINT_MESSAGE,
							 DOS_COMMAND ) ;

*/

#define IDENT_CHAR_1 0xAA
#define IDENT_CHAR_2 0xFF
#define IDENT_CHAR_3 0xFF

#define GRAPHICS_MODE_CHANGE 0x0
#define GRAPHICS_MOVE_CURSOR 0x1
#define GRAPHICS_COLOUR_CHANGE 0x2
#define GRAPHICS_SET_POINT 0x3
#define GRAPHICS_CLEAR_SCREEN 0x4
#define GRAPHICS_DRAW_LINE 0X5
#define GRAPHICS_DRAW_CIRCLE 0x6
#define GRAPHICS_READ_PIXEL_COLOUR 0x7
#define FILE_CREATE 0x8
#define FILE_OPEN 0x9
#define CHANGE_DIR 0xA
#define FILE_CLOSE 0xB
#define FILE_DELETE 0xC
#define MAKE_DIR 0xD
#define FILE_READ_BIN 0xE
#define FILE_READ_TXT 0xF
#define FILE_SEEK 0x10
#define FILE_WRITE_BIN 0x11
#define FILE_WRITE_TXT 0x12
#define DOS_COMMAND 0x15
#define EMULATE_SELECT 0x16
#define READ_SECTOR 0x17
#define WRITE_SECTOR 0x18
// new in V2.1
#define FILE_GET_FILE_SIZE 0x13
#define FILE_GET_FILE_POS 0x14
#define FILE_END_OF_FILE 0x19
#define FILE_SET_FILE_SIZE 0x1A
#define DELETE_DIR 0x1B
#define READ_DIR_ENTRY 0x1C
#define SET_TEXT_ATTR 0x1D
#define SET_TEXT_MODE 0x1E
// new in V2.2
#define FILE_DEV_READY 0x1F
#define SET_COM_CONFIG 0x20
#define READ_RTC 0x21
// new in V3.0
#define OPEN_WINDOW 0x30
#define CLOSE_WINDOW 0x31
#define SET_ACTIVE_WINDOW 0x32
#define GET_ACTIVE_WINDOW 0x33
#define ACTIVE_WINDOW 0x34
// new in V4.0
#define GRAPHICS_DEFINE_PAL_ENTRY 0x35

// file access modes
#define READ 1
#define WRITE 2
#define UPDATE 3
// new in V3.1
#define READ_DIR 0x81

// error codes
const unsigned char OKAY = 0 ;
const unsigned char FILE_NOT_ACCESSABLE = 1 ;
const unsigned char BUFFER_FAILURE = 2 ;
const unsigned char READ_FAILURE = 3 ;
const unsigned char WRITE_FAILURE = 4 ;
const unsigned char END_OF_FILE = 5 ;
const unsigned char GENERAL_IO_FAILURE = 6 ;
// new in V2.2
const unsigned char NOT_READY = 7 ;

/*

 MESSG_STATE : MESSG_STATE_TYPE ;
*/
char messg_state ;
int selected_window = 0 ;
/*
Provides Operations
 MESSAGE_EVENT
 PROCESS_MESSAGE
 OUTPUT_MESSAGE
*/
/*

Internal Operations
 PROCESS_SUPERVISOR_MESSAGE
 PROCESS_HDISK_MESSAGE
 PROCESS_FTP_MESSAGE
 PROCESS_VIDEO_CTRL_MESSAGE
 PROCESS_MISC_MESSAGE
 BUFFER_INPUT_DATA
 TRANSFER_OUTPUT_DATA

END.
*/
// fetch unsigned long from memory
unsigned long fetch_long ( char *src_data )
{
unsigned char byte_array[4] ;
unsigned long return_long ;

	byte_array[0] = src_data[3] ;
	byte_array[1] = src_data[2] ;
	byte_array[2] = src_data[1] ;
	byte_array[3] = src_data[0] ;

	memcpy ( &return_long, &byte_array, 4 ) ;
	return return_long ;
}

// fetch unsigned short from memory
unsigned short fetch_word ( char *src_data )
{
unsigned char byte_array[2] ;
unsigned short return_word ;

	byte_array[0] = src_data[1] ;
	byte_array[1] = src_data[0] ;

	memcpy ( &return_word, &byte_array, 2 ) ;
	return return_word ;
}

// write an unsigned word to the i/o port
void write_word ( unsigned short target_word )
{
  unsigned char byte_array[2] ;

  byte_array[0] = (target_word >> 8) & 0xff ;
  byte_array[1] = target_word & 0xff ; ;
	write_data ( byte_array,sizeof(byte_array) ) ;
}

// write an unsigned long to the i/o port
void write_long ( unsigned long target_long )
{
  unsigned char byte_array[4] ;

  byte_array[0] = (target_long >> 24) & 0xff ;
  byte_array[1] = (target_long >> 16) & 0xff ;
  byte_array[2] = (target_long >> 8) & 0xff ;
  byte_array[3] = target_long & 0xff ;
  write_data ( byte_array, sizeof ( byte_array ) ) ;
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION PROCESS_HDISK_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON April 26 1996, FOR V2.1 BY J.Bird

Procedure Process_HDisk_Message is
*/
void process_hdisk_message( unsigned char command, char *hdisk_packet )
/*
DESCRIPTION

This operation determines the requested HDisk message, initiates the transfer
of data across the HDisk interface and utilises the HDisk object to store the
data on the shared hard disk.

USES OPERATIONS
 BUFFER_INPUT_DATA
 TRANSFER_OUTPUT_BUFFER
 HDISK.READ_SECTOR
 HDISK.WRITE_SECTOR
 IN_PORT.READ_DATA

Types/Data
 Sector_Buffer : 256 byte buffer
 HDisk_Command
 LSN : long_int ;
*/
{
unsigned char sector_buffer[512] ;
unsigned long LSN ;
unsigned char status ;
/*
PSUEDO CODE

 HDisk_Command := In_Port.Read_Data ( HDisk_Port ) ;

 Read_LSN_from_Hdisk_Port ;

 if HDisk_Command = Write_Sector
 then

	 Sector_Buffer := Buffer_Input_Data ( HDisk_Port, 256 ) ;

	 HDisk.Write_Sector ( LSN, Sector_Buffer ) ;

 else

	 Sector_Buffer := HDisk.Read_Sector ( LSN ) ;

	 Transfer_Output_Buffer ( HDisk_Port, Sector_Buffer, 256 ) ;

 end if ;

CODE
*/
		LSN = fetch_long ( &hdisk_packet[1] ) ;

		if ( command == WRITE_SECTOR )
		{
      read_data ( sector_buffer, system_config.sector_size ) ;
			status = write_sector ( hdisk_packet[0], LSN, sector_buffer ) ;
			output_message() ;
			if ( status )
        WRITE_STATUS ( OKAY ) ;
      else
        WRITE_STATUS ( WRITE_FAILURE ) ;
		}
		else
		{
			status = read_sector ( hdisk_packet[0], LSN, sector_buffer ) ;
			output_message() ;
			if ( status )
			{
        WRITE_STATUS ( OKAY ) ;
        write_data ( sector_buffer, system_config.sector_size ) ;
			}
			else
        WRITE_STATUS ( READ_FAILURE ) ;
		}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION PROCESS_FTP_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON October 8, 1995, FOR V2.03 BY J.BIRD
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON October 29 1997, FOR V3.02 BY J.Bird
MOD ON November 21 1997, FOR V3.1 BY J.Bird

Procedure Process_Ftp_Message ( Command, FTP_Packet ) is
*/
void process_file_handler_message ( unsigned char command, char *file_packet )
/*
DESCRIPTION

This procedure is responsible for determining the required transfer, and then
reading or writing blocks of data to/from the Video_Port and passing them to
the File_Transfer object for reading or writing to disk.

USES OPERATIONS
 OUTPUT_MESSAGE
 BUFFER_INPUT_DATA
 BUFFER_OUTPUT_DATA
 FILE_TRANSFER.OPEN_FILE
 FILE_TRANSFER.CLOSE_FILE
 FILE_TRANSFER.WRITE_BLOCK
 FILE_TRANSFER.READ_BLOCK

Types/Data
 Filename : string ;
 File_Length : long_int ;
 File_Buffer : 256 byte buffer ;
 Buffer_Length : byte ;
*/
{
short int cr_posn ;
char fcb, eof_flag, byte ;
char success = FALSE ;
char *buffer, *memptr ;
unsigned short block_size, count ;
long file_size, file_posn ;
char buffer_allocated = FALSE ;
unsigned char acc_mode, bsize ;
/*
PSUEDO CODE


CODE
*/

	switch ( command )
	{
		case FILE_CREATE :
		{
			// replace the CR with a null for C strings
			cr_posn = strcspn ( &file_packet[1], "\x0D" ) ;
			file_packet[cr_posn+1] = 0 ;
			acc_mode = (unsigned char)file_packet[0] ;

			// call file handler appropriatly
			switch ( acc_mode )
			{
				case WRITE :
				{
					success = open_file ( &file_packet[1], "wb", &fcb ) ;
					break ;
				}

				case UPDATE :
				{
					success = open_file ( &file_packet[1], "w+b", &fcb ) ;
					break ;
				}

				default:

					success = FALSE ;
					break ;

			}

			output_message() ;

			// process errors
			if (!success)
				WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
			else
      {
				WRITE_STATUS ( OKAY ) ;
				write_data ( &fcb, 1 ) ;
			}
			break ;
		}

		case FILE_OPEN :
		{
			cr_posn = strcspn ( &file_packet[1], "\x0D" ) ;
			file_packet[cr_posn+1] = 0 ;

			acc_mode = (unsigned char)file_packet[0] ;
			// new in V3.1
			switch ( acc_mode )
			{
				case READ :
				{
					success = open_file ( &file_packet[1], "rb", &fcb ) ;
					break ;
				}

				case WRITE :
				{
					// not limited to write
					success = open_file ( &file_packet[1], "r+b", &fcb ) ;
					break ;
				}

				case UPDATE :
				{
					success = open_file ( &file_packet[1], "r+b", &fcb ) ;
					break ;
				}

				case READ_DIR :

					success = open_file ( &file_packet[1], "rdir", &fcb ) ;
					break ;

				default :

					success = FALSE ;
					break ;
			}

			output_message() ;
			if (!success)
				WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
			else
      {
				WRITE_STATUS ( OKAY ) ;
				write_data ( &fcb, 1 ) ;
			}
			break ;
		}

		case CHANGE_DIR :
		{
			cr_posn = strcspn ( file_packet, "\x0D" ) ;
			file_packet[cr_posn] = 0 ;

			output_message() ;

			// pass dir name to file_handler
			if ( !change_dir ( file_packet ) )
				WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
			else
			{
				// return current working directory
				if ( !get_current_dir ( file_packet ) )
					WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
				else
				{
					WRITE_STATUS ( OKAY ) ;
					strcat ( file_packet, "\x0d" ) ;
					bsize = strlen ( file_packet ) ;
					write_data ( &bsize, 1 ) ;
          write_data ( file_packet, bsize ) ;
				}
			}
			break ;
		}

		case FILE_CLOSE :
		{
			output_message() ;

			// close fcbn specified in packet
			if ( !close_file ( file_packet[0] ) )
				WRITE_STATUS ( GENERAL_IO_FAILURE ) ;
			else
				WRITE_STATUS ( OKAY ) ;

			break ;
		}

		case FILE_DELETE :
		{
			cr_posn = strcspn ( file_packet, "\x0D" ) ;
			file_packet[cr_posn] = 0 ;

			output_message() ;

			if ( !delete_file ( file_packet ) )
				WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
			else
				WRITE_STATUS ( OKAY ) ;

			break ;
		}

		case MAKE_DIR :
		{
			cr_posn = strcspn ( file_packet, "\x0D" ) ;
			file_packet[cr_posn] = 0 ;

			output_message() ;

			if ( !make_dir ( file_packet ) )
				WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
			else
				WRITE_STATUS ( OKAY ) ;

			break ;
		}

		case FILE_READ_BIN :
		{
			fcb = file_packet[0] ;
			success = OKAY ;

			if ( !end_of_file (fcb ) )
			{
				// get buffer for transfer
				block_size = fetch_word ( &file_packet[1] ) ;
				if (( buffer = ( char* ) malloc ( block_size )) == NULL )
					success = BUFFER_FAILURE ;
				else
				{
					buffer_allocated = TRUE ;
					// do the read
					if ( !read_file(fcb, buffer, &block_size, BINARY ))
					{
						// if eof encountered during read, then this is okay
						// eof will be reported on next request,
						// except when no data is available for transfer
						// must report it now
						if ( !end_of_file ( fcb ) )
							success = READ_FAILURE ;
						else
							if ( block_size == 0 )
								success = END_OF_FILE ;
					}
				}
			}
			else
				success = END_OF_FILE ;

			output_message() ;

			// o/p status byte
			WRITE_STATUS ( success ) ;

			// o/p the block size
			if ( success == OKAY )
			{
				write_word ( block_size ) ;
				// output the data block
        write_data ( buffer, block_size ) ;
			}

			if ( buffer_allocated )
				free ( buffer ) ;

			break ;
		}

		case FILE_READ_TXT :
		{
			fcb = file_packet[0] ;
			success = OKAY ;

			if ( !end_of_file ( fcb ) )
			{
				// get buffer for transfer, plus terminator byte
				block_size = fetch_word ( &file_packet[1] ) ;
				if (( buffer = ( char* ) malloc ( block_size + 1 )) == NULL )
					success = BUFFER_FAILURE ;
				else
				{
					buffer_allocated = TRUE ;
					// do the read
					if ( !read_file(fcb, buffer, &block_size, TEXT ) )
					{
						// if eof encountered during read, then report it now
						// as text reads will read right up till EOF
						// otherwise report error
						if ( end_of_file ( fcb ) )
							success = END_OF_FILE ;
						else
							success = READ_FAILURE ;
					}
				}
			}
			else
				success = END_OF_FILE ;

			output_message() ;

			// o/p status byte
			WRITE_STATUS ( success ) ;

			// o/p the block size
			if ( success == OKAY )
			{
				write_word ( block_size ) ;
				// output the data block
        write_data ( buffer, block_size ) ;
			}

			if ( buffer_allocated )
				free ( buffer ) ;

			break ;
		}

		case FILE_SEEK :
		{

			output_message() ;

			if ( !seek_file ( file_packet[0], fetch_long ( &file_packet[1] ) ) )
				WRITE_STATUS ( GENERAL_IO_FAILURE ) ;
			else
				WRITE_STATUS ( OKAY ) ;

			break ;
		}

		case FILE_WRITE_BIN :
		{

			fcb = file_packet[0] ;
			block_size = fetch_word ( &file_packet[1] ) ;
			success = OKAY ;

			// get a buffer for the incoming data
			if (( buffer = ( char* ) malloc ( block_size )) == NULL )
				success = BUFFER_FAILURE ;
			else
			{
        read_data ( buffer, block_size ) ;

				// write to file & check for errors
				// discard partial writes
				if ( !write_file ( fcb, buffer, &block_size, BINARY ) )
					success = WRITE_FAILURE ;

				free ( buffer ) ;
			}

			output_message() ;

			WRITE_STATUS ( success ) ;
			if ( success == OKAY )
				write_word ( block_size ) ;

			break ;
		}

		case FILE_WRITE_TXT :
		{

			fcb = file_packet[0] ;
			block_size = fetch_word ( &file_packet[1] ) ;
			success = OKAY ;

			// get a buffer for the incoming data, plus space for terminators
			if (( buffer = ( char* ) malloc ( block_size + 2 )) == NULL )
				success = BUFFER_FAILURE ;
			else
			{
				read_data ( buffer, block_size ) ;

				// terminate the string correctly
				buffer[block_size+1] = 0 ;

				// look for first CR in the string
				memptr = strchr ( buffer, '\x0D' ) ;

				// NOTE: THIS SHOULD NOT BE DONE HERE - SHOULD BE
				// A PCSBF FUNCTION TO ADD LF TO MEET DOS PROTOCOL
				// if CR is found, add LF & update block_size
				// Note: may cause problems as block_size could
				// feasably exceed requested block size
				if ( memptr != NULL )
				{
					memptr[1]= 0x0A ;
					// terminate string here
					memptr[2]=0 ;
				}

				// write to file & check for errors
				// discard partial writes
				if ( !write_file ( fcb, buffer, &block_size, TEXT ) )
					success = WRITE_FAILURE ;

				free ( buffer ) ;
			}

			output_message() ;

			WRITE_STATUS ( success ) ;
			if ( success == OKAY )
				write_word ( block_size ) ;

			break ;
		}

		case FILE_GET_FILE_SIZE :
		{
			fcb = file_packet[0] ;
			output_message() ;

			if ( get_file_size ( fcb, &file_size ) == TRUE )
			{
				WRITE_STATUS ( OKAY ) ;
				write_long ( file_size ) ;
			}
			else
				WRITE_STATUS ( GENERAL_IO_FAILURE ) ;

			break ;
		}

		case FILE_GET_FILE_POS :
		{
			fcb = file_packet[0] ;
			output_message() ;

			if ( get_file_posn ( fcb, &file_posn ) == TRUE )
			{
				WRITE_STATUS ( OKAY ) ;
				write_long ( file_posn ) ;
			}
			else
				WRITE_STATUS ( GENERAL_IO_FAILURE ) ;
			break ;
		}

		case FILE_END_OF_FILE :
		{
			fcb = file_packet[0] ;
			output_message() ;

			if ( end_of_file ( fcb ) == TRUE )
				WRITE_STATUS ( END_OF_FILE ) ;
			else
				WRITE_STATUS ( OKAY ) ;
			break ;
		}

		case FILE_SET_FILE_SIZE :
		{
			fcb = file_packet[0] ;
			output_message() ;

			file_size = fetch_long ( &file_packet[1] ) ;
			if ( set_file_size ( fcb, file_size ) == TRUE )
				WRITE_STATUS ( OKAY ) ;
			else
				WRITE_STATUS ( GENERAL_IO_FAILURE ) ;
			break ;
		}

		case DELETE_DIR :
		{
			cr_posn = strcspn ( file_packet, "\x0D" ) ;
			file_packet[cr_posn] = 0 ;

			output_message() ;

			if ( !delete_dir ( file_packet ) )
				WRITE_STATUS ( FILE_NOT_ACCESSABLE ) ;
			else
				WRITE_STATUS ( OKAY ) ;

			break ;
		}

		case READ_DIR_ENTRY :
		{
			char name[256] ;
			int i;
			unsigned short entry_no ;

			output_message() ;

			// get entry & process error state
			entry_no = fetch_word ( &file_packet[1] ) ;
			fcb = file_packet[0] ;

			if ( read_dir_entry ( fcb, entry_no, name ) )
			{
				// entry is okay
				WRITE_STATUS ( OKAY ) ;

				// add on CR to name string
				strcat ( name, "\x0D" ) ;

				// send byte count of pathname to follow
        bsize = strlen ( name ) ;
				write_data ( &bsize, 1 ) ;

				// write filename
        write_data ( name, bsize ) ;
			}
			else
				WRITE_STATUS ( END_OF_FILE ) ;
			break ;
		}
		case FILE_DEV_READY :
		{
			// new in V2.2
			fcb = file_packet[0] ;
			output_message() ;

			if ( device_ready ( fcb ) )
				WRITE_STATUS ( OKAY ) ;
			else
				WRITE_STATUS ( NOT_READY ) ;
			break ;
		}
		case SET_COM_CONFIG :
		{
			set_com_config ( file_packet[0], file_packet[1], file_packet[2],
				file_packet[3], file_packet[4] ) ;
			break ;
		}
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION PROCESS_VID_CTRL_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON October 29 1997, FOR V3.02 BY J.Bird
MOD ON June 21 1998, FOR V4.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Process_Vid_Ctrl_Message ( Command, Video_Packet ) is
*/
void process_vid_ctrl_message ( unsigned char command, char *video_packet )
/*

DESCRIPTION

This operation is responsible for interpreting the video data packet, and
passing it in the required format to the Video_Ctrl object for execution.

USES OPERATIONS
 OUTPUT_MESSAGE
 VIDEO_CTRL.GRAPHICS_MODE_CHANGE
 VIDEO_CTRL.GRAPHICS_CURSOR_UPDATE
 VIDEO_CTRL.GRAPHICS_SET_POINT
 VIDEO_CTRL.GRAPHICS_DRAW_LINE
 VIDEO_CTRL.GRAPHICS_DRAW_CIRCLE
 VIDEO_CTRL.GRAPHICS_COLOUR_CHANGE
 VIDEO_CTRL.GRAPHICS_CLEAR_SCREEN
 VIDEO_CTRL.GRAPHICS_READ_PIXEL

Types/Data
 X,Y
 X1,Y1
 X2,Y2
 Mode_id
 Radius
 Colour
*/
{
	short int i, cr_posn ;
	float x_scale, y_scale ;
	char *paste_buffer ;
	unsigned short buffer_size ;
  unsigned char bsize, win_id ;
/*

PSUEDO CODE

 case Command is

 when graphics_mode_change

	 Read mode id from packet ;
	 Video_Ctrl.Graphics_Mode_Change ( Mode_Id ) ;

 when graphics_move_cursor

	 Extract XY co-ord from packet ;
	 Video_Ctrl.Graphics_Cursor_Update ( X, Y ) ;

 when graphics_colour_change

	 Extract colour from packet ;
	 Video_Ctrl.Graphics_Colour_Change ( Colour ) ;

 when graphics_set_point

	 Extract XY co-ord from packet ;
	 Extract colour from packet ;
	 Video_Ctrl.Graphics_Set_Point ( X, Y, Colour ) ;

 when graphics_clear_screen

	 Extract colour from packet ;
	 Video_Ctrl.Graphics_Clear_Screen ( Colour ) ;

 when graphics_draw_line

	 Extract XY1 co-ord from packet ;
	 Extract XY2 co-ord from packet ;
	 Extract colour from packet ;
	 Video_Ctrl.Graphics_Draw_Line ( X1, Y1, X2, Y2, Colour ) ;

 when graphics_draw_circle

	 Extract radius from packet ;
	 Extract XY co-ord from packet ;
	 Extract colour from packet ;
	 Video_Ctrl.Graphics_Draw_Circle ( X, Y, Radius, Colour ) ;

 when graphics_read_pixel_colour

	 Extract XY co-ord from packet ;
	 Colour := Video_Ctrl.Graphics_Read_Pixel ( X, Y ) ;
	 Store Colour in packet ;
	 Output_Message ( Graphics_Read_Pixel_Colour, Video_Packet ) ;

 end case ;

CODE
*/

	switch ( command )
	{
		case GRAPHICS_MODE_CHANGE :
		{
			if ( active_windows[selected_window] != NULL )
			{
				// get scaling multiply factor from source data
				// 0 = +ve, 1 = -ve
				if ( !video_packet[4] )
					x_scale = 1 ;
				else
					x_scale = -1 ;
				if ( !video_packet[5] )
					y_scale = 1 ;
				else
					y_scale = -1 ;

				// activate graphics mode with supplied parameters
				active_windows[selected_window]->graphics_set_mode (
					fetch_word ( &video_packet[0] ), fetch_word ( &video_packet[2] ),
					x_scale, y_scale ) ;
			}
			break ;
		}
		case GRAPHICS_MOVE_CURSOR :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_cursor_update (
					fetch_word ( &video_packet[0] ), fetch_word ( &video_packet[2] ) ) ;
			break ;
		}
		case GRAPHICS_COLOUR_CHANGE :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_colour_change ( video_packet[0] ) ;
			break ;
		}
		case GRAPHICS_SET_POINT :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_set_point (
					fetch_word ( &video_packet[0] ),
						fetch_word ( &video_packet[2] ), video_packet[4] ) ;
			break ;
		}
		case GRAPHICS_CLEAR_SCREEN :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_clear_screen ( video_packet[0] ) ;
			break ;
		}
		case GRAPHICS_DRAW_LINE :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_draw_line (
					fetch_word ( &video_packet[0] ),
						fetch_word ( &video_packet[2] ), fetch_word ( &video_packet[4] ),
							fetch_word ( &video_packet[6] ), video_packet[8],
              video_packet[9], video_packet[10] ) ;
			break ;
		}
		case GRAPHICS_DRAW_CIRCLE :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_draw_circle (
					fetch_word ( &video_packet[2] ),
						fetch_word ( &video_packet[4] ), fetch_word ( &video_packet[0] ),
							video_packet[6], video_packet[7], video_packet[8] ) ;
			break ;
		}
		case GRAPHICS_READ_PIXEL_COLOUR :
		{
			output_message() ;
      bsize = 1 ;
			write_data ( &bsize, 1 ) ;      //V2.1 bug fix - message length
			if ( active_windows[selected_window] != NULL )
      {
        bsize = active_windows[selected_window]->graphics_read_pixel (
					fetch_word ( &video_packet[0] ), fetch_word ( &video_packet[2] ) ) ;
				write_data ( &bsize, 1 ) ;
      }
			else
      {
        bsize = 0 ;
				write_data ( &bsize, 1 ) ;
      }
			break ;
		}
		case EMULATE_SELECT :
		{
			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->select_video_emulation ( video_packet[0] ) ;
			break ;
		}

		case SET_TEXT_ATTR :

			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->set_text_attr ( video_packet[0], video_packet[1] ) ;
			break ;

		// new in V3.0, open,close,set,get window msgs
		case OPEN_WINDOW :

			// null terminate the window name
			cr_posn = strcspn ( video_packet, "\x0D" ) ;
			video_packet[cr_posn] = 0 ;
			// issue start of response msg
			output_message() ;
      bsize = 1 ;
			write_data ( &bsize, 1 ) ;
			// init win_id to 0 (default failed to open window)
			win_id = 0 ;
			// loop thru available windows
			for (i=0 ; i<MAX_WINDOWS ; i++ )
			{
				// if window not active, open window & update window id
				if ( active_windows[i] == NULL )
				{
					// begin exception catch for window opening
					try
					{
						active_windows[i] = new window_access ( video_packet, i ) ;
						selected_window = i ;
						win_id = i+1 ;
					}
					catch ( xalloc )
					{
						MessageBox ( Main_hWnd, "Memory Allocation Failure", "Memory Error",
							MB_OK | MB_ICONSTOP ) ;
					}

					i=MAX_WINDOWS ;
				}
			}
			// return either window id or 0 if no free windows
			write_data ( &win_id, 1 ) ;
			break ;

		case CLOSE_WINDOW :

			win_id = video_packet[0]-1 ;
			// de-activate window if it is active
			if ( active_windows[win_id] != NULL )
			{
				delete active_windows[win_id] ;
				active_windows[win_id] = NULL ;
			}
			break ;

		case SET_ACTIVE_WINDOW :

			// make the specified window active
			selected_window = video_packet[0]-1 ;
			break ;

		case GET_ACTIVE_WINDOW :

			// get window id and initiate data handshake
			win_id = video_packet[0]-1 ;
			output_message() ;
			if ( active_windows[win_id] != NULL )
			{
				// retrieve ptr to paste buffer, size & txfr to remote
				paste_buffer = active_windows[win_id]->get_paste_buffer ( &buffer_size ) ;
        bsize = (unsigned char)buffer_size ;
				write_data ( &bsize, 1 ) ;
        write_data ( paste_buffer, bsize ) ;
			}
			else
      {
        bsize = 0 ;
				write_data ( &bsize, 0 ) ;
      }
			break ;

		case ACTIVE_WINDOW :

			// return current selected_window to host
			output_message() ;
      bsize = 1 ;
			write_data ( &bsize, 1 ) ;
			win_id = 0 ;
			for ( i=0 ; i< MAX_WINDOWS ; i++ )
			{
				// active_hwnd maintained by vid_ctrl module
				if ( active_hwnd == active_windows[i]->MainWnd )
					win_id = i ;
			}
      win_id++ ;
			write_data ( &win_id, 1 ) ;
			break ;

		// new in v4.0 - define an entry in the windows palette
		case GRAPHICS_DEFINE_PAL_ENTRY :

			if ( active_windows[selected_window] != NULL )
				active_windows[selected_window]->graphics_define_palette_entry
					( (int)video_packet[0], video_packet[1], video_packet[2],
						video_packet[3] ) ;
			break ;

	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION PROCESS_MISC_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON October 8, 1995, FOR V2.03 BY J.BIRD
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON November 25 1996, FOR V3.0 BY J.Bird
MOD ON October 29 1997, FOR V3.02 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Process_Misc_Message ( Command, Data_Packet ) is
*/
void process_misc_message ( unsigned char command, char *data_packet )
/*
DESCRIPTION

This operation is responsible for interpreting the miscellaneous command data
packet, and passing it on to the Misc_Utils object for execution.

USES OPERATIONS
 MISC_UTILS.PRINT_DATA_BLOCK
 MISC_UTILS.EXECUTE_DOS_COMMAND

Types/Data
 TBD

PSUEDO CODE
 This operation is TBD release 2 software.

CODE
*/
{
	// size of RTC packet info
  // v4.0 - Y2K, provide 4 digit date code
	const unsigned char rtc_packet_size = 7 ;

	// V3.0 - DOS command forking not supported

	// new in V2.2 - read PC clock
	if ( command == READ_RTC )
	{
		read_system_clock ( data_packet ) ;
		output_message() ;
		write_data ( (void*)&rtc_packet_size,1 ) ;
    write_data ( data_packet, rtc_packet_size ) ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
INTERNAL OPERATION PROCESS_SUPERVISOR_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON October 8, 1995, FOR V2.03 BY J.BIRD
MOD ON April 26 1996, FOR V2.1 BY J.Bird
MOD ON November 25 1996, FOR V3.0 BY J.Bird

Procedure Process_Supervisor_Message is
*/
void process_supervisor_message(void)
/*
DESCRIPTION

This operation will determine the type of message being transmitted, and read
in the required number of bytes from the data port. The completed message is
then passed on:

-- if a file transfer message is received, the data packet is passed to
Process_FTP_Message
-- if a print message or DOS command is received, the data packet is passed to
Process_Misc_Message
-- if it is none of the above, the data packet is passed to
Process_Vid_Ctrl_Message

USES OPERATIONS
 PROCESS_FTP_MESSAGE
 PROCESS_MISC_MESSAGE
 PROCESS_VID_CTRL_MESSAGE
 BUFFER_INPUT_DATA
 IN_PORT.READ_DATA

Types/Data
 Command     : byte ;
 Param_Count : byte ;
 Packet_Buffer ;
*/
{
typedef struct {
  unsigned char Cmd ;
  unsigned char Count ; } CmdPktType ;
CmdPktType CmdPkt ;
unsigned char packet_buffer[256] ;
/*
PSUEDO CODE

 Command := In_Port.Read_Data ( Video_Port ) ;

 Param_Count := In_Port.Read_Data ( Video_Port ) ;

 Packet_Buffer := Buffer_Input_Data ( Video_Port, Param_Count ) ;

 if Command = File_Read
 or Command = File_Write
 then

	 Process_FTP_Message ( Command, Packet_Buffer ) ;

 else if Command = Print_Message
	  or Command = DOS_Command
 then

	 Process_Misc_Message ( Command, Packet_Buffer ) ;

 else

	 Process_Video_Ctrl_Message ( Command, Packet_Buffer ) ;

 end if ;

CODE
*/
	// read header info
  read_data ( &CmdPkt, sizeof(CmdPkt) ) ;
  read_data ( packet_buffer, CmdPkt.Count ) ;

	// dispatch message
	if ( ( CmdPkt.Cmd == READ_SECTOR ) || ( CmdPkt.Cmd == WRITE_SECTOR ) )
	{
		process_hdisk_message ( CmdPkt.Cmd, packet_buffer ) ;
		return ;
	}

	if ( CmdPkt.Cmd == READ_RTC )
	{
		process_misc_message ( CmdPkt.Cmd, packet_buffer ) ;
		return ;
	}

	if ( ( CmdPkt.Cmd <= GRAPHICS_READ_PIXEL_COLOUR ) ||
			( CmdPkt.Cmd == EMULATE_SELECT ) ||
			( CmdPkt.Cmd == SET_TEXT_ATTR ) ||
			( CmdPkt.Cmd >= OPEN_WINDOW ) )
	{
		process_vid_ctrl_message ( CmdPkt.Cmd, packet_buffer ) ;
		return ;
	}

	if ( ( ( CmdPkt.Cmd >= FILE_CREATE ) && ( CmdPkt.Cmd <= FILE_GET_FILE_POS ) ) ||
		( ( CmdPkt.Cmd >= FILE_END_OF_FILE ) && ( CmdPkt.Cmd <= READ_DIR_ENTRY ) ) ||
		( ( CmdPkt.Cmd >= FILE_DEV_READY ) && ( CmdPkt.Cmd <= SET_COM_CONFIG ) ) )
	{
		process_file_handler_message( CmdPkt.Cmd, packet_buffer ) ;
		return ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION PROCESS_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON October 8, 1995, FOR V2.03 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure PROCESS_MESSAGE is
*/
void process_message(void)
/*
DESCRIPTION

This operation determines the source of the incoming message and performs the
following processes:

-- if data is available on the HDisk port, calls the Process_HDisk_Message
operation to handle the data.
-- if data is available on the Video port, attempts to identify a message
header.
-- if a message header is identified, then the Process_Supervisor_Message
operation is called to handle the message.
-- if a message header is not identified, then the data is passed to the
Video_Ctrl object as text to be processed.

USES OPERATIONS
 PROCESS_SUPERVISOR_MESSAGE
 PROCESS_HDISK_MESSAGE
 VIDEO_CTRL.OUTPUT_TEXT
 IN_PORT.PORT_STATUS
 IN_PORT.READ_DATA

Types/Data
 Data_Byte : Byte ;
*/
{
unsigned char data_byte ;
/*

PSUEDO CODE

 -- check for video bus data
 if In_Port.Port_Status ( Video_Port )
 then

	 Data_Byte := In_Port.Read_Data ( Video_Port )

	 case Messg_State is

	 When Normal

		 if Data_Byte = Messg_Ident_Char1
		 then

			 Messg_State := Ident_1 ;

		 else

			 Video_Ctrl.Output_Text( Data_Byte ) ;

		 endif ;

	 When Ident_1

		 if Data_Byte = Messg_Ident_Char2
		 then

			 Messg_State := Ident_2 ;

		 else

			 Video_Ctrl.Output_Text ( Messg_Ident_Char1 ) ;
			 Video_Ctrl.Output_Text ( Data_Byte ) ;
			 Messg_State := Normal ;

		 end if ;

	 When Ident_2

		 if Data_Byte = Messg_Ident_Char3
		 then

			 Messg_State := Normal ;
			 Process_Supervisor_Message ;

		 else

			 Video_Ctrl.Output_Text ( Messg_Ident_Char1 ) ;
			 Video_Ctrl.Output_Text ( Messg_Ident_Char2 ) ;
			 Video_Ctrl.Output_Text ( Data_Byte ) ;
			 Messg_State := Normal ;

		  end if ;

		end case ;

 else if In_Port.Port_Status ( HDisk_Port )
 then

	 Process_HDisk_Message ;

 end if ;

CODE
*/
  read_data ( &data_byte, 1 ) ;

	switch ( messg_state )
	{
		case NORMAL :
		{
			if ( data_byte == IDENT_CHAR_1 )
				messg_state = IDENT_1 ;
			else
				if ( active_windows[selected_window] != NULL )
					active_windows[selected_window]->output_text ( data_byte ) ;
			break ;
		}
		case IDENT_1 :
		{
			if ( data_byte == IDENT_CHAR_2 )
				messg_state = IDENT_2 ;
			else
			{
				if ( active_windows[selected_window] != NULL )
					active_windows[selected_window]->output_text ( IDENT_CHAR_1 ) ;
				if ( data_byte != IDENT_CHAR_1 )
				{
					if ( active_windows[selected_window] != NULL )
						active_windows[selected_window]->output_text ( data_byte  ) ;
					messg_state = NORMAL ;
				}
			}
			break ;
		}
		case IDENT_2 :
		{
			if ( data_byte == IDENT_CHAR_3 )
			{
				messg_state = NORMAL ;
				process_supervisor_message() ;
			}
			else
			{
				if ( active_windows[selected_window] != NULL )
				{
					active_windows[selected_window]->output_text ( IDENT_CHAR_1 ) ;
					active_windows[selected_window]->output_text ( IDENT_CHAR_2 ) ;
				}
				if ( data_byte == IDENT_CHAR_1 )
					messg_state = IDENT_1 ;
				else
				{
					if ( active_windows[selected_window] != NULL )
						active_windows[selected_window]->output_text ( data_byte ) ;
					messg_state = NORMAL ;
				}
			}
			break ;
		}
		default : break ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION MESSAGE_EVENT
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD

Function MESSAGE_EVENT return boolean is
*/
int message_event(void)
/*
DESCRIPTION

This operation determines if data is waiting to be processed and returns a
boolean flag to indicate the state. Each of the valid input ports is examined
in turn to determine the status.

USES OPERATIONS
 IN_PORT.PORT_STATUS

Types/Data
 None

PSUEDO CODE

 if In_Port.Port_Status ( Video_Port )
 then

	 return true ;

 else if In_Port.Port_Status ( HDisk_Port )
 then

	 return true ;

 else

	 return false ;

 end if ;

CODE
*/
{
	return in_port_status() ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION OUTPUT_MESSAGE
CREATED ON February 1, 1995, 07:14 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD

Procedure Output_Message ( Command, Data_Packet ) is
*/
void output_message ( void )
/*
DESCRIPTION

This operation will send the requested video port message id & length followed
by the data packet supplied. The data packet should comprise the parameter
data for the required command.

USES OPERATIONS
 OUT_PORT.WRITE_DATA

Types/Data
 Message_Length ;
*/

/*
PSUEDO CODE

 Determine_Message_Length ;
 Out_Port.Write_Data ( Video_Port, Command ) ;
 Out_Port.Write_Data ( Video_Port, Message_Length ) ;

 Transfer_Output_Buffer ( Video_Port, Data_Packet, Message_Length ) ;

CODE
*/
{
  const unsigned char Hdr[] = { IDENT_CHAR_1, IDENT_CHAR_2, IDENT_CHAR_3 } ;
  write_data ( (void*)Hdr, sizeof ( Hdr ) ) ;
}
/*
END.
*/

void open_message_processor(void)
{
	open_hdisk() ;
	set_port_base ( system_config.base_addr, system_config.irq, system_config.dma ) ;
//	set_port_base ( 0x378, 0x7, 0 ) ;

}

void close_message_processor(void)
{
	int i ;
	close_hdisk() ;
	// delete active windows
	for (i=0 ; i<MAX_WINDOWS ; i++ )
	{
		if ( active_windows[i] != NULL )
			delete active_windows[i] ;
	}
	close_io_port() ;
}

