/* IO module for standard LPT parallel port for PC-Share V4.1x or later. */

#ifdef _Windows
	#include <windows.h>
#endif
#ifdef __WIN32__
	#include "portio32.h"
#else
	#error 16-bit builds not supported post PCShare v4.0x
#endif

#include "globals.h"
#include "io_ports.h"

// port states
#define INPUT 0
#define OUTPUT 1

// lpt offsets
#define DATA_REG 0
#define STATUS_REG 1
#define CTRL_REG 2

// lpt data
#define STROBE_BIT 0x01
#define DIR_BIT 32
#define ACK_BIT 64

// current direction
int current_dir, base_addr ;

// internals
void init_out_port ( void ) ;
void init_in_port ( void ) ;
void send_strobe ( void ) ;

#ifdef _Windows
/* the DLL stuff for Windows build */

#ifdef __WIN32__

/* 32-bit version */
#pragma argsused
BOOL WINAPI DllEntryPoint(HINSTANCE hDLLInst, DWORD  fdwReason,
					LPVOID lpvReserved)
{
	return TRUE ;
}

#else

/* 16-bit version */
// Turn off warning: Parameter '' is never used
#pragma argsused
int FAR PASCAL LibMain( HINSTANCE hInstance,
								WORD wDataSegment,
								WORD wHeapSize,
								LPSTR lpszCmdLine )
{

// The startup code for the DLL initializes the local heap(if there is one)
// with a call to LocalInit which locks the data segment.

	 if ( wHeapSize != 0 )
		  UnlockData( 0 );
	 return 1;   // Indicate that the DLL was initialized successfully.
}

// Turn off warning: Parameter '' is never used
#pragma argsused
int FAR PASCAL WEP ( int bSystemExit )
{
	unsigned char irq_mask ;

	 return 1;
}

/* end of win16/32 specific */
#endif

/* end of Win specific */
#endif

// toggle strobe line of printer port
void send_strobe ( void )
{
	unsigned char ctrl_state ;

	ctrl_state = inp ( base_addr+CTRL_REG ) ;
	outp ( base_addr+CTRL_REG, ctrl_state | STROBE_BIT ) ;
	outp ( base_addr+CTRL_REG, ctrl_state & (~STROBE_BIT)) ;
}

// set base addr of port to be used
#pragma argsused
void far __pascal _export set_port_base ( int port_base, int irq, int dma )
{
	unsigned char irq_mask ;
	int int_vect ;

	// set port base
	base_addr = port_base ;

	// clear strobe bit
	outp ( base_addr+CTRL_REG, inp ( base_addr+CTRL_REG ) & (~STROBE_BIT)) ;

	// initialise port for input
	init_in_port() ;
}

// get in_port status
int far __pascal _export in_port_status ( void )
{
	if ( current_dir != INPUT )
		init_in_port() ;

	if ( ( inp ( base_addr+STATUS_REG ) & ACK_BIT ) == 0 )
		return TRUE ;
	else
  {
    Sleep(0) ;
		return FALSE ;
  }
}

// read data block from port
void far __pascal _export read_data ( void *buf, int count )
{
	unsigned char data_read ;
  unsigned char *p = buf ;

	if ( current_dir != INPUT )
		init_in_port() ;

  while ( count )
  {
	  // data available?
	  while ( !in_port_status() )
	  {
	  }

    *p++ = inp ( base_addr + DATA_REG ) ;
	  send_strobe() ;
    count-- ;
  }
}

// initialise 8255 for input
void init_in_port ( void )
{

	current_dir = INPUT ;

	// set port for input
	outp ( base_addr+CTRL_REG, inp ( base_addr+CTRL_REG ) | DIR_BIT ) ;

	// compatability with 8255 - do 'reset' pulse
	send_strobe() ;
}

// initialise 8255 for output
void init_out_port ( void )
{
	// set port for output
	outp ( base_addr+CTRL_REG, inp ( base_addr+CTRL_REG ) & (~DIR_BIT) ) ;

	outp ( base_addr+DATA_REG, 0 ) ;
	current_dir = OUTPUT ;

	// compatability with 8255 - do 'reset' pulse
	send_strobe() ;
}

// write data block to outport port
void far __pascal _export write_data ( void *buf, int count )
{
  unsigned char *p = buf ;

	if ( current_dir != OUTPUT )
	{
		init_out_port() ;
		outp ( base_addr+DATA_REG, *p++ ) ;
    count-- ;
	}
	else
	{
		outp ( base_addr+DATA_REG, *p++ ) ;
		send_strobe() ;
    count-- ;
	}

  while ( count )
  {
	  while ( !out_port_status() )
	  {
	  }
		outp ( base_addr+DATA_REG, *p++ ) ;
		send_strobe() ;
    count-- ;
  }
}

// get status of output port
// returns TRUE if data still due to go out
int far __pascal _export out_port_status ( void )
{
	if ( current_dir != OUTPUT )
		init_out_port() ;

	if ( ( inp ( base_addr+STATUS_REG ) & ACK_BIT ) == 0 )
		return TRUE ;
	else
  {
    Sleep(0) ;
		return FALSE ;
  }
}

#ifndef _WINNT
// does nothing for all but formal NT build
void far __pascal _export close_io_port ( void )
{
}
#endif
