/* IO module for 6809 emulator to link into
	PCShare V3.x - 32bit version. */

#include <windows.h>
#include <alloc.h>
#include <string.h>
#include "io_ports.h"

/* structure of the fifo used for param passing */
#define FIFO_SIZE 0xA000
struct fifo_type {
	unsigned short read_pos ;
	unsigned short write_pos ;
	unsigned short count ;
	unsigned char fifo[FIFO_SIZE] ; } ;

// the fifos - private to each process
static struct fifo_type *in_fifo, *out_fifo ;

// shared memory handle
static LPVOID lpvmem ;
static HANDLE hMapObject = NULL;   /* handle to file mapping */
static const int shared_mem_size = sizeof ( struct fifo_type ) * 2 ;

/* 32-bit version */
// use Win32 file mapping to setup shared in and out
// fifos which will be used to pass data in & out of PCShare
#pragma argsused
BOOL WINAPI DllEntryPoint(HINSTANCE hDLLInst, DWORD  fdwReason,
					LPVOID lpvReserved)
{
	BOOL fInit ;

	switch (fdwReason)
	{
		/*
		 * The DLL is attaching to a process due to process
		 * initialization or a call to LoadLibrary.
		 */
		 case DLL_PROCESS_ATTACH:

			/* Create a named file mapping object */
			hMapObject = CreateFileMapping(
				 (HANDLE) 0xFFFFFFFF, /* use paging file    */
				 NULL,                /* no security attr.  */
				 PAGE_READWRITE,      /* read/write access  */
				 0,                   /* size: high 32-bits */
				 shared_mem_size,     /* size: low 32-bits  */
				 "dllmemfilemap");    /* name of map object */
			if (hMapObject == NULL)
				 return FALSE;

			/* The first process to attach initializes memory. */
			fInit = (GetLastError() != ERROR_ALREADY_EXISTS);

			/* Get a pointer to the file-mapped shared memory */
			lpvmem = MapViewOfFile(
					 hMapObject,     /* object to map view of    */
					 FILE_MAP_WRITE, /* read/write access        */
					 0,              /* high offset:   map from  */
					 0,              /* low offset:    beginning */
					 0);             /* default: map entire file */

			if ( lpvmem == NULL)
				 return FALSE;

			/* init and setup fifos */
			if (fInit)
			{
				/* setup for first process */
				memset(lpvmem, '\0', shared_mem_size );
				in_fifo = (struct fifo_type*)lpvmem ;
				out_fifo = in_fifo+1 ;
			}
			else
			{
				/* second process transposes mapping so process 1's
				in becomes 2's out etc. */
				out_fifo = (struct fifo_type*)lpvmem ;
				in_fifo = out_fifo+1 ;
			}
			break;

		 case DLL_PROCESS_DETACH:

			/* Unmap shared memory from the process's address space. */
			UnmapViewOfFile( lpvmem );

			/* Close the process's handle to the file-mapping object. */
			CloseHandle(hMapObject);
			break;

		 default:
			break;
	}
	return TRUE;
}

// set base addr of port to be used
#pragma argsused
void far __pascal _export set_port_base ( int port_base, int irq, int dma )
{
	// does nothing
}

// get in_port status
int far __pascal _export in_port_status ( void )
{
	return ( ( in_fifo->count != 0 ) ) ;
}

// read a data byte from fifo
unsigned char far __pascal _export read_data ( void )
{
	unsigned char ret_byte ;

	while ( !in_port_status() )
	{
	}
	ret_byte = in_fifo->fifo[in_fifo->read_pos] ;
	in_fifo->read_pos++ ;
	if ( in_fifo->read_pos == FIFO_SIZE )
		in_fifo->read_pos = 0 ;
	in_fifo->count-- ;

	return ret_byte ;
}

// initialise 8255 for input
void init_in_port ( void )
{
}

// initialise 8255 for output
void init_out_port ( void )
{
}

// write data byte to outport port
void far __pascal _export write_data ( unsigned char data_byte )
{
	out_fifo->fifo[out_fifo->write_pos] = data_byte ;
	out_fifo->write_pos++ ;
	if ( out_fifo->write_pos == FIFO_SIZE )
		out_fifo->write_pos = 0 ;
	out_fifo->count++ ;

	while ( !out_port_status() )
	{
	}
}

// get status of output port
int far __pascal _export out_port_status ( void )
{
	return ( ( out_fifo->count != FIFO_SIZE ) ) ;
}

void far __pascal _export close_io_port ( void )
{
}

