// RS232 interface package access protocol will
// differ depending on whether this is a DOS or
// Windows application. The compiler will decide which
// variant to build based on the _Windows macro defined by BCC

// MOD ON January 25 1997, FOR PCShare V3.0 BY J.Bird
// MOD ON October 29 1997, FOR V3.02 BY J.Bird
// MOD ON January 01 1998, FOR V3.1 BY J.Bird

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include "rs232.h"

HANDLE hCom ;
int device_handle ;
BOOL device_open = FALSE ;

// read byte from COM port
unsigned char read_rs232 ( void )
{
	unsigned char data ;
	unsigned long count ;

	// wait for data available at port
	while ( !rs232_ready() ) ;

#ifndef __WIN32__

	ReadComm ( device_handle, &data, 1 ) ;

#else

	ReadFile ( hCom, &data, 1, &count, NULL ) ;

#endif

	return data ;
}

// write byte to COM port
void write_rs232 ( unsigned char out_byte )
{
#ifndef __WIN32__

	// add data to tx queue
	WriteComm ( device_handle, &out_byte, 1 ) ;

#else

	unsigned long count ;

	WriteFile ( hCom, &out_byte, 1, &count, NULL ) ;

#endif
}

int init_rs232 ( serial_232_t *port_init )
{
	DCB dms_config ;
	char temp_str[50], parity ;
	const UINT pcsbaud_to_baud[] = { 110, 150, 300, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 0 } ;
	const char pcsparity_to_parity[] = { 'n', 'o', 'e', 'm', 's' } ;

	if ( device_open )
		return FALSE ;

	// create COMx string from in params
	sprintf ( temp_str, "COM%d", port_init->comport ) ;

#ifndef __WIN32__

	// open the comms port with a 1K TX/RX bfr
	device_handle = OpenComm ( temp_str, 1024, 1024 ) ;
	if ( device_handle < 0 )
		return FALSE ;
	GetCommState ( device_handle, &dms_config ) ;
#else

	hCom = CreateFile( temp_str,
	 GENERIC_READ | GENERIC_WRITE,
	 0,    /* comm devices must be opened w/exclusive-access */
	 NULL, /* no security attrs */
	 OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
	 0,    /* not overlapped I/O */
	 NULL  /* hTemplate must be NULL for comm devices */
	 );

	if (hCom == INVALID_HANDLE_VALUE)
		return FALSE ;
	GetCommState ( hCom, &dms_config ) ;

#endif

	// create DOS type MODE string from input parameters
	sprintf ( temp_str, "COM%d:%d,%c,%d,%d", port_init->comport,
		pcsbaud_to_baud[port_init->baud], pcsparity_to_parity[port_init->parity],
		port_init->bits, port_init->stops ) ;

	// create COM DCB for port
	BuildCommDCB ( temp_str, &dms_config ) ;

#ifndef __WIN32__
	// configure device
	SetCommState ( &dms_config ) ;
#else
	SetCommState ( hCom, &dms_config ) ;
#endif

	device_open = TRUE ;

	return TRUE ;
}

void de_init_rs232 ( void )
{
#ifndef __WIN32__
	CloseComm ( device_handle ) ;
#else
	CloseHandle ( hCom ) ;
#endif

	device_open = FALSE ;

}

int rs232_ready ( void )
{
	COMSTAT com_status ;
	unsigned long errors ;

#ifndef __WIN32__
	GetCommError ( device_handle, &com_status ) ;
#else
	ClearCommError ( hCom, &errors, &com_status ) ;
#endif
	return com_status.cbInQue ;
}

