/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PACKAGE VIDEO_CTRL
CREATED ON February 12, 1995, 04:27 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON February 19, 1996, FOR V2.1 BY J.Bird
MOD ON October 06, 1996, FOR V3.0 BY J.Bird

Package Video_Ctrl is

DESCRIPTION

The Video_Ctrl package is responsible for handling the video related functions
that are requested by the external computer. These will involve textual data
to be displayed, and graphics operations to be executed.

The package comprises 2 internal objects:

-- Graphics for handling all of the graphical operations.
-- VT100_Emul for performing the DEC VT100 terminal emulation.

The majority of the video processing is handled by these two objects, the
Video_Ctrl object primarily serving the purpose of a single interface to the
outside world.

REQUIRED INTERFACE
 GRAPHICS
 VT100_EMUL
*/
#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include "pcshare.h"
#include "config.h"
#include "commemu.h"
#include "graphics.h"
#include "video_ctrl.h"
#include "go51_emu.h"
#include "vt100_em.h"
#include <toolhelp.h>
#include "fileio.h"

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif
/*
Types/Data
*/

// paint operations
#define MARK_COPY 1
#define REDRAW_ALL 2

// text attributes
#define ATTR_NONE 0x0
#define ATTR_REV 0x1
#define ATTR_BOLD 0x2
#define ATTR_ULINE 0x4
#define ATTR_BLINK 0x8
#define ATTR_INVIS 0x10

window_access *active_windows[MAX_WINDOWS] = { NULL } ;
static int modules_active=0 ;
UINT cursor_timer ;
HWND active_hwnd ;

window_access *find_owner_class( HWND hWnd ) ;
void host_port_delay(void) ;
void CALLBACK _export CursorTimer ( HWND, UINT, UINT, DWORD ) ;
BOOL CALLBACK _export DisplayOptions (HWND, UINT, WPARAM, LPARAM);

//struct text_info save_info ;
/*
Provides Operations
 OUTPUT_TEXT
 GRAPHICS_MODE_CHANGE
 GRAPHICS_CURSOR_UPDATE
 GRAPHICS_SET_POINT
 GRAPHICS_DRAW_LINE
 GRAPHICS_DRAW_CIRCLE
 GRAPHICS_READ_PIXEL
 GRAPHICS_COLOUR_CHANGE
 GRAPHICS_CLEAR_SCREEN
*/
//BOOL CALLBACK _export EditProc (HWND, UINT, WPARAM, LPARAM);

/*
Internal Operations
 None

END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION OUTPUT_TEXT
CREATED ON February 12, 1995, 04:27 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON February 19, 1996, FOR V2.1 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Output_Text ( Character ) is
*/
void window_access::output_text ( char textcr )
/*
DESCRIPTION

This operation is responsible for displaying a text character on the screen.
It uses the VT100_Emul object to determine:

-- if an escape sequence is in progress, then the character is passed to the
object for processing.

-- if the character signifies the start of an escape sequence, then the
chracter is passed to the object for processing.

-- otherwise it is displayed at the current cursor position.

USES OPERATIONS
 VT100_EMUL.VT_DETECT
 VT100_EMUL.VT_IN_PROGRESS
 VT100_EMUL.PROCESS_VT_SEQUENCE

Types/Data
 None

PSUEDO CODE

 if VT100_Emul.VT_In_Progress
 then

	  VT100_Emul.Process_VT_Sequence ( Character ) ;

 else if VT100_Emul.VT_Detect ( Character )
 then

	  VT100_Emul.Process_VT_Sequence ( Character ) ;

 else

	  Print ( Character ) ;

 end if ;

CODE
*/
{
	// add text character to fifo
	text_fifo[write_ptr]=textcr ;
	write_ptr++ ;
	if ( write_ptr == FIFO_SIZE )
	{
		write_ptr = 0 ;
	}

	// if copy region is currently selected, remove it
	if ( copying_selected )
		clear_copy_region() ;

	if ( !paint_pending )
	{
		// request paint message to command shell window
		InvalidateRect ( CommHwnd, NULL, FALSE ) ;
		paint_pending = TRUE ;
	}
	host_port_delay() ;
}
/*
END.
*/

// set emulation defaults
void window_access::select_video_emulation ( unsigned char emulation_code )
{
	display_options.emulation = emulation_code ;

	switch ( emulation_code )
	{
		case VT100_EMUL :

			// for VT100 displays, set a fixed 80*25 display
			display_options.max_x = 80 ;
			display_options.max_y = 25 ;
			display_options.wrap_flag = TRUE ;
			display_options.fixed_size = TRUE ;
			break ;

		case GO51_EMUL :

			// for GO51 mode set a fixed 80*24 display
			display_options.max_x = 80 ;
			display_options.max_y = 24 ;
			display_options.wrap_flag = TRUE ;
			display_options.fixed_size = TRUE ;
			break ;
	}
	set_display_options() ;
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_MODE_CHANGE
CREATED ON February 12, 1995, 04:27 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Mode_Change ( Mode_Id ) is
*/
void window_access::graphics_set_mode ( unsigned x_pixels, unsigned y_pixels, float x_scale, float y_scale )
/*
DESCRIPTION

This operation initiates a Graphics mode change by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.MODE_CHANGE

Types/Data
 None

PSUEDO CODE

 Graphics.Mode_Change ( Mode_Id ) ;

CODE
*/
{
	HDC hdc ;

	if ( graphic_win != NULL )
	{
		delete graphic_win ;
		graphic_win = NULL ;
		paint_op = REDRAW_ALL ;
		return ;
	}

	hdc = GetDC ( CommHwnd ) ;
	graphic_win = new graphics_support ( hdc, x_pixels, y_pixels, x_scale, y_scale );
	ReleaseDC ( CommHwnd, hdc ) ;
	paint_op = REDRAW_ALL ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_CURSOR_UPDATE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Cursor_Update ( X_Coord, Y_Coord ) is
*/
void window_access::graphics_cursor_update ( unsigned x_coord, unsigned y_coord )
/*
DESCRIPTION

This operation initiates a Graphics cursor update by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.CURSOR_UPDATE

Types/Data
 None

PSUEDO CODE

 Graphics.Cursor_Update ( X_Coord, Y_Coord ) is

CODE
*/
{
	if ( graphic_win != NULL )
		graphic_win->cursor_update ( x_coord, y_coord ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_SET_POINT
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Set_Point ( X_Coord, Y_Coord, Colour ) is
*/
void window_access::graphics_set_point ( unsigned x_coord, unsigned y_coord, unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics set point by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.SET_POINT

Types/Data
 None

PSUEDO CODE

 Graphics.Set_Point ( X_Coord, Y_Coord, Colour ) ;

CODE
*/
{
	if ( graphic_win != NULL )
	{
		graphic_win->set_point ( x_coord, y_coord, colour ) ;
		paint_op = REDRAW_ALL ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_DRAW_LINE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Draw_Line ( X1_Coord, Y1_Coord, X2_Coord, Y2_Coord, Colour
) is
*/
void window_access::graphics_draw_line ( unsigned x1_coord, unsigned y1_coord, unsigned x2_coord, unsigned y2_coord, unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics draw line by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.DRAW_LINE

Types/Data
 None

PSUEDO CODE

 Graphics.Draw_Line ( X1_Coord, Y1_Coord, X2_Coord, Y2_Coord, Colour ) ;

CODE
*/
{
	if ( graphic_win != NULL )
	{
		graphic_win->draw_line ( x1_coord, y1_coord, x2_coord, y2_coord, colour ) ;
		paint_op = REDRAW_ALL ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_DRAW_CIRCLE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Draw_Circle ( X_Coord, Y_Coord, Radius, Colour ) is
*/
void window_access::graphics_draw_circle ( unsigned x_coord, unsigned y_coord, unsigned radius, unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics draw circle by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.DRAW_CIRCLE

Types/Data
 None

PSUEDO CODE

 Graphics.Draw_Circle ( X_Coord, Y_Coord, Radius, Colour ) ;

CODE
*/
{
	 if ( graphic_win != NULL )
	 {
		graphic_win->draw_circle ( x_coord, y_coord, radius, colour ) ;
		paint_op = REDRAW_ALL ;
	 }
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_READ_PIXEL
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Function Graphics_Read_Pixel ( X_Coord, Y_Coord ) return colour is
*/
unsigned char window_access::graphics_read_pixel ( unsigned x_coord, unsigned y_coord )
/*
DESCRIPTION

This operation initiates a Graphics read pixel by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.READ_PIXEL

Types/Data
 None

PSUEDO CODE

 return Graphics.Read_Pixel ( X_Coord, Y_Coord ) ;

CODE
*/
{
	if ( graphic_win != NULL )
		return ( graphic_win->read_pixel ( x_coord, y_coord ) ) ;
	else
		return 0 ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_COLOUR_CHANGE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Colour_Change ( Colour ) is
*/
void window_access::graphics_colour_change ( unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics colour change by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.COLOUR_CHANGE

Types/Data
 None

PSUEDO CODE

 Graphics.Colour_Change ( Colour ) ;

CODE
*/
{
	 if ( graphic_win != NULL )
		graphic_win->colour_change ( colour ) ;
}

/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_CLEAR_SCREEN
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Clear_Screen ( Colour ) is
*/
void window_access::graphics_clear_screen ( unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics clear screen by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.CLEAR_SCREEN

Types/Data
 None

PSUEDO CODE

 Graphics.Clear_Screen ( Colour ) ;

CODE
*/
{
	 if ( graphic_win != NULL )
	 {
		graphic_win->clear_window ( colour ) ;
		paint_op = REDRAW_ALL ;
	 }
}
/*
END.
*/

// initialise window
window_access::window_access( char *window_name )
{
HDC hDC ;
TEXTMETRIC tm ;
RECT rc ;
int i,j ;

	hnewsfont = CreateFontIndirect((LPLOGFONT) &system_config.font );
	font_definition = system_config.font ;

	// initialise variables and flags
	cursor_x=0 ;
	cursor_y=0 ;
	x_char = 0 ;
	y_char = 0 ;
	read_ptr = 0 ;
	write_ptr = 0 ;
	paint_pending = FALSE ;
	paint_op = 0 ;
	copying_active = FALSE ;
	copying_selected = FALSE ;
	keyb_write = 0 ;
	paste_buffer = NULL ;
	keystate = NORMAL_CHAR ;
	keymap = system_config.keymap ;
	keypress_discard = FALSE ;
	screen_emulator = NULL ;

	// init text_map to all spaces
	memset ( text_map, 32, MAX_POS_X*MAX_POS_Y ) ;
	// init attribute map
	memset ( attr_map, ATTR_NONE, MAX_POS_X*MAX_POS_Y ) ;

	// create command shell
	CommHwnd = CreateDialog ( hInst, MAKEINTRESOURCE ( EDIT_WINDOW ), Main_hWnd, (DLGPROC)EditProc ) ;

	SetWindowText ( CommHwnd, window_name ) ;

	// get text sizes and shell dimensions
	hDC = GetDC ( CommHwnd ) ;
	holdsfont=SelectObject(hDC, hnewsfont);

	GetTextMetrics(hDC, &tm);
	x_char = tm.tmAveCharWidth ;
	y_char = tm.tmHeight + tm.tmExternalLeading;
	graphic_win = NULL ;

	// get max size of characters in window
	GetClientRect ( CommHwnd, &rc ) ;
	max_y = ( rc.bottom - rc.top ) / y_char ;
	max_x = ( rc.right - rc.left ) / x_char ;
	display_options.max_x = max_x ;
	display_options.max_y = max_y ;
	display_options.fixed_size = FALSE ;
	display_options.wrap_flag = FALSE ;

	// set video emulation & screen colours to system default
	select_video_emulation ( system_config.emulation ) ;
	fore_colour = system_config.fore_colour ;
	back_colour = system_config.back_colour ;

	SelectObject ( hDC, holdsfont ) ;
	ReleaseDC ( CommHwnd, hDC ) ;

	cursor_state = FALSE ;
	display_attr = ATTR_NONE ;

	if ( modules_active == 0 )
	{
		// setup cursor timer
		cursor_timer = SetTimer ( NULL, 0, 500, (TIMERPROC)CursorTimer ) ;
	}
	// increment count of windows active
	modules_active++ ;
}

// close window - by remote
window_access::~window_access()
{

	delete screen_emulator ;

	if ( graphic_win != NULL )
		delete graphic_win ;

	// release paste buffer if still allocated
	if ( paste_buffer != NULL )
		free ( paste_buffer ) ;

	// kill cursor timer if last module active
	modules_active-- ;
	if ( modules_active == 0 )
	{
		KillTimer ( NULL, cursor_timer ) ;
	}

	DeleteObject(hnewsfont) ;
	DestroyWindow ( CommHwnd ) ;
}

// general window paint routine. Called whenever a paint
// message is sent to the command window
void window_access::paint_window ( HDC hDC )
{
char out_str[200] ;
int str_pos = 0 ;
int start_char,i ;
unsigned char op_code ;
BOOL abort_flag, char_processed ;

	// select in font
	holdsfont=SelectObject(hDC, hnewsfont);

	// clear cursor if it's drawn
	if ( cursor_state )
		flash_cursor ( hDC, FALSE ) ;

	// set current drawing attributes
	set_display_attr ( hDC, display_attr ) ;

	// initialise the paint abort flag
	abort_flag = FALSE ;

	// if window has just been sized to push the current_y cursor
	// out of the new window, this scrolls the text and cursor back in
	if ( cursor_y > max_y )
	{
		int delta_y = cursor_y-max_y, y, new_max_y ;

		new_max_y = max_y ;
		max_y = cursor_y ;

		for ( y=0 ; y<delta_y ; y++ )
			scroll_down() ;
		max_y = new_max_y ;
		cursor_y = max_y ;

		// set abort flag to prevent any more painting till
		// the new scroll region comes in
		abort_flag = TRUE ;
	}

	// initialise emulator variables
	op_code = NO_ACTION ;
	char_processed = FALSE ;

	// loop to print new text
	while ( ( read_ptr != write_ptr ) && ( !abort_flag ) )
	{
		// check for any outstanding screen emulation sequences
		// perform all pending sequences until none are left or
		// the abort_flag is set
		do
		{
			op_code = screen_emulator->emulation_sequence_pending() ;
			if ( op_code != NO_ACTION )
				abort_flag = process_emulation_code ( hDC, op_code ) ;
		} while ( (!abort_flag) && (op_code != NO_ACTION) ) ;

		if ( !abort_flag )
		{
			// call screen emulator to check if an escape sequence is in progress
			if ( screen_emulator->emulation_in_progress() )
			{
				// pass the incoming text character for processing & set flag to indicate done
				op_code = screen_emulator->process_emulation_sequence ( text_fifo[read_ptr] ) ;
				char_processed = TRUE ;
			}
			else
			{
				// see if this character is the start of an emulation sequence
				if ( screen_emulator->emulation_detect ( text_fifo[read_ptr] ) )
				{
					// if so, pass to emulator for processing & set flag
					op_code = screen_emulator->process_emulation_sequence ( text_fifo[read_ptr] ) ;
					char_processed = TRUE ;
				}
			}

			// if emulator has just processed this char it can't then be used as
			// part of a normal text string, otherwise add to string being built
			if ( !char_processed )
			{
				int new_x_pos, wrap_x ;

				// absolute x-cursor position for text
				new_x_pos = cursor_x+str_pos ;
				// set possible text wrapping point
				if ( display_options.wrap_flag )
					wrap_x = max_x ;
				else
					wrap_x = MAX_POS_X ;

				// if position is less than wrap point, add to string
				if ( new_x_pos < wrap_x )
				{
					out_str[str_pos]=text_fifo[read_ptr] ;
					text_map[cursor_y][new_x_pos] = text_fifo[read_ptr] ;
					attr_map[cursor_y][new_x_pos] = display_attr ;
					str_pos++ ;
				}
				new_x_pos++ ;

				// if required to wrap text & at wrapping point
				if ( display_options.wrap_flag && new_x_pos == wrap_x )
				{
					// print text accumulated so far
					if ( str_pos !=0 )
					{
						TextOut ( hDC, cursor_x*x_char, cursor_y*y_char, out_str, str_pos ) ;
						str_pos=0 ;
					}
					// now simulate a CR/LF combination
					cursor_x = 0 ;
					char_processed = TRUE ;
					op_code = Line_Feed ;
				}
				// note: if text wrapping is not supported, further text is now ignored
				// until the cursor_x position falls into window range
			}

			if ( char_processed )
			{
				// screen emulator has processed text char
				// and requested an operation be performed
				if ( op_code != NO_ACTION )
				{
					// print text accumulated so far
					if ( str_pos !=0 )
					{
						TextOut ( hDC, cursor_x*x_char, cursor_y*y_char, out_str, str_pos ) ;
						cursor_x+=str_pos ;
						str_pos=0 ;
					}
					// go and process the emulation code, possibly aborting
					// further processing in this paint loop
					abort_flag = process_emulation_code ( hDC, op_code ) ;
				}
				char_processed = FALSE ;
			}

			// increment and wrap text fifo as necessary
			read_ptr++ ;
			if ( read_ptr == FIFO_SIZE )
				read_ptr = 0 ;
		} // abort flag not set
	} // main loop

	// print any remaining text
	if ( str_pos !=0 )
	{
		TextOut ( hDC, cursor_x*x_char, cursor_y*y_char, out_str, str_pos ) ;
		cursor_x+=str_pos ;
	}

	// all painting can be stopped by setting the abort flag
	if ( !abort_flag )
	{
		// if the paint has not been aborted, and a request to
		// re-draw the entire display has occured OR
		// if paint_pending hasn't been set then this paint message
		// has not been induced internally then re-draw it anyway
		if ( (paint_op == REDRAW_ALL) || ( !paint_pending ) )
		{
			// if in mid-copy draw with highlighted region else re-draw normally
			if ( copying_selected )
				paint_copy_region ( hDC ) ;
			else
			{
				// first re-draw graphic bitmap
				if ( graphic_win != NULL )
				{
					RECT rc ;

					GetClientRect ( CommHwnd, &rc ) ;
					graphic_win->txfr_to_display_window ( hDC, rc.right-rc.left, rc.bottom-rc.top ) ;

					// ensure text bkgnd does not destroy graphics
					SetBkMode ( hDC, TRANSPARENT ) ;
				}
				// redraw text
				redraw_display( hDC ) ;
			}
		}

		// if paint msg caused by a request to mark a copy region
		if ( paint_op == MARK_COPY )
			paint_copy_region ( hDC ) ;

		// flash the cursor
		cursor_state=!cursor_state ;
		flash_cursor ( hDC, cursor_state ) ;

		// clear down flags
		paint_op = 0 ;
		paint_pending = FALSE ;
	}
	else
	{
		// if abort flag has been set, then a self induced paint is required
		// can't manually post a WM_PAINT message or call InvdalidateRect
		// so post a user defined msg which will be picked up by the
		// dialog box scheduler
		PostMessage ( CommHwnd, WM_COMMAND, AUTO_REPAINT, 0 ) ;
	}
	SelectObject ( hDC, holdsfont ) ;
}

// adjust display extents due to windows sizing operation
void window_access::resize_window ( int x_size, int y_size )
{
	max_x = x_size / x_char ;
	max_y = ( y_size / y_char ) - 1 ;

	// clip display size to maximum specified if in fixed mode
	if ( display_options.fixed_size )
	{
		if ( max_x > display_options.max_x )
			max_x = display_options.max_x ;
		if ( max_y > display_options.max_y )
			max_y = display_options.max_y ;
	}

	paint_op = REDRAW_ALL ;

	// if graphic mode is active, will need to re-draw entire
	// window not just the newly uncovered bit
	if ( graphic_win != NULL )
		InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

// draw or clear the cursor from this device context
void window_access::flash_cursor ( HDC hDC, BOOL draw_flag )
{
RECT rc ;
HBRUSH cursor_brush ;

	rc.left = cursor_x*x_char ;
	rc.top = cursor_y*y_char ;
	rc.right = rc.left+x_char ;
	rc.bottom = rc.top+y_char ;

	if ( !draw_flag )
	{
		cursor_brush = CreateSolidBrush ( back_colour ) ;
		FillRect ( hDC, &rc, cursor_brush ) ;
		draw_attr_text ( hDC, cursor_x, cursor_y, 1 ) ;
	}
	else
	{
		cursor_brush = CreateSolidBrush ( fore_colour ) ;
		FillRect ( hDC, &rc, cursor_brush ) ;
	}
	DeleteObject ( cursor_brush ) ;
}

// draws text onto the device context at the x & y cursor co-ordinates
// from the text and attribute maps
void window_access::draw_attr_text ( HDC hDC, int x, int y, int length )
{
	unsigned char op_code, saved_attr ;
	int new_length, sx ;

	// save current display attribute
	saved_attr = display_attr ;

	// initialise the display attributes from the current attribute map
	op_code = attr_map[y][x] ;
	set_display_attr ( hDC, op_code ) ;

	// initialise local counters - new string length and starting x-pos
	new_length = 0 ;
	sx = x ;

	// while there's still stuff to do
	while ( length > 0 )
	{
		// if attributes match current, update counters
		if ( attr_map[y][x] == op_code )
		{
			new_length++ ;
			length-- ;
			x++ ;
		}
		else
		{
			// output done so far
			TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], new_length ) ;
			// get new code and adjust context appropriately
			op_code = attr_map[y][x] ;
			set_display_attr ( hDC, op_code ) ;
			new_length = 0 ;
			sx=x ;
		}
	}
	// print any part processed string
	if ( new_length != 0 )
		TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], new_length ) ;

	// restore current attributes
	set_display_attr ( hDC, saved_attr ) ;
	display_attr = saved_attr ;
}

// re-draw entire text window based on text map
void window_access::redraw_display( HDC hDC )
{
int y ;

	for ( y=0 ; y<=max_y ; y++ )
		draw_attr_text ( hDC, 0, y, max_x ) ;
}

// scroll the window down adjusting the text map accordingly
void window_access::scroll_down()
{
int y_pos ;
RECT rc ;

	if ( display_options.fixed_size )
	{
		rc.left = 0 ;
		rc.top = 0 ;
		rc.right = max_x * x_char ;
		rc.bottom = max_y * y_char ;
		ScrollWindow ( CommHwnd, 0, -y_char, &rc, NULL ) ;
	}
	else
		ScrollWindow ( CommHwnd, 0, -y_char, NULL, NULL ) ;

	for ( y_pos=0 ; y_pos < max_y ; y_pos++ )
	{
		memcpy ( &text_map[y_pos][0], &text_map[y_pos+1][0], MAX_POS_X ) ;
		memcpy ( &attr_map[y_pos][0], &attr_map[y_pos+1][0], MAX_POS_X ) ;
	}
	memset ( &text_map[cursor_y][0], 32, MAX_POS_X ) ;
	memset ( &attr_map[cursor_y][0], ATTR_NONE, MAX_POS_X ) ;

}

// scroll the window up adjusting the text map accordingly
void window_access::scroll_up()
{
int y_pos ;
RECT rc ;

	// for fixed sized windows, only scroll the fixed area
	if ( display_options.fixed_size )
	{
		rc.left = 0 ;
		rc.top = 0 ;
		rc.right = max_x * x_char ;
		rc.bottom = max_y * y_char - y_char ;
		ScrollWindow ( CommHwnd, 0, y_char, &rc, NULL ) ;
	}
	else
		ScrollWindow ( CommHwnd, 0, y_char, NULL, NULL ) ;

	for ( y_pos=max_y-1 ; y_pos >0 ; y_pos-- )
	{
		memcpy ( &text_map[y_pos][0], &text_map[y_pos-1][0], MAX_POS_X ) ;
		memcpy ( &attr_map[y_pos][0], &attr_map[y_pos-1][0], MAX_POS_X ) ;
	}
	memset ( text_map, 32, MAX_POS_X ) ;
	memset ( attr_map, ATTR_NONE, MAX_POS_X ) ;

}

// determine start x & y start co-ords of new copy region
void window_access::start_copy_region( int x_pos, int y_pos )
{
	csx = x_pos / x_char ;
	csy = y_pos / y_char ;
	if ( copying_selected )
		clear_copy_region() ;
	copying_active = TRUE ;
}

// routine requests copy region be cleared
void window_access::clear_copy_region ( void )
{
	copying_selected = FALSE ;
	paint_op = REDRAW_ALL ;
	paint_pending = TRUE ;
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

void window_access::end_copy_region ( void )
{
	// clear copy active flag
	copying_active = FALSE ;

	// if text FIFO read/write ptrs do not match
	// then text has come in during the copying and been held off till now
	// so clear the region and display the text
	if ( read_ptr != write_ptr )
		clear_copy_region() ;
	else
		copying_selected = TRUE ;
}

// determine region selected
void window_access::mark_copy_region ( int x_pos, int y_pos )
{
	// determine new position of cursor
	x_pos = x_pos / x_char ;
	y_pos = y_pos / y_char ;

	// save new co-ords
	cex = x_pos ;
	cey = y_pos ;

	// request paint message to command shell window
	InvalidateRect ( CommHwnd, NULL, FALSE ) ;
	paint_op = MARK_COPY ;
	paint_pending = TRUE ;
}

// copy selected region to clipboard
void window_access::copy_to_clipboard ( void )
{
	HGLOBAL mem_handle ;
	char *copy_bfr, *ptr ;
	int px,y, x, space_pos ;

	// if all co-ordinates of selected region are the same
	// then no-region is available for copy
	if ( cex == csx && cey == csy )
		return ;

	if ( OpenClipboard ( CommHwnd ) )
	{
		EmptyClipboard() ;

		// get a block of memory to store the text in
		// including space for file CR/LF and NULL
		mem_handle = GlobalAlloc ( GHND, abs(cey-csy)*MAX_POS_X+abs(cex-csx)+3 ) ;
		copy_bfr = (char*)GlobalLock ( mem_handle )  ;
		ptr = copy_bfr ;

		// px is the top left x-cordinate in the text map to begin copying from
		px=sx ;
		// loop through selected lines
		for(y=sy;y<ey;y++)
		{
			// look for the last non-space char in the text map
			// assumed to be end of copy line
			space_pos=px ;
			for(x=px;x<MAX_POS_X;x++)
			{
				if ( text_map[y][x]!=' ' )
					space_pos=x ;
			}
			// copy valid part of line into copy buffer & add CR/LF
			memcpy ( ptr, &text_map[y][px], space_pos-px+1 ) ;
			ptr+=space_pos-px+1 ;
			*ptr = 0xd ;
			ptr++ ;
			*ptr = 0xa ;
			ptr++ ;
			// reset starting x position to 0 for start of next line
			px=0 ;
		}
		// paste in the last bit of the line, adding CR/LF & terminating null
		if ( ex>0 )
		{
			memcpy ( ptr, &text_map[ey][px], ex ) ;
			ptr+=ex-px ;
			*ptr = 0xd ;
			ptr++ ;
			*ptr = 0xa ;
			ptr++ ;
		}
		*ptr = 0 ;

		// release memory and paste it on the clipboard
		GlobalUnlock ( mem_handle ) ;
		SetClipboardData ( CF_OEMTEXT, mem_handle ) ;
		CloseClipboard() ;
	}
}

// this routine re-draws the entire text window, highlighting
// the selected region
void window_access::paint_copy_region ( HDC hDC )
{
	int x,y ;
	BOOL marking ;
	COLORREF high_fore_colour, high_back_colour ;

	high_fore_colour = ~fore_colour ;
	high_back_colour = ~back_colour ;

	// determine starting & ending points into sx & sy
	// if a single line is highlighted, determine the greater x-coord
	if ( cey == csy )
	{
		if ( cex < csx )
		{
			sx = cex ;
			ex = csx ;
		}
		else
		{
			sx = csx ;
			ex = cex ;
		}
		sy = csy ;
		ey = csy ;
	}
	else
	{
		// if ending line is now less than the starting line
		// top left x must now be the last co-ord
		if ( cey < csy )
		{
			sy = cey ;
			ey = csy ;
			sx = cex ;
			ex = csx ;
		}
		else
		{
			sy = csy ;
			ey = cey ;
			sx = csx ;
			ex = cex ;
		}
	}

	marking = FALSE ;

	// loop thru the no. of selected lines
	for ( y=0 ; y<=cursor_y ; y++ )
	{
		// if only a single line is marked draw in the normal and
		// selected portions
		if ( y==sy && y==ey )
		{
			draw_attr_text ( hDC, 0, y, sx ) ;
			SetTextColor ( hDC, high_fore_colour ) ;
			SetBkColor ( hDC, high_back_colour ) ;
			TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], ex-sx ) ;
			draw_attr_text ( hDC, ex, y, max_x-ex ) ;
		}
		else
		{
			// if the starting line is reached, highlight from the starting
			// x-coord and leave the highlight pen selected
			if ( y==sy )
			{
				draw_attr_text ( hDC, 0, y, sx ) ;
				SetTextColor ( hDC, high_fore_colour ) ;
				SetBkColor ( hDC, high_back_colour ) ;
				TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], max_x-sx ) ;
				marking = TRUE ;
			}
			else
			{
				// if the ending line is reached, highlight the remaining
				// bit and restore the normal pen
				if ( y==ey )
				{
					TextOut ( hDC, 0, y*y_char, &text_map[y][0], ex ) ;
					draw_attr_text ( hDC, ex, y, max_x-ex ) ;
					marking = FALSE ;
				}
				else
					if ( marking )
						TextOut ( hDC, 0, y*y_char, &text_map[y][0], max_x ) ;
					else
						draw_attr_text ( hDC, 0, y, max_x ) ;
			}
		}
	}
}

// this operation will update the device context, and window class
// variables to perform a specific screen emulation operation defined
// from either commemu.h, go51_emu.h or vt_emu.h
// the return parameter will indicate whether painting can continue
// after this emulation code has been processed
BOOL window_access::process_emulation_code ( HDC hDC, unsigned char op_code )
{
	int count ;
	BOOL return_flag ;
	LOGFONT attr_font ;

	switch ( op_code )
	{
		case Carriage_Return :

			cursor_x = 0 ;
			return FALSE ;

		case Line_Feed :

			if ( cursor_y < max_y-1 )
			{
				cursor_y++ ;
				return FALSE ;
			}
			else
			{
				scroll_down() ;
				return TRUE ;
			}

		case CURSOR_RIGHT :

			count = screen_emulator->new_x ;
			while ( count > 0 )
			{
				if ( cursor_x < max_x )
					cursor_x++ ;
				count-- ;
			}
			return FALSE ;

		case CURSOR_LEFT :

			count = screen_emulator->new_x ;
			while ( count > 0 )
			{
				if ( cursor_x > 0 )
					cursor_x-- ;
				count-- ;
			}
			return FALSE ;

		case CURSOR_UP :

			count = screen_emulator->new_y ;
			while ( count > 0 )
			{
				if ( cursor_y > 0 )
				{
					cursor_y-- ;
					return_flag = FALSE ;
				}
				else
				{
					scroll_up() ;
					return_flag = TRUE ;
				}
				count-- ;
			}
			return return_flag ;

		case CURSOR_DOWN :

			count = screen_emulator->new_y ;
			while ( count > 0 )
			{
				if ( cursor_y < max_y-1 )
				{
					cursor_y++ ;
					return_flag = FALSE ;
				}
				else
				{
					scroll_down() ;
					return_flag = TRUE ;
				}
				count-- ;
			}
			return return_flag ;

		case REVERSE_ON :

			display_attr = display_attr | ATTR_REV ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case REVERSE_OFF :

			display_attr = display_attr & (~ATTR_REV) ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case Bell :

			return FALSE ;

		case Backspace :

			if ( cursor_x == 0 )
			{
				if ( cursor_y != 0 )
				{
					cursor_y-- ;
					cursor_x = max_x-1 ;
				}
			}
			else
			{
					cursor_x-- ;
			}
			return FALSE ;

		case Tab :

			cursor_x = screen_emulator->find_next_tab ( cursor_x ) ;
			return FALSE ;

		case Home :

			cursor_x = 0 ;
			cursor_y = 0 ;
			return FALSE ;

		case Clear :

			cursor_x = 0 ;
			cursor_y = 0 ;

		case CLEAR_AOS :

			memset ( text_map, VK_SPACE, MAX_POS_X*MAX_POS_Y ) ;
			memset ( attr_map, ATTR_NONE, MAX_POS_X*MAX_POS_Y ) ;
			paint_op = REDRAW_ALL ;
			paint_pending = TRUE ;
			InvalidateRect ( CommHwnd, NULL, TRUE ) ;
			return TRUE ;

		case UNDERLINE_ON :

			display_attr = display_attr | ATTR_ULINE ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case UNDERLINE_OFF :

			display_attr = display_attr & (~ATTR_ULINE) ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case CLEAR_EOS :

			memset ( &text_map[cursor_y][cursor_x], VK_SPACE,
				MAX_POS_Y*MAX_POS_X-(cursor_y*MAX_POS_X+cursor_x) ) ;
			memset ( &attr_map[cursor_y][cursor_x], ATTR_NONE,
				MAX_POS_Y*MAX_POS_X-(cursor_y*MAX_POS_X+cursor_x) ) ;
			paint_op = REDRAW_ALL ;
			paint_pending = TRUE ;
			return TRUE ;

		case CLEAR_EOL :

			memset ( &text_map[cursor_y][cursor_x], 32, MAX_POS_X - cursor_x ) ;
			memset ( &attr_map[cursor_y][cursor_x], ATTR_NONE, MAX_POS_X - cursor_x ) ;
			paint_op = REDRAW_ALL ;
			paint_pending = TRUE ;
			return TRUE ;

		case SET_CURSOR :

			cursor_x = screen_emulator->new_x ;
			cursor_y = screen_emulator->new_y ;
			return FALSE ;

		case SCROLL_UP :

			scroll_up() ;
			return TRUE ;

		case SCROLL_DOWN :

			scroll_down() ;
			return TRUE ;

		case SAVE_ATTRS :

			attr_save_x = cursor_x ;
			attr_save_y = cursor_y ;
			saved_attr = display_attr ;
			return FALSE ;

		case REST_ATTRS :

			cursor_x = attr_save_x ;
			cursor_y = attr_save_y ;
			display_attr = saved_attr ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case SET_TAB :

			screen_emulator->modify_tab ( cursor_x, TRUE ) ;
			return FALSE ;

		case RESET_TERM :

			return FALSE ;

		case NO_ATTRS :

			display_attr = ATTR_NONE ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case BOLD_ON :

			display_attr = display_attr | ATTR_BOLD ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case BOLD_OFF :

			display_attr = display_attr & (~ATTR_BOLD) ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case BLINK_ON :

			display_attr = display_attr | ATTR_BLINK ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case INVISIBLE_ON :

			display_attr = display_attr | ATTR_INVIS ;
			set_display_attr ( hDC, display_attr ) ;
			return FALSE ;

		case WIDTH_132 :

			if ( display_options.fixed_size )
			{
				display_options.max_x = 132 ;
				if ( max_x > display_options.max_x )
					max_x = display_options.max_x ;
			}
			return FALSE ;

		case WIDTH_80 :

			if ( display_options.fixed_size )
			{
				display_options.max_x = 80 ;
				if ( max_x > display_options.max_x )
					max_x = display_options.max_x ;
			}
			return FALSE ;

		case AUTO_WRAP :

			display_options.wrap_flag = TRUE ;
			return FALSE ;

		case RESET_WRAP :

			display_options.wrap_flag = FALSE ;
			return FALSE ;

		case CLEAR_TAB :

			screen_emulator->modify_tab ( cursor_x, FALSE ) ;
			return FALSE ;

		case CLEAR_LEFT :

			memset ( &text_map[cursor_y][0], VK_SPACE, cursor_x ) ;
			memset ( &attr_map[cursor_y][0], ATTR_NONE, cursor_x ) ;
			paint_op = REDRAW_ALL ;
			paint_pending = TRUE ;
			return TRUE ;

		case CLEAR_TOS :

			memset ( text_map, VK_SPACE, cursor_y*MAX_POS_X+cursor_x ) ;
			memset ( attr_map, ATTR_NONE, cursor_y*MAX_POS_X+cursor_x ) ;
			paint_op = REDRAW_ALL ;
			paint_pending = TRUE ;
			return TRUE ;

		case CLEAR_LINE :

			memset ( &text_map[cursor_y][0], VK_SPACE, MAX_POS_X ) ;
			memset ( &attr_map[cursor_y][0], ATTR_NONE, MAX_POS_X ) ;
			paint_op = REDRAW_ALL ;
			paint_pending = TRUE ;
			return TRUE ;

		default :
			return FALSE ;
	}
}

// sets the display attributes according to the supplied attribute code
void window_access::set_display_attr ( HDC hDC, unsigned char attr_code )
{
	BOOL new_font = FALSE ;
	COLORREF new_fore_colour, new_back_colour ;

	// determine text colours based on reverse attribute setting
	if ( ( attr_code & ATTR_REV ) == ATTR_REV )
	{
		new_fore_colour = fore_colour ;
		new_back_colour = back_colour ;
	}
	else
	{
		new_back_colour = fore_colour ;
		new_fore_colour = back_colour ;
	}

	// select a colour for blinking text
	if ( attr_code & ATTR_BLINK )
		new_fore_colour = RGB ( 255, 128, 128 ) ;

	// if invisible attribute set, make foreground colour same as bkgnd
	if ( attr_code & ATTR_INVIS )
		new_fore_colour = new_back_colour ;

	// setup the screen colours
	SetBkColor ( hDC, new_fore_colour ) ;
	SetTextColor ( hDC, new_back_colour ) ;

	// check for bold test.
	if ( ( attr_code & ATTR_BOLD ) == ATTR_BOLD )
	{
		// if bold text not already selected, set attr in font structure
		if ( font_definition.lfWeight != FW_BOLD )
		{
			font_definition.lfWeight = FW_BOLD ;
			new_font = TRUE ;
		}
	}
	else
		// if bold selected in font, it is no longer reqd
		if ( font_definition.lfWeight == FW_BOLD )
		{
			font_definition.lfWeight = FW_NORMAL ;
			new_font = TRUE ;
		}

	// perform similar checks for underline text
	if ( ( attr_code & ATTR_ULINE ) == ATTR_ULINE )
	{
		if ( !font_definition.lfUnderline )
		{
			font_definition.lfUnderline = TRUE ;
			new_font = TRUE ;
		}
	}
	else
		if ( font_definition.lfUnderline )
		{
			font_definition.lfUnderline = FALSE ;
			new_font = TRUE ;
		}

	// if a new font is deemed necessary
	if ( new_font )
	{
		// de-select current font from context
		SelectObject ( hDC, holdsfont ) ;

		// delete font and create new one
		DeleteObject(hnewsfont) ;
		hnewsfont = CreateFontIndirect((LPLOGFONT) &font_definition );

		// select in new font
		holdsfont=SelectObject(hDC, hnewsfont);
	}
}

// set display options for window size and terminal emulation mode
void window_access::set_display_options ( void )
{
	RECT rc ;

	if ( display_options.fixed_size )
	{
		// range check the display size
		if ( display_options.max_x > MAX_POS_X )
			display_options.max_x = MAX_POS_X ;
		if ( display_options.max_y > MAX_POS_Y )
			display_options.max_y = MAX_POS_Y ;

		// clip display extents to those specified
		if ( max_x > display_options.max_x )
			max_x = display_options.max_x ;
		if ( max_y > display_options.max_y )
			max_y = display_options.max_y ;
	}
	else
	{
		// get size of window and set display extents to it
		GetClientRect ( CommHwnd, &rc ) ;
		max_y = ( rc.bottom - rc.top ) / y_char ;
		max_x = ( rc.right - rc.left ) / x_char ;
	}

	if ( screen_emulator != NULL )
		delete screen_emulator ;

	// activate the required screen emulation module
	switch ( display_options.emulation )
	{
		case NO_EMULATION :

			screen_emulator = new common_emulator ;
			break ;

		case GO51_EMUL :

			screen_emulator = new go51_emulator ;
			break ;

		case VT100_EMUL :

			screen_emulator = new vt100_emulator ;
			break ;
	}

	// force re-paint to take into account new changes
	paint_pending = TRUE ;
	paint_op = REDRAW_ALL ;
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

// returns a pointer to either paste buffer or keyboard buffer
// and byte count (up to 255 bytes) of data to be passed to remote
char *window_access::get_paste_buffer ( unsigned *buffer_size )
{
	char *bfr_ptr ;
	const unsigned max_paste_return = 255 ;

	// if paste bfr exists
	if ( paste_buffer != NULL )
	{
		// if paste bfr >255 bytes, return just a 255 byte chunk
		// and update ptrs
		if ( paste_bfr_size > max_paste_return )
		{
			paste_bfr_size-=max_paste_return ;
			*buffer_size = max_paste_return ;
			bfr_ptr = paste_ptr ;
			paste_ptr+=max_paste_return ;
			return bfr_ptr ;
		}

		// if buffer empty free buffer
		if ( paste_bfr_size == 0 )
		{
			free ( paste_buffer ) ;
			paste_buffer = NULL ;
		}
		else
		{
			// return remaining chunk in paste bfr
			*buffer_size=paste_bfr_size ;
			bfr_ptr = paste_ptr ;
			paste_bfr_size = 0 ;
			return bfr_ptr ;
		}
	}

	// no paste buffer available, so check if anything in keyboard bfr
	if ( keyb_write != 0 )
	{
		 *buffer_size = keyb_write ;
		 keyb_write = 0 ;
		 return keyb_buffer ;
	}

	// no data is available
	*buffer_size = 0 ;
	return NULL ;
}

// copy the windows clipboard into the local paste buffer - up to 64K
void window_access::copy_clipboard_to_paste_buffer ( void )
{
	unsigned long i ;
	const unsigned long max_clipboard_size = 65535 ;
	char *clip_ptr, *clipbrd_start ;
	HANDLE clipboard_handle ;

	paste_bfr_size = 0 ;

	// open and get handle to clipboard data
	if ( !OpenClipboard ( CommHwnd ) )
		return ;
	clipboard_handle = GetClipboardData ( CF_TEXT ) ;
	if ( clipboard_handle == NULL )
		return ;
	clipbrd_start = (unsigned char*)GlobalLock ( clipboard_handle ) ;
	clip_ptr = clipbrd_start ;

	// look for terminating null at end of clipboard
	for (i=0 ; i<max_clipboard_size ; i++ )
	{
		if ( *clip_ptr == 0 )
		{
			paste_bfr_size = i ;
			i=max_clipboard_size ;
		}
		clip_ptr++ ;
	}

	if ( paste_bfr_size != 0 )
	{
		// grab paste buffer, deleting old one first if necessary
		if ( paste_buffer != NULL )
			free ( paste_buffer ) ;
		paste_buffer = (char*)malloc ( paste_bfr_size ) ;
		// copy clipboard text to local bfr
		memcpy ( paste_buffer, clipbrd_start, paste_bfr_size ) ;
		// working data pointer
		paste_ptr = paste_buffer ;
	}
	// release clipboard
	GlobalUnlock ( clipboard_handle ) ;
	CloseClipboard() ;
}

// copy the passed keypress into the keyboard buffer
void window_access::copy_keypress_to_paste_buffer ( char keypress, BOOL special )
{
	char keycode ;

	// if special keycode passed in check for a mapping
	// in our keymap
	if ( special )
	{
		keycode = keymap.keys[keystate][vtkey_to_udfkey[keypress]] ;
		// if no match found no further processing needed
		if ( keycode == 0 )
			return ;

		// set flag indicating that possibly the next WM_CHAR message
		// is to be discarded
		keypress_discard = TRUE ;

		// discard character if insufficient room
		if ( keyb_write == 256 )
			return ;

		// otherwise add character to keyboard buffer
		keyb_buffer[keyb_write] = keycode ;
		keyb_write++ ;
	}
	else
	{
		// if flag set to discard this char, clear flag and exit
		if ( keypress_discard )
		{
			keypress_discard = FALSE ;
			return ;
		}
		// discard character if insufficient room
		if ( keyb_write == 256 )
			return ;
		// add normal char to buffer
		keyb_buffer[keyb_write] = keypress ;
		keyb_write++ ;
	}
}

void window_access::new_display_font( void )
{
	HDC hDC ;
	TEXTMETRIC tm ;
	RECT rc ;

	// create new user font
	get_new_font ( CommHwnd, &font_definition, &fore_colour ) ;
	DeleteObject(hnewsfont) ;
	hnewsfont = CreateFontIndirect((LPLOGFONT) &font_definition );

	// get new text sizes
	hDC = GetDC ( CommHwnd ) ;
	holdsfont=SelectObject(hDC, hnewsfont);
	GetTextMetrics(hDC, &tm);
	x_char = tm.tmAveCharWidth ;
	y_char = tm.tmHeight + tm.tmExternalLeading;
	SelectObject ( hDC, holdsfont ) ;
	ReleaseDC ( CommHwnd, hDC ) ;
	// re-define window in terms of new font
	GetClientRect ( CommHwnd, &rc ) ;
	resize_window ( rc.right, rc.bottom ) ;
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

// command shell window
// this handles all the windows messages sent to PC-Share command shell
BOOL CALLBACK _export EditProc (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
window_access *user_window ;
FILE *key_file ;
PAINTSTRUCT ps;

	switch ( Message )
	{
		case WM_INITDIALOG :

			return 1 ;

		case WM_COMMAND :

			// menu commands
			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case CM_EDITCOPY :

					// if copy region selected, copy it to the clipboard
					user_window = find_owner_class ( hDlg ) ;
					if ( user_window->copying_selected )
						user_window->copy_to_clipboard() ;
					break ;

				case CM_EDITPASTE :

					// transfer clipboard into local paste buffer
					find_owner_class ( hDlg )->copy_clipboard_to_paste_buffer() ;
					break ;

				case AUTO_REPAINT :

					// special message caused within paint operations
					// to force paint message
					InvalidateRect ( hDlg, NULL, FALSE ) ;
					break ;

				case DISPLAY_OPTIONS :

					user_window = find_owner_class ( hDlg ) ;
					// activate dialog box to get new display options
					DialogBoxParam ( hInst, MAKEINTRESOURCE ( DISPLAY_DIALOG ), hDlg,
						(DLGPROC)DisplayOptions, (LPARAM)&user_window->display_options ) ;
					user_window->set_display_options() ;
					break ;

				case DISPLAY_FONT :

					// get new font and update current display
					find_owner_class ( hDlg )->new_display_font() ;
					break ;

				case DISPLAY_FORECOLOUR :

					// get new foreground colour and re-draw display
					user_window = find_owner_class ( hDlg ) ;
					get_new_colour ( hDlg, &user_window->fore_colour ) ;
					InvalidateRect ( hDlg, NULL, TRUE ) ;
					break ;

				case DISPLAY_BACKCOLOUR :

					// get new background colour and re-draw display
					user_window = find_owner_class ( hDlg ) ;
					get_new_colour ( hDlg, &user_window->back_colour ) ;
					InvalidateRect ( hDlg, NULL, TRUE ) ;
					break ;

				case DISPLAY_KEYS :

					// define new keymap for this window
					user_window = find_owner_class ( hDlg ) ;
					DialogBoxParam ( hInst, MAKEINTRESOURCE ( KEYMAPS ), Main_hWnd,
						(DLGPROC)DefineKeyMap, (LPARAM)&user_window->keymap ) ;
					break ;

				case SAVE_KEYMAP :

					// save keymap to disk file
					user_window = find_owner_class ( hDlg ) ;
					key_file = OpenSaveFile ( hDlg, "keys", "map", "wb" ) ;
					if ( key_file != NULL )
					{
						fwrite ( &user_window->keymap, sizeof ( struct keymap_type ), 1, key_file ) ;
						fclose ( key_file ) ;
					}
					break ;

				case LOAD_KEYMAP :

					// load keymap from disk file
					user_window = find_owner_class ( hDlg ) ;
					key_file = OpenLoadFile ( hDlg, "keys", "map", "rb" ) ;
					if ( key_file != NULL )
					{
						fread ( &user_window->keymap, sizeof ( struct keymap_type ), 1, key_file ) ;
						fclose ( key_file ) ;
					}
					break ;
			}
			return 0 ;

		case WM_LBUTTONDOWN :

			// if mouse button is pressed, notify the system a copy region is
			// about to be selected
			user_window = find_owner_class ( hDlg ) ;
			user_window->start_copy_region ( LOWORD ( lParam ), HIWORD ( lParam ) ) ;
			return 0 ;

		case WM_LBUTTONUP :

			// if button released, set update state flags
			user_window = find_owner_class ( hDlg ) ;
			if ( user_window->copying_active )
				user_window->end_copy_region() ;
			return 0 ;

		case WM_MOUSEMOVE :

			// if mouse is moved whilst button is depressed, highlight the selected
			// region
			user_window = find_owner_class ( hDlg ) ;
			if ( user_window->copying_active && ( ( wParam & MK_LBUTTON ) == MK_LBUTTON ) )
				user_window->mark_copy_region ( LOWORD ( lParam ), HIWORD ( lParam ) ) ;
			return 0 ;

		case WM_SIZE :

			// on a re-size, get the new co-ordinates of the window
			// and indicate the arriving paint message has been caused by a re-size
			user_window = find_owner_class( hDlg ) ;
			if ( ( user_window != NULL ) && ( wParam != SIZE_MINIMIZED ) )
				user_window->resize_window ( LOWORD ( lParam ), HIWORD ( lParam ) ) ;
			return 0 ;

		case WM_PAINT :

			// call the window class paint handler
			BeginPaint ( hDlg, &ps ) ;
			user_window = find_owner_class( hDlg ) ;
			if ( user_window != NULL )
				user_window->paint_window ( ps.hdc ) ;
			EndPaint ( hDlg, &ps ) ;
			return 0 ;

		case WM_ERASEBKGND :
		{
			HBRUSH back_brush ;
			RECT rc ;
			HDC hdc = (HDC)wParam ;

			// erase background to selected window's colour
         // or system default if it is not yet active
			user_window = find_owner_class( hDlg ) ;
			if ( user_window != NULL )
				back_brush = CreateSolidBrush ( user_window->back_colour ) ;
			else
         	back_brush = CreateSolidBrush ( system_config.back_colour ) ;
			GetClientRect ( hDlg, &rc ) ;
			FillRect ( hdc, &rc, back_brush ) ;
			DeleteObject ( back_brush ) ;
			return 1 ;
		}

		case WM_CHAR :

			// transfer keypress to paste bfr
			find_owner_class ( hDlg )->copy_keypress_to_paste_buffer ( (char)wParam, FALSE ) ;
			return 0 ;

		case WM_KEYDOWN :

			user_window = find_owner_class ( hDlg ) ;
			// if keypress is SHIFT or CTRL, adjust keystate accordingly
			// otherwise pass to window for processing
			switch ( wParam )
			{
				case VK_SHIFT :
					user_window->keystate = SHIFTED ;
					break ;
				case VK_CONTROL :
					user_window->keystate = CONTROL ;
					break ;
				default :
					find_owner_class ( hDlg )->copy_keypress_to_paste_buffer ( (char)wParam, TRUE ) ;
			}
			break ;

		case WM_KEYUP :

			// if shift or ctrl key has been released, modify keyboard state
			// for current window
			// if keypress discard flag has not been cleared, clear it now
			user_window = find_owner_class ( hDlg ) ;
			switch ( wParam )
			{
				case VK_SHIFT :
				case VK_CONTROL :
					user_window->keystate = NORMAL_CHAR ;
					break ;
				default :
					if ( user_window->keypress_discard )
						user_window->keypress_discard = FALSE ;
			}
			break ;

		case WM_SETFOCUS :

			active_hwnd = hDlg ;
			return 0 ;

		case WM_CLOSE :

			// prompt user for confirmation of closing window as it will be closed
			// without the knowledge of the host computer
			if ( MessageBox ( hDlg, "Are you sure you want to close this active window?",
				"Close Active Window", MB_OKCANCEL | MB_ICONQUESTION ) == IDOK )
			{
				int i ;

				// find the window and delete it - don't call destroywindow
				// as the class destructor will do it
				for(i=0;i<MAX_WINDOWS;i++)
				{
					if ( active_windows[i] != NULL )
						if ( active_windows[i]->CommHwnd == hDlg )
						{
							delete active_windows[i] ;
							active_windows[i] = NULL ;
						}
				}
			}
			break ;

		default:
			return 0 ;
	}
	return 0 ;
}

// this operation returns a pointer to the class
// which owns the specified window identity
window_access *find_owner_class( HWND hWnd )
{
	int i ;
	for(i=0;i<MAX_WINDOWS;i++)
	{
		if ( active_windows[i] != NULL )
			if ( active_windows[i]->CommHwnd == hWnd )
				return active_windows[i] ;
	}
	return NULL ;
}

// timer procedure called once a second by windows used
// to flash cursors in all active windows
#pragma argsused
void CALLBACK _export CursorTimer ( HWND hWnd, UINT msg, UINT idTimer, DWORD dwTime )
{
	int i ;

	// issue paint msg to all active windows not currently awaiting painting
	for (i=0 ; i<MAX_WINDOWS;i++)
	{
		if ( active_windows[i] != NULL )
			if ( !active_windows[i]->paint_pending )
			{
				active_windows[i]->paint_pending = TRUE ;
				InvalidateRect ( active_windows[i]->CommHwnd, NULL, FALSE ) ;
			}
	}
}

// operation to delay a short period to enable
// the host to present another data item to it's
// IO port prior to a paint msg being executed. Should prevent
// paint messages from occuring to only draw single characters at a time
void host_port_delay(void)
{
	TIMERINFO ti ;
	DWORD start_time ;
	const DWORD wait_time = 1 ;

	ti.dwSize = sizeof ( TIMERINFO ) ;

	TimerCount ( &ti ) ;
	start_time = ti.dwmsSinceStart ;

	while ( ti.dwmsSinceStart-start_time < wait_time )
		TimerCount ( &ti ) ;
}

// dialog box procedure for entering display options
BOOL CALLBACK _export DisplayOptions (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
	static struct display_options_type *display_options ;
	BOOL tx ;

	switch ( Message )
	{
		case WM_INITDIALOG :

			// get parameter passed to dialog box
			display_options = (struct display_options_type*)lParam ;

			// initialise dialog box options with supplied structure
			switch ( display_options->emulation )
			{
				case NO_EMULATION :

					CheckDlgButton ( hDlg, EMULATION_NONE, 1 ) ;
					break ;

				case VT100_EMUL :

					CheckDlgButton ( hDlg, EMULATION_VT100, 1 ) ;
					break ;

				case GO51_EMUL :

					CheckDlgButton ( hDlg, EMULATION_GO51, 1 ) ;
					break ;
			}

			SetDlgItemInt ( hDlg, X_CHARS, display_options->max_x, FALSE ) ;
			SetDlgItemInt ( hDlg, Y_CHARS, display_options->max_y, FALSE ) ;

			if ( display_options->fixed_size )
				CheckDlgButton ( hDlg, DISPLAY_FIX, 1 ) ;
			else
			{
				CheckDlgButton ( hDlg, DISPLAY_VAR, 1 ) ;
				EnableWindow ( GetDlgItem ( hDlg, X_CHARS ), FALSE ) ;
				EnableWindow ( GetDlgItem ( hDlg, Y_CHARS ), FALSE ) ;
			}
			if ( display_options->wrap_flag )
				CheckDlgButton ( hDlg, TEXT_WRAP, 1 ) ;

			return 1 ;

		case WM_COMMAND :

			// menu commands
			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case DISPLAY_FIX :

					CheckDlgButton ( hDlg, DISPLAY_FIX, 1 ) ;
					CheckDlgButton ( hDlg, DISPLAY_VAR, 0 ) ;
					EnableWindow ( GetDlgItem ( hDlg, X_CHARS ), TRUE ) ;
					EnableWindow ( GetDlgItem ( hDlg, Y_CHARS ), TRUE ) ;
					break ;

				case DISPLAY_VAR :

					CheckDlgButton ( hDlg, DISPLAY_FIX, 0 ) ;
					CheckDlgButton ( hDlg, DISPLAY_VAR, 1 ) ;
					EnableWindow ( GetDlgItem ( hDlg, X_CHARS ), FALSE ) ;
					EnableWindow ( GetDlgItem ( hDlg, Y_CHARS ), FALSE ) ;
					break ;

				case EMULATION_NONE :

					CheckDlgButton ( hDlg, EMULATION_NONE, 1 ) ;
					CheckDlgButton ( hDlg, EMULATION_VT100, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_GO51, 0 ) ;
					break ;

				case EMULATION_VT100 :

					CheckDlgButton ( hDlg, EMULATION_NONE, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_VT100, 1 ) ;
					CheckDlgButton ( hDlg, EMULATION_GO51, 0 ) ;
					break ;

				case EMULATION_GO51 :

					CheckDlgButton ( hDlg, EMULATION_NONE, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_VT100, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_GO51, 1 ) ;
					break ;

				case IDOK :

					// load up return information from dialog box
					if ( IsDlgButtonChecked ( hDlg, EMULATION_NONE ) )
						display_options->emulation = NO_EMULATION ;
					if ( IsDlgButtonChecked ( hDlg, EMULATION_VT100 ) )
						display_options->emulation = VT100_EMUL ;
					if ( IsDlgButtonChecked ( hDlg, EMULATION_GO51 ) )
						display_options->emulation = GO51_EMUL ;

					if ( IsDlgButtonChecked ( hDlg, DISPLAY_FIX ) )
					{
						display_options->fixed_size = TRUE ;
						display_options->max_x = GetDlgItemInt ( hDlg, X_CHARS, &tx, FALSE ) ;
						display_options->max_y = GetDlgItemInt ( hDlg, Y_CHARS, &tx, FALSE ) ;
					}
					else
						display_options->fixed_size = FALSE ;

					if ( IsDlgButtonChecked ( hDlg, TEXT_WRAP ) )
						display_options->wrap_flag = TRUE ;
					else
						display_options->wrap_flag = FALSE ;

					EndDialog ( hDlg, 0 ) ;
					break ;
			}
			return 0 ;
	}
	return 0 ;
}






















































