/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PACKAGE VIDEO_CTRL
CREATED ON February 12, 1995, 04:27 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON February 19, 1996, FOR V2.1 BY J.Bird
MOD ON October 06, 1996, FOR V3.0 BY J.Bird
MOD ON October 29 1997, FOR V3.02 BY J.Bird
MOD ON November 22 1997, FOR V3.1 BY J.Bird
MOD ON March 03 1998, FOR V3.11 BY J.Bird
MOD ON March 13 1998, FOR V4.0 BY J.Bird

Package Video_Ctrl is

DESCRIPTION

The Video_Ctrl package is responsible for handling the video related functions
that are requested by the external computer. These will involve textual data
to be displayed, and graphics operations to be executed.

The package comprises 2 internal objects:

-- Graphics for handling all of the graphical operations.
-- VT100_Emul for performing the DEC VT100 terminal emulation.

The majority of the video processing is handled by these two objects, the
Video_Ctrl object primarily serving the purpose of a single interface to the
outside world.

REQUIRED INTERFACE
 GRAPHICS
 VT100_EMUL
*/
#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include "pcshare.h"
#include "config.h"
#include "commemu.h"
#include "graphics.h"
#include "video_ct.h"
#include "go51_emu.h"
#include "vt100_em.h"
#include "fileio.h"

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif
/*
Types/Data
*/

// paint operations
#define MARK_COPY 1
#define REDRAW_ALL 2

// text attributes
#define ATTR_NONE 0x0
#define ATTR_REV 0x1
#define ATTR_BOLD 0x2
#define ATTR_ULINE 0x4
#define ATTR_BLINK 0x8
#define ATTR_INVIS 0x10

// active display window classes
window_access *active_windows[MAX_WINDOWS] = { NULL } ;
// count of windows active
static int modules_active=0 ;
// handle for cursor timer
UINT cursor_timer ;
// current active window
HWND active_hwnd ;
const char *winname = "DisplayWin" ;
const char *editname = "EditWin" ;

window_access *find_owner_class( HWND hWnd ) ;
void CALLBACK _export CursorTimer ( HWND, UINT, UINT, DWORD ) ;
BOOL CALLBACK _export DisplayOptions (HWND, UINT, WPARAM, LPARAM);
void clean_cmd_line ( char *cmdstring ) ;

/*
Provides Operations
 OUTPUT_TEXT
 GRAPHICS_MODE_CHANGE
 GRAPHICS_CURSOR_UPDATE
 GRAPHICS_SET_POINT
 GRAPHICS_DRAW_LINE
 GRAPHICS_DRAW_CIRCLE
 GRAPHICS_READ_PIXEL
 GRAPHICS_COLOUR_CHANGE
 GRAPHICS_CLEAR_SCREEN
*/

/*
Internal Operations
 None

END.
*/
/* clean up a command string prior to putting it in the
   command history listbox. currently just looks for and strips
   out backspace character. */
void clean_cmd_line ( char *cmdstring )
{
   char localcmd[CMD_BUF_SIZE] ;
	 char *sptr = cmdstring ;
   char *dptr = localcmd ;
   int i ;

   for ( i=0 ; i<strlen(cmdstring) ; i++ )
   {
		if ( *sptr == system_config.del_key )
      	dptr-- ;
    else
      	*dptr++ = *sptr ;
      sptr++ ;
   }
   *dptr = 0 ;
   strcpy ( cmdstring, localcmd ) ;
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION OUTPUT_TEXT
CREATED ON February 12, 1995, 04:27 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON February 19, 1996, FOR V2.1 BY J.Bird
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON March 13 1998, FOR V4.0 BY J.Bird

Procedure Output_Text ( Character ) is
*/
void window_access::output_text ( char textcr )
/*
DESCRIPTION

This operation is responsible for displaying a text character on the screen.
It uses the VT100_Emul object to determine:

-- if an escape sequence is in progress, then the character is passed to the
object for processing.

-- if the character signifies the start of an escape sequence, then the
chracter is passed to the object for processing.

-- otherwise it is displayed at the current cursor position.

USES OPERATIONS
 VT100_EMUL.VT_DETECT
 VT100_EMUL.VT_IN_PROGRESS
 VT100_EMUL.PROCESS_VT_SEQUENCE

Types/Data
 None

PSUEDO CODE

 if VT100_Emul.VT_In_Progress
 then

	  VT100_Emul.Process_VT_Sequence ( Character ) ;

 else if VT100_Emul.VT_Detect ( Character )
 then

	  VT100_Emul.Process_VT_Sequence ( Character ) ;

 else

	  Print ( Character ) ;

 end if ;

CODE
*/
{
	unsigned char op_code ;
	BOOL char_processed ;
	int wrap_x ;

	// clear cursor if it's drawn
	if ( cursor_state )
	{
		draw_cursor ( FALSE ) ;
		cursor_state = FALSE ;
	}

	// if copy region is currently selected, remove it
	if ( copying_selected )
		clear_copy_region() ;

	// if window has been scrolled, restore it
	if ( scroll_base_y != window_base_y )
	{
		remap_window ( window_base_y ) ;
		scroll_base_y = window_base_y ;
		SetScrollPos ( CommHwnd, SB_VERT, MAX_SCROLL_Y_POS, TRUE );
		scroll_pos = MAX_SCROLL_Y_POS ;
		InvalidateRect ( CommHwnd, NULL, TRUE ) ;
		UpdateWindow ( CommHwnd ) ;
	}

	// dump to capture file if open
	if ( capture_file != NULL )
		fputc ( textcr, capture_file ) ;

	// check command prompt match if end of line detected
	if ( textcr == '\r' || textcr == '\n' )
	{
		// correctly terminate string and reset ptr for next line
		cmd_line[cmd_pos] = 0 ;
		cmd_pos = 0 ;
		// do the check for command prompt against command line
		if ( strstr ( cmd_line, system_config.cmd_prompt ) != NULL )
		{
      	// if found, first strip out control chars to get a 'pure'
         // ascii variant
			clean_cmd_line ( cmd_line ) ;
         // if not just a null cmd line left...
			if ( strlen ( system_config.cmd_prompt ) != strlen ( cmd_line ) )
			{
				// if found, add to command history list
				SendDlgItemMessage ( ToolWnd, TOOL_HISTORY, CB_INSERTSTRING,
					  0, (LPARAM)((LPSTR)&cmd_line[strlen(system_config.cmd_prompt)] ) ) ;
				SendDlgItemMessage ( ToolWnd, TOOL_HISTORY, CB_SETCURSEL, 0, 0 ) ;
			}
		}
	}
	else
	{
		// not end of line, save char in cmd_line for later
		cmd_line[cmd_pos] = textcr ;
		cmd_pos++ ;
		// if reached end of cmd buf, discard and restart
		if ( cmd_pos == CMD_BUF_SIZE )
			cmd_pos = 0 ;
	}

	char_processed = FALSE ;
	// check for screen emulation sequence in progress
	if ( screen_emulator->emulation_in_progress() )
	{
		// pass the incoming text character for processing & set flag to indicate done
		op_code = screen_emulator->process_emulation_sequence ( textcr ) ;
		char_processed = TRUE ;
	}
	else
	{
		// see if this character is the start of an emulation sequence
		if ( screen_emulator->emulation_detect ( textcr ) )
		{
			// if so, pass to emulator for processing & set flag
			op_code = screen_emulator->process_emulation_sequence ( textcr ) ;
			char_processed = TRUE ;
		}
		else
		{
			// set possible text wrapping point
			if ( display_options.wrap_flag )
				wrap_x = max_x ;
			else
				wrap_x = MAX_POS_X ;

			// if position is less than wrap point, add to string
			if ( cursor_x < wrap_x )
			{
				TextOut ( hDC, cursor_x*x_char, cursor_y*y_char, &textcr, 1 ) ;
				text_map[cursor_y][cursor_x] = textcr ;
				attr_map[cursor_y][cursor_x] = display_attr ;
				cursor_x++ ;
			}

			// if required to wrap text & at wrapping point
			if ( display_options.wrap_flag && cursor_x == wrap_x )
			{
				// now simulate a CR/LF combination
				cursor_x = 0 ;
				char_processed = TRUE ;
				op_code = Line_Feed ;
			}
			// note: if text wrapping is not supported, further text is now ignored
			// until the cursor_x position falls into window range
		}
	}

	// if text character has been processed by a screen emulator
	// and action is required, process the operation code
	if ( char_processed )
		if ( op_code != NO_ACTION )
			process_emulation_code ( op_code ) ;

	// process any 'chained' escape sequences
	do
	{
		op_code = screen_emulator->emulation_sequence_pending() ;
		if ( op_code != NO_ACTION )
			process_emulation_code ( op_code ) ;
	} while ( op_code != NO_ACTION ) ;
}
/*
END.
*/

// set emulation defaults
void window_access::select_video_emulation ( unsigned char emulation_code )
{
	display_options.emulation = emulation_code ;

	switch ( emulation_code )
	{
		case VT100_EMUL :

			// for VT100 displays, set a fixed 80*25 display
			display_options.max_x = 80 ;
			display_options.max_y = 25 ;
			display_options.wrap_flag = TRUE ;
			display_options.fixed_size = TRUE ;
			break ;

		case GO51_EMUL :

			// for GO51 mode set a fixed 80*24 display
			display_options.max_x = 80 ;
			display_options.max_y = 24 ;
			display_options.wrap_flag = TRUE ;
			display_options.fixed_size = TRUE ;
			break ;
	}
	set_display_options() ;
}

// set text display attributes
void window_access::set_text_attr ( int fore_col, int back_col )
{
	fore_colour = fpalette[fore_col] ;
	back_colour = fpalette[back_col] ;
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_MODE_CHANGE
CREATED ON February 12, 1995, 04:27 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Graphics_Mode_Change ( Mode_Id ) is
*/
void window_access::graphics_set_mode ( unsigned x_pixels, unsigned y_pixels, float x_scale, float y_scale )
/*
DESCRIPTION

This operation initiates a Graphics mode change by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.MODE_CHANGE

Types/Data
 None

PSUEDO CODE

 Graphics.Mode_Change ( Mode_Id ) ;

CODE
*/
{

	// if graphic module already active, delete it
	if ( graphic_win )
		delete graphic_win ;

	// if x_pix & y_pix both spec. as 0 means close graphics down
	// otherwise launch the new module
	if ( x_pixels == 0 && y_pixels == 0 )
		graphic_win = NULL ;
	else
		graphic_win = new graphics_support ( hDC, x_pixels, y_pixels, x_scale, y_scale );
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_CURSOR_UPDATE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird

Procedure Graphics_Cursor_Update ( X_Coord, Y_Coord ) is
*/
void window_access::graphics_cursor_update ( unsigned x_coord, unsigned y_coord )
/*
DESCRIPTION

This operation initiates a Graphics cursor update by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.CURSOR_UPDATE

Types/Data
 None

PSUEDO CODE

 Graphics.Cursor_Update ( X_Coord, Y_Coord ) is

CODE
*/
{
	if ( graphic_win  )
		graphic_win->cursor_update ( x_coord, y_coord ) ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_SET_POINT
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Graphics_Set_Point ( X_Coord, Y_Coord, Colour ) is
*/
void window_access::graphics_set_point ( unsigned x_coord, unsigned y_coord, unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics set point by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.SET_POINT

Types/Data
 None

PSUEDO CODE

 Graphics.Set_Point ( X_Coord, Y_Coord, Colour ) ;

CODE
*/
{
	if ( graphic_win  )
	{
		graphic_win->set_point ( x_coord, y_coord, colour ) ;
		graphic_event = TRUE ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_DRAW_LINE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Graphics_Draw_Line ( X1_Coord, Y1_Coord, X2_Coord, Y2_Coord, Colour
) is
*/
void window_access::graphics_draw_line ( unsigned x1_coord,
                                         unsigned y1_coord,
                                         unsigned x2_coord,
                                         unsigned y2_coord,
                                         unsigned char colour,
                                         unsigned char width,
                                         unsigned char style )
/*
DESCRIPTION

This operation initiates a Graphics draw line by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.DRAW_LINE

Types/Data
 None

PSUEDO CODE

 Graphics.Draw_Line ( X1_Coord, Y1_Coord, X2_Coord, Y2_Coord, Colour ) ;

CODE
*/
{
	if ( graphic_win )
	{
		graphic_win->draw_line ( x1_coord,
                             y1_coord,
                             x2_coord,
                             y2_coord,
                             colour,
                             width,
                             style ) ;
		graphic_event = TRUE ;
	}
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_DRAW_CIRCLE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Graphics_Draw_Circle ( X_Coord, Y_Coord, Radius, Colour ) is
*/
void window_access::graphics_draw_circle ( unsigned x_coord,
                                           unsigned y_coord,
                                           unsigned radius,
                                           unsigned char colour,
                                           unsigned char width,
                                           unsigned char style )
/*
DESCRIPTION

This operation initiates a Graphics draw circle by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.DRAW_CIRCLE

Types/Data
 None

PSUEDO CODE

 Graphics.Draw_Circle ( X_Coord, Y_Coord, Radius, Colour ) ;

CODE
*/
{
	 if ( graphic_win  )
	 {
		graphic_win->draw_circle ( x_coord, y_coord, radius, colour, width, style ) ;
		graphic_event = TRUE ;
	 }
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_READ_PIXEL
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Function Graphics_Read_Pixel ( X_Coord, Y_Coord ) return colour is
*/
unsigned char window_access::graphics_read_pixel ( unsigned x_coord, unsigned y_coord )
/*
DESCRIPTION

This operation initiates a Graphics read pixel by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.READ_PIXEL

Types/Data
 None

PSUEDO CODE

 return Graphics.Read_Pixel ( X_Coord, Y_Coord ) ;

CODE
*/
{
	if ( graphic_win )
		return ( graphic_win->read_pixel ( x_coord, y_coord ) ) ;
	else
		return 0 ;
}
/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_COLOUR_CHANGE
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Graphics_Colour_Change ( Colour ) is
*/
void window_access::graphics_colour_change ( unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics colour change by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.COLOUR_CHANGE

Types/Data
 None

PSUEDO CODE

 Graphics.Colour_Change ( Colour ) ;

CODE
*/
{
	 if ( graphic_win )
		graphic_win->colour_change ( colour ) ;
}

/*
END.
*/

/*
HNC INDIVIDUAL PROJECT PC REMOTE SHARER SYSTEM
PROVIDED OPERATION GRAPHICS_CLEAR_SCREEN
CREATED ON February 12, 1995, 04:28 PM BY J.BIRD
MOD ON September 22, 1995, FOR V2.01 BY J.BIRD
MOD ON November 03, 1996 FOR V3.0 BY J.Bird
MOD ON August 07 1999, FOR V4.00 BY J.Bird

Procedure Graphics_Clear_Screen ( Colour ) is
*/
void window_access::graphics_clear_screen ( unsigned char colour )
/*
DESCRIPTION

This operation initiates a Graphics clear screen by requesting the Graphics
object to perform the change.

USES OPERATIONS
 GRAPHICS.CLEAR_SCREEN

Types/Data
 None

PSUEDO CODE

 Graphics.Clear_Screen ( Colour ) ;

CODE
*/
{
	 if ( graphic_win )
	 {
		graphic_win->clear_window ( colour ) ;
		graphic_event = TRUE ;
	 }
}
/*
END.
*/

// new in v4.0
void window_access::graphics_define_palette_entry ( int pal_entry,
		unsigned char red, unsigned char green, unsigned char blue )
{
	 if ( graphic_win  )
		graphic_win->define_palette_entry ( pal_entry, red, green, blue ) ;
}

// initialise window
window_access::window_access( char *window_name, int ident )
{
TEXTMETRIC tm ;
RECT rc ;
int i,j ;
WNDCLASS wndClass ;
RECT ToolRC, MainRC ;

	hnewsfont = CreateFontIndirect((LPLOGFONT) &system_config.font );
	font_definition = system_config.font ;

	// initialise variables and flags
	cursor_x=0 ;
	cursor_y=0 ;
	x_char = 0 ;
	y_char = 0 ;
	copying_active = FALSE ;
	copying_selected = FALSE ;
	keyb_write = 0 ;
	paste_buffer = NULL ;
	keystate = NORMAL_CHAR ;
	keymap = system_config.keymap ;
	keypress_discard = FALSE ;
	screen_emulator = NULL ;
	capture_file = NULL ;
	import_file = NULL ;
	window_base_y = 0 ;
	scroll_base_y = 0 ;
	scroll_pos = MAX_SCROLL_Y_POS ;
	cmd_pos = 0 ;
  GScroll = FALSE ;

	/* Text/Attr Regions & Maps:

		The text map stores the character contents of a display window much
		like memory mapped video RAM. The attr map stores the attribute (colours,
		underline etc.) settings for each character in the text map.

		The text/attr regions cover the maximum scrollable area which can be
		obtained by the window's scroll bar. At any one time, the text/attr maps
		are mapped somewhere into the text/attr region. As they are srolled up/down
		they move throughout the regions. When the scrollbar is used, the map positions
		are simply moved up/down the regions.
	*/

	// allocate the scrollable text & attr window
	// and initilise all areas to blank
	text_region = new char*[MAX_SCROLL_Y] ;
	attr_region = new unsigned char*[MAX_SCROLL_Y] ;
	for ( i=0 ; i < MAX_SCROLL_Y ; i++ )
	{
		text_region[i] = new char[MAX_SCROLL_X] ;
		memset ( text_region[i], VK_SPACE, MAX_SCROLL_X ) ;
		attr_region[i] = new unsigned char[MAX_SCROLL_X] ;
		memset ( attr_region[i], ATTR_NONE, MAX_SCROLL_X ) ;
	}

	// map the display window into the top left of the
	// display region
	text_map = new char*[MAX_POS_Y] ;
	attr_map = new unsigned char*[MAX_POS_Y] ;
	for ( i=0 ; i < MAX_POS_Y ; i++ )
	{
		text_map[i] = text_region[i] ;
		attr_map[i] = attr_region[i] ;
	}

	if ( modules_active == 0 )
	{
		// setup cursor timer
		cursor_timer = SetTimer ( NULL, 0, 500, (TIMERPROC)CursorTimer ) ;

		// register window class for edit window
		// this is given a private device context to minimize
		// the work needed to display each character from the host computer
		wndClass.style          = CS_HREDRAW | CS_VREDRAW | CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW | CS_OWNDC ;
		wndClass.lpfnWndProc    = EditProc ;
		wndClass.cbClsExtra     = 0;
		wndClass.cbWndExtra     = 0;
		wndClass.hInstance      = hInst ;
		wndClass.hIcon          = NULL ;
		wndClass.hCursor        = LoadCursor(NULL, IDC_ARROW);
		wndClass.hbrBackground  = (HBRUSH)GetStockObject ( WHITE_BRUSH ) ;
		wndClass.lpszMenuName   = NULL ;
		wndClass.lpszClassName  = editname ;

		RegisterClass(&wndClass) ;

		// register window class for main window. Edit window now becomes a child
		// of this class
		wndClass.style          = CS_HREDRAW | CS_VREDRAW | CS_BYTEALIGNCLIENT | CS_BYTEALIGNWINDOW ;
		wndClass.lpfnWndProc    = MainWndProc ;
		wndClass.cbClsExtra     = 0;
		wndClass.cbWndExtra     = 0;
		wndClass.hInstance      = hInst ;
		wndClass.hIcon          = LoadIcon ( hInst, MAKEINTRESOURCE ( EDIT_ICON ) ) ;
		wndClass.hCursor        = LoadCursor(NULL, IDC_ARROW);
		wndClass.hbrBackground  = (HBRUSH)GetStockObject ( WHITE_BRUSH ) ;
		wndClass.lpszMenuName   = MAKEINTRESOURCE ( EDIT_MENU ) ;
		wndClass.lpszClassName  = winname ;

		RegisterClass(&wndClass) ;

	}

	// create main window
	MainWnd = CreateWindowEx ( 0, winname, window_name, WS_POPUPWINDOW | WS_CAPTION | WS_OVERLAPPED |
			WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_THICKFRAME | WS_CLIPCHILDREN, 100, 100, 670, 467, NULL, NULL, hInst, NULL ) ;

	ShowWindow(MainWnd, SW_SHOWNORMAL );

   // v4.0 - create toolbar
	ToolWnd = CreateDialog ( hInst, MAKEINTRESOURCE ( EDIT_TOOLBAR ), MainWnd, (DLGPROC)EditToolBar ) ;

	// add it's name to the main menu's window list
	AppendMenu ( Window_hMenu, MF_STRING, WINDOW_BASE_ID+ident, window_name ) ;
	DrawMenuBar ( Main_hWnd ) ;
	window_id = ident ;

	GetClientRect ( MainWnd, &MainRC ) ;
	GetClientRect ( ToolWnd, &ToolRC ) ;
	rc.top = ToolRC.bottom ;
	rc.left = ToolRC.left ;
	rc.right = MainRC.right ;
	rc.bottom = MainRC.bottom-ToolRC.bottom ;

	// create command shell
	CommHwnd = CreateWindowEx ( 0, editname, window_name, WS_CHILD |
			WS_VSCROLL, rc.left, rc.top, rc.right, rc.bottom, MainWnd, 0, hInst, NULL ) ;
	ShowWindow(CommHwnd, SW_SHOWNORMAL );
	SetFocus ( MainWnd ) ;

	// get text sizes and shell dimensions
	hDC = GetDC ( CommHwnd ) ;
	holdsfont=SelectObject(hDC, hnewsfont);

	GetTextMetrics(hDC, &tm);
	x_char = tm.tmAveCharWidth ;
	y_char = tm.tmHeight + tm.tmExternalLeading;
	graphic_win = NULL ;

	// get max size of characters in window
	GetClientRect ( CommHwnd, &rc ) ;
	max_y = ( rc.bottom - rc.top ) / y_char ;
	max_x = ( rc.right - rc.left ) / x_char ;
	display_options.max_x = max_x ;
	display_options.max_y = max_y ;
	display_options.fixed_size = FALSE ;
	display_options.wrap_flag = FALSE ;

	// set video emulation & screen colours to system default
	select_video_emulation ( system_config.emulation ) ;
	fore_colour = system_config.fore_colour ;
	back_colour = system_config.back_colour ;

	cursor_state = FALSE ;
	display_attr = ATTR_NONE ;

	SetScrollRange (CommHwnd, SB_VERT, 0, MAX_SCROLL_Y_POS, TRUE);
	SetScrollPos   (CommHwnd, SB_VERT, MAX_SCROLL_Y_POS, TRUE );

	// increment count of windows active
	modules_active++ ;
}

// close window - by remote
window_access::~window_access()
{
	int i ;

	// release display windows
	delete[] text_map ;
	delete[] attr_map ;
	for ( i=0 ; i < MAX_SCROLL_Y ; i++ )
		delete[] text_region[i] ;
	delete[] text_region ;
	for ( i=0 ; i < MAX_SCROLL_Y ; i++ )
		delete[] attr_region[i] ;
	delete[] attr_region ;

	delete screen_emulator ;

	// release graphic module if active
	if ( graphic_win != NULL )
		delete graphic_win ;

	// release paste buffer if still allocated
	if ( paste_buffer != NULL )
		free ( paste_buffer ) ;

	// kill cursor timer if last module active
	modules_active-- ;

	// restore and release device context
	SelectObject ( hDC, holdsfont ) ;
	ReleaseDC ( CommHwnd, hDC ) ;

	// delete font and destroy windows
	DeleteObject(hnewsfont) ;
	DestroyWindow ( MainWnd ) ;

	// remove window name from window menu
	DeleteMenu ( Window_hMenu, WINDOW_BASE_ID+window_id, MF_BYCOMMAND ) ;
	DrawMenuBar ( Main_hWnd ) ;

	if ( modules_active == 0 )
	{
		KillTimer ( NULL, cursor_timer ) ;
		UnregisterClass ( winname, hInst ) ;
		UnregisterClass ( editname, hInst ) ;
	}
}

// general window paint routine. Called whenever a paint
// message is sent to the command window
void window_access::paint_window ( void )
{
int start_char,i ;
unsigned char op_code ;
BOOL abort_flag ;

	// clear cursor if it's drawn
	if ( cursor_state )
	{
		draw_cursor ( FALSE ) ;
		cursor_state = FALSE ;
	}

	// initialise the paint abort flag
	abort_flag = FALSE ;

   // v3.11 bug fix - window should not be scrolled in
   // when display has been scrolled
	if ( scroll_base_y == window_base_y )
	{
		// if window has just been sized to push the current_y cursor
		// out of the new window, this scrolls the text and cursor back in
		if ( cursor_y >= max_y )
		{
			int delta_y = cursor_y-max_y+1, y, new_max_y ;

			new_max_y = max_y ;
			max_y = cursor_y+1 ;

			for ( y=0 ; y<delta_y ; y++ )
				scroll_down() ;
			max_y = new_max_y ;
			cursor_y = max_y-1 ;

			// set abort flag to prevent any more painting till
			// the new scroll region comes in
			abort_flag = TRUE ;
		}
   }

	// all painting can be stopped by setting the abort flag
	if ( !abort_flag )
	{
		// if in mid-copy draw with highlighted region else re-draw normally
		if ( copying_selected )
			paint_copy_region () ;
		else
		{
			// first re-draw graphic bitmap
			if ( graphic_win != NULL )
			{
				RECT rc ;

				GetClientRect ( CommHwnd, &rc ) ;
				graphic_win->txfr_to_display_window ( hDC, rc.right-rc.left, rc.bottom-rc.top ) ;
				graphic_event = FALSE ;

				// ensure text bkgnd does not destroy graphics
				SetBkMode ( hDC, TRANSPARENT ) ;
			}
			// redraw text
			redraw_display() ;
			SetBkMode ( hDC, OPAQUE ) ;
		}
	}
	else
	{
		// if abort flag has been set, then a self induced paint is required
		// can't manually post a WM_PAINT message or call InvdalidateRect
		// so post a user defined msg which will be picked up by the
		// dialog box scheduler
		PostMessage ( CommHwnd, WM_COMMAND, AUTO_REPAINT, 0 ) ;
	}
}

// adjust display extents due to windows sizing operation
void window_access::resize_window ( int x_size, int y_size )
{
	int old_max_y,i ;

	old_max_y = max_y ;

	max_x = x_size / x_char ;
	max_y = ( y_size / y_char ) - 1 ;

	// clip display size to maximum specified if in fixed mode
	if ( display_options.fixed_size )
	{
		if ( max_x > display_options.max_x )
			max_x = display_options.max_x ;
		if ( max_y > display_options.max_y )
			max_y = display_options.max_y ;
	}

	// if new window size is now larger than previous
	// ensure the text & attr maps are initialised to prevent
	// old data being re-displayed
	if ( old_max_y < max_y )
	{
		for ( i=old_max_y ; i<max_y ; i++ )
		{
			memset ( text_map[i], VK_SPACE, MAX_POS_X ) ;
			memset ( attr_map[i], ATTR_NONE, MAX_POS_X ) ;
		}
	}

	// if graphic mode is active, will need to re-draw entire
	// window not just the newly uncovered bit
	if ( graphic_win != NULL )
		InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

// draw or clear the cursor from this device context
void window_access::draw_cursor ( BOOL draw_flag )
{
RECT rc ;

	rc.left = cursor_x*x_char ;
	rc.top = cursor_y*y_char ;
	rc.right = rc.left+x_char ;
	rc.bottom = rc.top+y_char ;

	InvertRect ( hDC, &rc ) ;
}

// toggle the cursor
void window_access::flash_cursor ( void )
{
	cursor_state=!cursor_state ;
	draw_cursor ( cursor_state ) ;

	// update graphic window if an event has occured
	if ( graphic_win != NULL && graphic_event )
	{
		RECT rc ;

		GetClientRect ( CommHwnd, &rc ) ;
		graphic_win->txfr_to_display_window ( hDC, rc.right-rc.left, rc.bottom-rc.top ) ;
		graphic_event = FALSE ;
	}
  // v4.0 - bug fix, redraw display if scrolled & graphics are displayed
  if ( GScroll )
  {
    InvalidateRect ( CommHwnd, NULL, TRUE ) ;
    GScroll = FALSE ;
  }
}

// draws text onto the device context at the x & y cursor co-ordinates
// from the text and attribute maps
void window_access::draw_attr_text ( int x, int y, int length )
{
	unsigned char op_code, saved_attr ;
	int new_length, sx ;

	// save current display attribute
	saved_attr = display_attr ;

	// initialise the display attributes from the current attribute map
	op_code = attr_map[y][x] ;
	set_display_attr ( op_code ) ;

	// initialise local counters - new string length and starting x-pos
	new_length = 0 ;
	sx = x ;

	// while there's still stuff to do
	while ( length > 0 )
	{
		// if attributes match current, update counters
		if ( attr_map[y][x] == op_code )
		{
			new_length++ ;
			length-- ;
			x++ ;
		}
		else
		{
			// output done so far
			TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], new_length ) ;
			// get new code and adjust context appropriately
			op_code = attr_map[y][x] ;
			set_display_attr ( op_code ) ;
			new_length = 0 ;
			sx=x ;
		}
	}
	// print any part processed string
	if ( new_length != 0 )
		TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], new_length ) ;

	// restore current attributes
	set_display_attr ( saved_attr ) ;
	display_attr = saved_attr ;
}

// re-draw entire text window based on text map
void window_access::redraw_display( void )
{
int y ;

	for ( y=0 ; y<max_y ; y++ )
		draw_attr_text ( 0, y, max_x ) ;
}

// scroll the window down adjusting the text map accordingly
void window_access::scroll_down()
{
int y_pos, i ;
RECT rc ;

	rc.left = 0 ;
	rc.top = 0 ;

	// if text is wrapping just scroll the fixed region, otherwise
	// take the whole width
	if ( display_options.wrap_flag )
		rc.right = max_x * x_char ;
	else
		rc.right = MAX_POS_X * x_char ;

	rc.bottom = max_y * y_char ;
	ScrollWindow ( CommHwnd, 0, -y_char, &rc, &rc ) ;

	window_base_y++ ;
	if ( window_base_y == MAX_SCROLL_Y )
		window_base_y = 0 ;

	scroll_base_y = window_base_y ;
	y_pos = window_base_y ;

	// scroll text & attr maps
	// move map position in region
	// v3.11 bug fix - use MAX_POS_Y over max_y to ensure
	//                 entire displayable area is remapped.
	//                 Bug Seen as: dynacalc display anomolies.
	for ( i=0 ; i < MAX_POS_Y ; i++ )
	{
		text_map[i] = text_region[y_pos] ;
		attr_map[i] = attr_region[y_pos] ;
		y_pos++ ;
		if ( y_pos == MAX_SCROLL_Y )
			y_pos = 0 ;
	}

	// initialise bottom row
	memset ( &text_map[cursor_y][0], VK_SPACE, MAX_POS_X ) ;
	memset ( &attr_map[cursor_y][0], ATTR_NONE, MAX_POS_X ) ;

	// force immediate update on display
//	UpdateWindow ( CommHwnd ) ;
  // set flag indicating scroll has occured whilst in graphics mode
	GScroll = ( graphic_win != NULL ) ;
}

// scroll the window up adjusting the text map accordingly
void window_access::scroll_up()
{
int y_pos,i ;
RECT rc ;

	rc.left = 0 ;
	rc.top = 0 ;

	// if text is wrapping just scroll the fixed region, otherwise
	// take the whole width
	if ( display_options.wrap_flag )
		rc.right = max_x * x_char ;
	else
		rc.right = MAX_POS_X * x_char ;

	rc.bottom = max_y * y_char ;
	ScrollWindow ( CommHwnd, 0, y_char, &rc, &rc ) ;

	window_base_y-- ;
	if ( window_base_y < 0 )
		window_base_y = MAX_SCROLL_Y-1 ;

	scroll_base_y = window_base_y ;
	y_pos = window_base_y ;

	// scroll text & attr maps
	// move map position in region
	// v3.11 bug fix - use MAX_POS_Y over max_y to ensure
	//                 entire displayable area is remapped.
	//                 Bug Seen as: dynacalc display anomolies.
	for ( i=0 ; i < MAX_POS_Y ; i++ )
	{
		text_map[i] = text_region[y_pos] ;
		attr_map[i] = attr_region[y_pos] ;
		y_pos++ ;
		if ( y_pos == MAX_SCROLL_Y )
			y_pos = 0 ;
	}

	memset ( text_map[0], VK_SPACE, MAX_POS_X ) ;
	memset ( attr_map[0], ATTR_NONE, MAX_POS_X ) ;

	UpdateWindow ( CommHwnd ) ;
  // set flag indicating scroll has occured whilst in graphics mode
	GScroll = ( graphic_win != NULL ) ;
}

// determine start x & y start co-ords of new copy region
void window_access::start_copy_region( int x_pos, int y_pos )
{
	csx = x_pos / x_char ;
	csy = y_pos / y_char ;
	cex = csx ;
	cey = csy ;
	if ( copying_selected )
		clear_copy_region() ;
	copying_active = TRUE ;
}

// routine requests copy region be cleared
void window_access::clear_copy_region ( void )
{
	copying_selected = FALSE ;
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

void window_access::end_copy_region ( void )
{
	// clear copy active flag
	copying_active = FALSE ;
	copying_selected = TRUE ;
}

// determine region selected
void window_access::mark_copy_region ( int x_pos, int y_pos )
{
	// determine new position of cursor
	x_pos = x_pos / x_char ;
	y_pos = y_pos / y_char ;

	// save new co-ords
	cex = x_pos ;
	cey = y_pos ;
	paint_copy_region() ;
}

// copy selected region to clipboard
void window_access::copy_selected_region ( SelectOpType SelectOp )
{
	HGLOBAL mem_handle ;
	char *copy_bfr, *ptr ;
	int px,y, x, space_pos ;
	FILE *export_file ;

	// if all co-ordinates of selected region are the same
	// then no-region is available for copy
	if ( cex == csx && cey == csy )
		return ;

	// get a block of memory to store the text in
	// including space for file CR/LF and NULL
	mem_handle = GlobalAlloc ( GHND, abs(cey-csy)*MAX_POS_X+abs(cex-csx)+3 ) ;
	if ( mem_handle == NULL )
		return ;

	copy_bfr = (char*)GlobalLock ( mem_handle )  ;
	ptr = copy_bfr ;

	// px is the top left x-cordinate in the text map to begin copying from
	px=sx ;
	// loop through selected lines
	for(y=sy;y<ey;y++)
	{
		// look for the last non-space char in the text map
		// assumed to be end of copy line
		space_pos=px ;
		for(x=px;x<MAX_POS_X;x++)
		{
			if ( text_map[y][x]!=' ' )
				space_pos=x ;
		}
		// copy valid part of line into copy buffer & add CR/LF
		memcpy ( ptr, &text_map[y][px], space_pos-px+1 ) ;
		ptr+=space_pos-px+1 ;
		*ptr = 0xd ;
		ptr++ ;
		*ptr = 0xa ;
		ptr++ ;
		// reset starting x position to 0 for start of next line
		px=0 ;
	}
	// paste in the last bit of the line, adding CR/LF & terminating null
	if ( ex>0 )
	{
		memcpy ( ptr, &text_map[ey][px], ex-px ) ;
		ptr+=ex-px ;
//		*ptr = 0xd ;
//		ptr++ ;
//		*ptr = 0xa ;
//		ptr++ ;
	}
	*ptr = 0 ;

	// if destination is marked for clipboard
	switch ( SelectOp )
	{
		case CLIPBOARD :

			// release memory and paste it on the clipboard
			GlobalUnlock ( mem_handle ) ;
			if ( OpenClipboard ( CommHwnd ) )
			{
				EmptyClipboard() ;
				SetClipboardData ( CF_OEMTEXT, mem_handle ) ;
				CloseClipboard() ;
			}
			break ;

		case DISK :

			// destination is export file
			export_file = OpenSaveFile ( CommHwnd, "export", "txt", "wb" ) ;
			if ( export_file != NULL )
			{
				fwrite ( copy_bfr, 1, ptr-copy_bfr, export_file ) ;
				fclose ( export_file ) ;
			}
			// v3.11 bug fix - memory never returned
			GlobalUnlock ( mem_handle ) ;
			GlobalFree ( mem_handle ) ;
			break ;

		case PASTEBFR :

			// new v4.0 - inject direct into paste buffer
			transfer_string_to_paste_bfr ( copy_bfr ) ;
			GlobalUnlock ( mem_handle ) ;
			GlobalFree ( mem_handle ) ;
			break ;
	}
}

// this routine re-draws the entire text window, highlighting
// the selected region
void window_access::paint_copy_region ( void )
{
	int x,y ;
	BOOL marking ;
	COLORREF high_fore_colour, high_back_colour ;

	// v3.11 - ensure top byte remains 00 - appears to cause
	// problems under WinNT
	high_fore_colour = (~fore_colour) & 0x00FFFFFFL ;
	high_back_colour = (~back_colour) & 0x00FFFFFFL ;

	// determine starting & ending points into sx & sy
	// if a single line is highlighted, determine the greater x-coord
	if ( cey == csy )
	{
		if ( cex < csx )
		{
			sx = cex ;
			ex = csx ;
		}
		else
		{
			sx = csx ;
			ex = cex ;
		}
		sy = csy ;
		ey = csy ;
	}
	else
	{
		// if ending line is now less than the starting line
		// top left x must now be the last co-ord
		if ( cey < csy )
		{
			sy = cey ;
			ey = csy ;
			sx = cex ;
			ex = csx ;
		}
		else
		{
			sy = csy ;
			ey = cey ;
			sx = csx ;
			ex = cex ;
		}
	}

	marking = FALSE ;

	// loop thru the no. of selected lines
	for ( y=0 ; y<max_y ; y++ )
	{
		// if only a single line is marked draw in the normal and
		// selected portions
		if ( y==sy && y==ey )
		{
			draw_attr_text ( 0, y, sx ) ;
			SetTextColor ( hDC, high_fore_colour ) ;
			SetBkColor ( hDC, high_back_colour ) ;
			TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], ex-sx ) ;
			draw_attr_text ( ex, y, max_x-ex ) ;
		}
		else
		{
			// if the starting line is reached, highlight from the starting
			// x-coord and leave the highlight pen selected
			if ( y==sy )
			{
				draw_attr_text ( 0, y, sx ) ;
				SetTextColor ( hDC, high_fore_colour ) ;
				SetBkColor ( hDC, high_back_colour ) ;
				TextOut ( hDC, sx*x_char, y*y_char, &text_map[y][sx], max_x-sx ) ;
				marking = TRUE ;
			}
			else
			{
				// if the ending line is reached, highlight the remaining
				// bit and restore the normal pen
				if ( y==ey )
				{
					TextOut ( hDC, 0, y*y_char, &text_map[y][0], ex ) ;
					draw_attr_text ( ex, y, max_x-ex ) ;
					marking = FALSE ;
				}
				else
					if ( marking )
						TextOut ( hDC, 0, y*y_char, &text_map[y][0], max_x ) ;
					else
						draw_attr_text (  0, y, max_x ) ;
			}
		}
	}
}

// this operation will update the device context, and window class
// variables to perform a specific screen emulation operation defined
// from either commemu.h, go51_emu.h or vt_emu.h
void window_access::process_emulation_code ( unsigned char op_code )
{
	int count, new_x, new_y ;
	LOGFONT attr_font ;

	switch ( op_code )
	{
		case Carriage_Return :

			cursor_x = 0 ;
			break ;

		case Line_Feed :

			if ( cursor_y < max_y-1 )
			{
				cursor_y++ ;
				break ;
			}
			else
			{
				scroll_down() ;
				break ;
			}

		case CURSOR_RIGHT :

			count = screen_emulator->new_x ;
			while ( count > 0 )
			{
				if ( cursor_x < max_x )
					cursor_x++ ;
				count-- ;
			}
			break ;

		case CURSOR_LEFT :

			count = screen_emulator->new_x ;
			while ( count > 0 )
			{
				if ( cursor_x > 0 )
					cursor_x-- ;
				count-- ;
			}
			break ;

		case CURSOR_UP :

			count = screen_emulator->new_y ;
			while ( count > 0 )
			{
				if ( cursor_y > 0 )
					cursor_y-- ;
				else
					scroll_up() ;
				count-- ;
			}
			break ;

		case CURSOR_DOWN :

			count = screen_emulator->new_y ;
			while ( count > 0 )
			{
				if ( cursor_y < max_y-1 )
					cursor_y++ ;
				else
					scroll_down() ;
				count-- ;
			}
			break ;

		case REVERSE_ON :

			display_attr = display_attr | ATTR_REV ;
			set_display_attr ( display_attr ) ;
			break ;

		case REVERSE_OFF :

			display_attr = display_attr & (~ATTR_REV) ;
			set_display_attr ( display_attr ) ;
			break ;

		case Bell :

//			sndPlaySound ("SystemExclamation", SND_SYNC);
			MessageBeep ( MB_ICONEXCLAMATION ) ;
			break ;

		case Backspace :

			if ( cursor_x == 0 )
			{
				if ( cursor_y != 0 )
				{
					cursor_y-- ;
					cursor_x = max_x-1 ;
				}
			}
			else
			{
					cursor_x-- ;
			}
			break ;

		case Tab :

			new_x = screen_emulator->find_next_tab ( cursor_x ) ;
			if ( new_x < MAX_POS_X )
				cursor_x = new_x ;
			break ;

		case Home :

			cursor_x = 0 ;
			cursor_y = 0 ;
			break ;

		case Clear :

			cursor_x = 0 ;
			cursor_y = 0 ;

		case CLEAR_AOS :

			for ( count=0 ; count < MAX_POS_Y ; count++ )
			{
				memset ( text_map[count], VK_SPACE, MAX_POS_X ) ;
				memset ( attr_map[count], ATTR_NONE, MAX_POS_X ) ;
			}
			InvalidateRect ( CommHwnd, NULL, TRUE ) ;
			break ;

		case UNDERLINE_ON :

			display_attr = display_attr | ATTR_ULINE ;
			set_display_attr ( display_attr ) ;
			break ;

		case UNDERLINE_OFF :

			display_attr = display_attr & (~ATTR_ULINE) ;
			set_display_attr ( display_attr ) ;
			break ;

		case CLEAR_EOS :

			memset ( &text_map[cursor_y][cursor_x], VK_SPACE, MAX_POS_X - cursor_x ) ;
			memset ( &attr_map[cursor_y][cursor_x], ATTR_NONE, MAX_POS_X - cursor_x ) ;
			for ( count=cursor_y ; count < MAX_POS_Y ; count++ )
			{
				memset ( text_map[count], VK_SPACE, MAX_POS_X ) ;
				memset ( attr_map[count], ATTR_NONE, MAX_POS_X ) ;
			}
			InvalidateRect ( CommHwnd, NULL, FALSE ) ;
			break ;

		case CLEAR_EOL :

			memset ( &text_map[cursor_y][cursor_x], 32, MAX_POS_X - cursor_x ) ;
			memset ( &attr_map[cursor_y][cursor_x], ATTR_NONE, MAX_POS_X - cursor_x ) ;
			InvalidateRect ( CommHwnd, NULL, FALSE ) ;
			break ;

		case SET_CURSOR :

			new_x = screen_emulator->new_x ;
			new_y = screen_emulator->new_y ;
			if ( new_x >=0 && new_x < MAX_POS_X )
				cursor_x = new_x ;
			if ( new_y >=0 && new_y < MAX_POS_Y )
				cursor_y = new_y ;
			break ;

		case SCROLL_UP :

			scroll_up() ;
			break ;

		case SCROLL_DOWN :

			scroll_down() ;
			break ;

		case SAVE_ATTRS :

			attr_save_x = cursor_x ;
			attr_save_y = cursor_y ;
			saved_attr = display_attr ;
			break ;

		case REST_ATTRS :

			cursor_x = attr_save_x ;
			cursor_y = attr_save_y ;
			display_attr = saved_attr ;
			set_display_attr ( display_attr ) ;
			break ;

		case SET_TAB :

			screen_emulator->modify_tab ( cursor_x, TRUE ) ;
			break ;

		case RESET_TERM :

			break ;

		case NO_ATTRS :

			display_attr = ATTR_NONE ;
			set_display_attr ( display_attr ) ;
			break ;

		case BOLD_ON :

			display_attr = display_attr | ATTR_BOLD ;
			set_display_attr ( display_attr ) ;
			break ;

		case BOLD_OFF :

			display_attr = display_attr & (~ATTR_BOLD) ;
			set_display_attr ( display_attr ) ;
			break ;

		case BLINK_ON :

			display_attr = display_attr | ATTR_BLINK ;
			set_display_attr ( display_attr ) ;
			break ;

		case INVISIBLE_ON :

			display_attr = display_attr | ATTR_INVIS ;
			set_display_attr ( display_attr ) ;
			break ;

		case WIDTH_132 :

			if ( display_options.fixed_size )
			{
				display_options.max_x = 132 ;
				if ( max_x > display_options.max_x )
					max_x = display_options.max_x ;
			}
			break ;

		case WIDTH_80 :

			if ( display_options.fixed_size )
			{
				display_options.max_x = 80 ;
				if ( max_x > display_options.max_x )
					max_x = display_options.max_x ;
			}
			break ;

		case AUTO_WRAP :

			display_options.wrap_flag = TRUE ;
			break ;

		case RESET_WRAP :

			display_options.wrap_flag = FALSE ;
			break ;

		case CLEAR_TAB :

			screen_emulator->modify_tab ( cursor_x, FALSE ) ;
			break ;

		case CLEAR_LEFT :

			memset ( &text_map[cursor_y][0], VK_SPACE, cursor_x ) ;
			memset ( &attr_map[cursor_y][0], ATTR_NONE, cursor_x ) ;
			InvalidateRect ( CommHwnd, NULL, FALSE ) ;
			break ;

		case CLEAR_TOS :

			for ( count=0 ; count<cursor_y ; count++ )
			{
				memset ( text_map[count], VK_SPACE, MAX_POS_X ) ;
				memset ( attr_map[count], ATTR_NONE, MAX_POS_X ) ;
			}
			memset ( &text_map[cursor_y][0], VK_SPACE, cursor_x ) ;
			memset ( &attr_map[cursor_y][0], ATTR_NONE, cursor_x ) ;
			InvalidateRect ( CommHwnd, NULL, FALSE ) ;
			break ;

		case CLEAR_LINE :

			memset ( &text_map[cursor_y][0], VK_SPACE, MAX_POS_X ) ;
			memset ( &attr_map[cursor_y][0], ATTR_NONE, MAX_POS_X ) ;
			InvalidateRect ( CommHwnd, NULL, FALSE ) ;
			break ;

	}
}

// sets the display attributes according to the supplied attribute code
void window_access::set_display_attr ( unsigned char attr_code )
{
	BOOL new_font = FALSE ;
	COLORREF new_fore_colour, new_back_colour ;

	// determine text colours based on reverse attribute setting
	if ( ( attr_code & ATTR_REV ) == ATTR_REV )
	{
		new_fore_colour = fore_colour ;
		new_back_colour = back_colour ;
	}
	else
	{
		new_back_colour = fore_colour ;
		new_fore_colour = back_colour ;
	}

	// select a colour for blinking text
	if ( attr_code & ATTR_BLINK )
		new_fore_colour = RGB ( 255, 128, 128 ) ;

	// if invisible attribute set, make foreground colour same as bkgnd
	if ( attr_code & ATTR_INVIS )
		new_fore_colour = new_back_colour ;

	// setup the screen colours
	SetBkColor ( hDC, new_fore_colour ) ;
	SetTextColor ( hDC, new_back_colour ) ;

	// check for bold test.
	if ( ( attr_code & ATTR_BOLD ) == ATTR_BOLD )
	{
		// if bold text not already selected, set attr in font structure
		if ( font_definition.lfWeight != FW_BOLD )
		{
			font_definition.lfWeight = FW_BOLD ;
			new_font = TRUE ;
		}
	}
	else
		// if bold selected in font, it is no longer reqd
		if ( font_definition.lfWeight == FW_BOLD )
		{
			font_definition.lfWeight = FW_NORMAL ;
			new_font = TRUE ;
		}

	// perform similar checks for underline text
	if ( ( attr_code & ATTR_ULINE ) == ATTR_ULINE )
	{
		if ( !font_definition.lfUnderline )
		{
			font_definition.lfUnderline = TRUE ;
			new_font = TRUE ;
		}
	}
	else
		if ( font_definition.lfUnderline )
		{
			font_definition.lfUnderline = FALSE ;
			new_font = TRUE ;
		}

	// if a new font is deemed necessary
	if ( new_font )
	{
		// de-select current font from context
		SelectObject ( hDC, holdsfont ) ;

		// delete font and create new one
		DeleteObject(hnewsfont) ;
		hnewsfont = CreateFontIndirect((LPLOGFONT) &font_definition );

		// select in new font
		holdsfont=SelectObject(hDC, hnewsfont);
	}
}

// set display options for window size and terminal emulation mode
void window_access::set_display_options ( void )
{
	RECT rc ;
	int old_max_y,i ;

	old_max_y = max_y ;

	if ( display_options.fixed_size )
	{
		// range check the display size
		if ( display_options.max_x > MAX_POS_X )
			display_options.max_x = MAX_POS_X ;
		if ( display_options.max_y > MAX_POS_Y )
			display_options.max_y = MAX_POS_Y ;

		// clip display extents to those specified
		if ( max_x > display_options.max_x )
			max_x = display_options.max_x ;
		if ( max_y > display_options.max_y )
			max_y = display_options.max_y ;
	}
	else
	{
		// get size of window and set display extents to it
		GetClientRect ( CommHwnd, &rc ) ;
		max_y = ( rc.bottom - rc.top ) / y_char ;
		max_x = ( rc.right - rc.left ) / x_char ;
	}

	// if new window size is now larger than previous
	// ensure the text & attr maps are initialised to prevent
	// old data being re-displayed
	if ( old_max_y < max_y )
	{
		for ( i=old_max_y ; i<max_y ; i++ )
		{
			memset ( text_map[i], VK_SPACE, MAX_POS_X ) ;
			memset ( attr_map[i], ATTR_NONE, MAX_POS_X ) ;
		}
	}

	if ( screen_emulator != NULL )
		delete screen_emulator ;

	// activate the required screen emulation module
	switch ( display_options.emulation )
	{
		case NO_EMULATION :

			screen_emulator = new common_emulator ;
			break ;

		case GO51_EMUL :

			screen_emulator = new go51_emulator ;
			break ;

		case VT100_EMUL :

			screen_emulator = new vt100_emulator ;
			break ;
	}

	// force re-paint to take into account new changes
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

// returns a pointer to either paste buffer or keyboard buffer
// and byte count (up to 255 bytes) of data to be passed to remote
char *window_access::get_paste_buffer ( unsigned short *buffer_size )
{
	char *bfr_ptr ;
	const unsigned short max_paste_return = 255 ;

	// if import file open
	if ( import_file != NULL )
	{
		// read a chunk from the file
		*buffer_size = fread ( &import_bfr, 1, max_paste_return, import_file ) ;
		// if a smaller chunk is retrieved, assume this to be end of file
		// and close it
		if ( *buffer_size != max_paste_return )
		{
			fclose ( import_file ) ;
			import_file = NULL ;
		}
		return import_bfr ;
	}

	// if paste bfr exists
	if ( paste_buffer != NULL )
	{
		// if paste bfr >255 bytes, return just a 255 byte chunk
		// and update ptrs
		if ( paste_bfr_size > max_paste_return )
		{
			paste_bfr_size-=max_paste_return ;
			*buffer_size = max_paste_return ;
			bfr_ptr = paste_ptr ;
			paste_ptr+=max_paste_return ;
			return bfr_ptr ;
		}

		// if buffer empty free buffer
		if ( paste_bfr_size == 0 )
		{
			free ( paste_buffer ) ;
			paste_buffer = NULL ;
		}
		else
		{
			// return remaining chunk in paste bfr
			*buffer_size=paste_bfr_size ;
			bfr_ptr = paste_ptr ;
			paste_bfr_size = 0 ;
			return bfr_ptr ;
		}
	}

	// no paste buffer available, so check if anything in keyboard bfr
	if ( keyb_write != 0 )
	{
		 *buffer_size = keyb_write ;
		 keyb_write = 0 ;
		 return keyb_buffer ;
	}

	// no data is available
	*buffer_size = 0 ;
	return NULL ;
}

// copy the windows clipboard into the local paste buffer - up to 64K
void window_access::copy_clipboard_to_paste_buffer ( void )
{
	unsigned long i ;
	const unsigned long max_clipboard_size = 65535 ;
	char *clip_ptr, *clipbrd_start ;
	HANDLE clipboard_handle ;

	paste_bfr_size = 0 ;

	// open and get handle to clipboard data
	if ( !OpenClipboard ( CommHwnd ) )
		return ;
	clipboard_handle = GetClipboardData ( CF_TEXT ) ;
	if ( clipboard_handle == NULL )
		return ;
	clipbrd_start = (unsigned char*)GlobalLock ( clipboard_handle ) ;
	clip_ptr = clipbrd_start ;

	// look for terminating null at end of clipboard
	for (i=0 ; i<max_clipboard_size ; i++ )
	{
		if ( *clip_ptr == 0 )
		{
			paste_bfr_size = i ;
			i=max_clipboard_size ;
		}
		clip_ptr++ ;
	}

	if ( paste_bfr_size != 0 )
	{
		// grab paste buffer, deleting old one first if necessary
		if ( paste_buffer != NULL )
			free ( paste_buffer ) ;
		paste_buffer = (char*)malloc ( paste_bfr_size ) ;
		if ( paste_buffer == NULL )
		{
			GlobalUnlock ( clipboard_handle ) ;
			CloseClipboard() ;
			return ;
		}

		// copy clipboard text to local bfr
		memcpy ( paste_buffer, clipbrd_start, paste_bfr_size ) ;
		// working data pointer
		paste_ptr = paste_buffer ;
	}
	// release clipboard
	GlobalUnlock ( clipboard_handle ) ;
	CloseClipboard() ;
}

// copy the passed keypress into the keyboard buffer
void window_access::copy_keypress_to_paste_buffer ( unsigned char keypress, BOOL special )
{
	char keycode ;
	int i ;

	// if special keycode passed in check for a mapping
	// in our keymap
	if ( special )
	{
		// if no match found no further processing needed
		if ( vtkey_to_udfkey[keypress] == 0xff )
			return ;

		// if key has not been mapped, no further processing is required
		if ( keymap.keys[keystate][vtkey_to_udfkey[keypress]][0] == 0 )
			return ;

		// set flag indicating that possibly the next WM_CHAR message
		// is to be discarded
		keypress_discard = TRUE ;

		// loop thru mapping definitions in keymap
		for ( i=0 ; i< CHARS_PER_KEY ; i++ )
		{
			keycode = keymap.keys[keystate][vtkey_to_udfkey[keypress]][i] ;

			// if entry not set to 0 and space in keyboard buffer
			// copy entry into buffer
			if ( keycode != 0 && keyb_write < 256 )
			{
				keyb_buffer[keyb_write] = keycode ;
				keyb_write++ ;
			}
		}
	}
	else
	{
		// if flag set to discard this char, clear flag and exit
		if ( keypress_discard )
		{
			keypress_discard = FALSE ;
			return ;
		}
		// discard character if insufficient room
		if ( keyb_write == 256 )
			return ;
		// add normal char to buffer
		keyb_buffer[keyb_write] = (char)keypress ;
		keyb_write++ ;
	}
}

void window_access::new_display_font( void )
{
	TEXTMETRIC tm ;
	RECT rc ;

	// create new user font
	SelectObject ( hDC, holdsfont ) ;
	get_new_font ( CommHwnd, &font_definition, &fore_colour ) ;
	DeleteObject(hnewsfont) ;
	hnewsfont = CreateFontIndirect((LPLOGFONT) &font_definition );
	holdsfont = SelectObject ( hDC, hnewsfont ) ;

	// get new text sizes
	GetTextMetrics(hDC, &tm);
	x_char = tm.tmAveCharWidth ;
	y_char = tm.tmHeight + tm.tmExternalLeading;
	// re-define window in terms of new font
	GetClientRect ( CommHwnd, &rc ) ;
	resize_window ( rc.right, rc.bottom ) ;
	InvalidateRect ( CommHwnd, NULL, TRUE ) ;
}

// map display window starting at specified posn in region
void window_access::remap_window ( int top_row )
{
	int y_pos, i ;

	// restore window position
	y_pos = top_row ;
	for ( i=0 ; i < max_y ; i++ )
	{
		text_map[i] = text_region[y_pos] ;
		attr_map[i] = attr_region[y_pos] ;
		y_pos++ ;
		if ( y_pos == MAX_SCROLL_Y )
			y_pos = 0 ;
	}
}

void window_access::process_scroll_request ( WPARAM wParam, LPARAM lParam )
{

	int nVscrollInc ;           // scroll increments
	int delta_map_y ;
	RECT rc ;

	// React to the various vertical scroll related actions.
	switch (GET_WM_VSCROLL_CODE(wParam, lParam))
	{
		case SB_TOP:
			nVscrollInc =-scroll_pos ;
			break;

		case SB_BOTTOM:
			nVscrollInc = MAX_SCROLL_Y_POS - scroll_pos ;
			break;

		case SB_LINEUP:
			nVscrollInc = -1;
			break;

		case SB_LINEDOWN:
			nVscrollInc = 1;
			break;

		case SB_PAGEUP:
			nVscrollInc = -max(1, max_y );
			break;

		case SB_PAGEDOWN:
			nVscrollInc = max(1, max_y );
			break;

		case SB_THUMBPOSITION:
			nVscrollInc = GET_WM_VSCROLL_POS(wParam, lParam) - scroll_pos ;
			break;

		case SB_THUMBTRACK:
			nVscrollInc = GET_WM_VSCROLL_POS(wParam, lParam) - scroll_pos ;
			break;

		default:
			nVscrollInc = 0 ;
	}
	nVscrollInc = max(-scroll_pos,
		min(nVscrollInc, MAX_SCROLL_Y_POS - scroll_pos));

	// if scrolling has occurred
	if (nVscrollInc)
	{
		// clear cursor if it's drawn
		if ( cursor_state )
		{
			draw_cursor ( FALSE ) ;
			cursor_state = FALSE ;
		}

		// determine new scroll position
		scroll_pos += nVscrollInc;

		// scroll bar is normally at max extent, so determine a delta from this point
		// this gives the number of lines needed to move in the text/attr regions
		delta_map_y = MAX_SCROLL_Y_POS - scroll_pos ;

		// determine absolute top left of new display window based on current
		// dealing with wrap as necessary
		scroll_base_y = (window_base_y - delta_map_y) ;
		if ( scroll_base_y < 0 )
			scroll_base_y = MAX_SCROLL_Y + scroll_base_y ;

		// remap the text window to this new position
		remap_window ( scroll_base_y ) ;
		// adjust the scrollbar position
		SetScrollPos(CommHwnd, SB_VERT, scroll_pos, TRUE);

		rc.left = 0 ;
		rc.top = 0 ;

		// if text is wrapping just scroll the fixed region, otherwise
		// take the whole width
		if ( display_options.wrap_flag )
			rc.right = max_x * x_char ;
		else
			rc.right = MAX_POS_X * x_char ;

		rc.bottom = max_y * y_char ;
		ScrollWindow ( CommHwnd, 0, -y_char * nVscrollInc, &rc, &rc ) ;

		UpdateWindow( CommHwnd );
	}
}

// new in V4.00 - transfer a text string into paste buffer
void window_access::transfer_string_to_paste_bfr ( char *string )
{
	// determine size of paste buffer needed
	paste_bfr_size = strlen ( string ) ;

	// grab paste buffer, deleting old one first if necessary
	if ( paste_buffer != NULL )
		free ( paste_buffer ) ;
	paste_buffer = (char*)malloc ( paste_bfr_size ) ;

	// copy in string
	memcpy ( paste_buffer, string, paste_bfr_size ) ;

	// working data pointer
	paste_ptr = paste_buffer ;
}

// command shell window
// this handles all the windows messages sent to PC-Share command shell
LRESULT CALLBACK  MainWndProc (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
window_access *user_window ;
FILE *key_file ;
HMENU hMenu ;
RECT rc, ToolRC, MainRC ;

	switch ( Message )
	{
		case WM_INITDIALOG :

			return 1 ;

		case WM_COMMAND :

			// menu commands
			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case CM_FILEIMPORT :

					user_window = find_owner_class ( hDlg ) ;
					if ( user_window->import_file != NULL )
						fclose ( user_window->import_file ) ;
					user_window->import_file = OpenLoadFile ( hDlg, "import", "txt", "rb" ) ;
					break ;

				case CM_FILEEXPORT :

					// copy selected region to file
					find_owner_class ( hDlg )->copy_selected_region( DISK ) ;
					break ;

				case CM_FILECAPTURE :

					user_window = find_owner_class ( hDlg ) ;
					hMenu = GetMenu ( hDlg ) ;
					if ( user_window->capture_file != NULL )
					{
						fclose ( user_window->capture_file ) ;
						user_window->capture_file = NULL ;
						CheckMenuItem ( hMenu, CM_FILECAPTURE, MF_UNCHECKED ) ;
					}
					else
					{
						user_window->capture_file = OpenSaveFile ( hDlg, "capture", "txt", "wt" ) ;
						if ( user_window->capture_file != NULL )
							CheckMenuItem ( hMenu, CM_FILECAPTURE, MF_CHECKED ) ;
					}
					break ;

				case CM_EDITCOPY :

					// if copy region selected, copy it to the clipboard
					user_window = find_owner_class ( hDlg ) ;
					if ( user_window->copying_selected )
						user_window->copy_selected_region( CLIPBOARD ) ;
					break ;

				case CM_EDITPASTE :

					// transfer clipboard into local paste buffer
					find_owner_class ( hDlg )->copy_clipboard_to_paste_buffer() ;
					break ;

				case CM_EDITSELECT_ALL :
				{
					RECT rc ;

					// get size of window client
					GetClientRect ( hDlg, &rc ) ;
					// mark entire client area as selected
					user_window = find_owner_class ( hDlg ) ;
					user_window->start_copy_region ( rc.left, rc.top ) ;
					user_window->mark_copy_region ( rc.right, rc.bottom ) ;
					user_window->end_copy_region() ;
				}
					break ;

				case DISPLAY_OPTIONS :

					user_window = find_owner_class ( hDlg ) ;
					// activate dialog box to get new display options
					DialogBoxParam ( hInst, MAKEINTRESOURCE ( DISPLAY_DIALOG ), hDlg,
						(DLGPROC)DisplayOptions, (LPARAM)&user_window->display_options ) ;
					user_window->set_display_options() ;
					break ;

				case DISPLAY_FONT :

					// get new font and update current display
					find_owner_class ( hDlg )->new_display_font() ;
					break ;

				case DISPLAY_FORECOLOUR :

					// get new foreground colour and re-draw display
					user_window = find_owner_class ( hDlg ) ;
					get_new_colour ( hDlg, &user_window->fore_colour ) ;
					InvalidateRect ( hDlg, NULL, TRUE ) ;
					break ;

				case DISPLAY_BACKCOLOUR :

					// get new background colour and re-draw display
					user_window = find_owner_class ( hDlg ) ;
					get_new_colour ( hDlg, &user_window->back_colour ) ;
					InvalidateRect ( hDlg, NULL, TRUE ) ;
					break ;

				case DISPLAY_KEYS :

					// define new keymap for this window
					user_window = find_owner_class ( hDlg ) ;
					DialogBoxParam ( hInst, MAKEINTRESOURCE ( KEYMAPS ), hDlg,
						(DLGPROC)DefineKeyMap, (LPARAM)&user_window->keymap ) ;
					break ;

				case SAVE_KEYMAP :

					// save keymap to disk file
					user_window = find_owner_class ( hDlg ) ;
					key_file = OpenSaveFile ( hDlg, "keys", "map", "wb" ) ;
					if ( key_file != NULL )
					{
						fwrite ( &user_window->keymap, sizeof ( struct keymap_type ), 1, key_file ) ;
						fclose ( key_file ) ;
					}
					break ;

				case LOAD_KEYMAP :

					// load keymap from disk file
					user_window = find_owner_class ( hDlg ) ;
					key_file = OpenLoadFile ( hDlg, "keys", "map", "rb" ) ;
					if ( key_file != NULL )
					{
						fread ( &user_window->keymap, sizeof ( struct keymap_type ), 1, key_file ) ;
						fclose ( key_file ) ;
					}
					break ;

				case KEYS_DEFAULT :

					// check before updating the system config keyboard map
					if ( MessageBox ( hDlg, "Overwrite System Keyboard Map with current map?",
						"Save as System Keymap", MB_OKCANCEL | MB_ICONQUESTION ) == IDOK )
					{
						user_window = find_owner_class ( hDlg ) ;
						system_config.keymap = user_window->keymap ;
						save_config_data() ;
					}
					break ;
			}
			return 0 ;

		case WM_SIZE :

			// get size of toolbar window
			user_window = find_owner_class( hDlg ) ;
			if ( ( user_window != NULL ) && ( wParam != SIZE_MINIMIZED ) )
			{
				GetClientRect ( user_window->ToolWnd, &ToolRC ) ;
				rc.top = ToolRC.bottom ;
				rc.left = ToolRC.left ;
				// resize editable display window to new size
				// taking off toolbar size
				MoveWindow ( user_window->CommHwnd,
					ToolRC.left, ToolRC.bottom, LOWORD ( lParam ),
						HIWORD ( lParam ) - (ToolRC.bottom-ToolRC.top), TRUE ) ;
			}
			return 0 ;

		case WM_CHAR :

			// transfer keypress to paste bfr
			find_owner_class ( hDlg )->copy_keypress_to_paste_buffer ( (unsigned char)wParam, FALSE ) ;
			return 0 ;

		case WM_KEYDOWN :

			user_window = find_owner_class ( hDlg ) ;
			// if keypress is SHIFT or CTRL, adjust keystate accordingly
			// otherwise pass to window for processing
			switch ( wParam )
			{
				case VK_SHIFT :
					user_window->keystate = SHIFTED ;
					break ;
				case VK_CONTROL :
					user_window->keystate = CONTROL ;
					break ;
				default :
					user_window->copy_keypress_to_paste_buffer ( (unsigned char)wParam, TRUE ) ;
			}
			break ;

		case WM_KEYUP :

			// if shift or ctrl key has been released, modify keyboard state
			// for current window
			// if keypress discard flag has not been cleared, clear it now
			user_window = find_owner_class ( hDlg ) ;
			switch ( wParam )
			{
				case VK_SHIFT :
				case VK_CONTROL :
					user_window->keystate = NORMAL_CHAR ;
					break ;
				default :
					if ( user_window->keypress_discard )
						user_window->keypress_discard = FALSE ;
			}
			break ;

		case WM_SETFOCUS :

			active_hwnd = hDlg ;
			return 0 ;

		case WM_CLOSE :

			// prompt user for confirmation of closing window as it will be closed
			// without the knowledge of the host computer
			if ( MessageBox ( hDlg, "Are you sure you want to close this active window?",
				"Close Active Window", MB_OKCANCEL | MB_ICONQUESTION ) == IDOK )
			{
				int i ;

				// find the window and delete it - don't call destroywindow
				// as the class destructor will do it
				for(i=0;i<MAX_WINDOWS;i++)
				{
					if ( active_windows[i] != NULL )
						if ( active_windows[i]->MainWnd == hDlg )
						{
							delete active_windows[i] ;
							active_windows[i] = NULL ;
						}
				}
			}
			break ;

		default:
				return DefWindowProc(hDlg, Message, wParam, lParam);
	}
	return 0 ;
}

// this operation returns a pointer to the class
// which owns the specified window identity
window_access *find_owner_class( HWND hWnd )
{
	int i ;
	for(i=0;i<MAX_WINDOWS;i++)
	{
		if ( active_windows[i] != NULL )
			if ( active_windows[i]->MainWnd == hWnd )
				return active_windows[i] ;
	}
	return NULL ;
}

// timer procedure called once a second by windows used
// to flash cursors in all active windows
#pragma argsused
void CALLBACK _export CursorTimer ( HWND hWnd, UINT msg, UINT idTimer, DWORD dwTime )
{
	int i ;

	// tell all active windows to flash the cursor
	for (i=0 ; i<MAX_WINDOWS;i++)
	{
		if ( active_windows[i] != NULL )
			active_windows[i]->flash_cursor() ;
	}
}

// dialog box procedure for entering display options
BOOL CALLBACK _export DisplayOptions (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
	static struct display_options_type *display_options ;
	BOOL tx ;

	switch ( Message )
	{
		case WM_INITDIALOG :

			// get parameter passed to dialog box
			display_options = (struct display_options_type*)lParam ;

			// initialise dialog box options with supplied structure
			switch ( display_options->emulation )
			{
				case NO_EMULATION :

					CheckDlgButton ( hDlg, EMULATION_NONE, 1 ) ;
					break ;

				case VT100_EMUL :

					CheckDlgButton ( hDlg, EMULATION_VT100, 1 ) ;
					break ;

				case GO51_EMUL :

					CheckDlgButton ( hDlg, EMULATION_GO51, 1 ) ;
					break ;
			}

			SetDlgItemInt ( hDlg, X_CHARS, display_options->max_x, FALSE ) ;
			SetDlgItemInt ( hDlg, Y_CHARS, display_options->max_y, FALSE ) ;

			if ( display_options->fixed_size )
				CheckDlgButton ( hDlg, DISPLAY_FIX, 1 ) ;
			else
			{
				CheckDlgButton ( hDlg, DISPLAY_VAR, 1 ) ;
				EnableWindow ( GetDlgItem ( hDlg, X_CHARS ), FALSE ) ;
				EnableWindow ( GetDlgItem ( hDlg, Y_CHARS ), FALSE ) ;
			}
			if ( display_options->wrap_flag )
				CheckDlgButton ( hDlg, TEXT_WRAP, 1 ) ;

			return 1 ;

		case WM_COMMAND :

			// menu commands
			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case DISPLAY_FIX :

					CheckDlgButton ( hDlg, DISPLAY_FIX, 1 ) ;
					CheckDlgButton ( hDlg, DISPLAY_VAR, 0 ) ;
					EnableWindow ( GetDlgItem ( hDlg, X_CHARS ), TRUE ) ;
					EnableWindow ( GetDlgItem ( hDlg, Y_CHARS ), TRUE ) ;
					break ;

				case DISPLAY_VAR :

					CheckDlgButton ( hDlg, DISPLAY_FIX, 0 ) ;
					CheckDlgButton ( hDlg, DISPLAY_VAR, 1 ) ;
					EnableWindow ( GetDlgItem ( hDlg, X_CHARS ), FALSE ) ;
					EnableWindow ( GetDlgItem ( hDlg, Y_CHARS ), FALSE ) ;
					break ;

				case EMULATION_NONE :

					CheckDlgButton ( hDlg, EMULATION_NONE, 1 ) ;
					CheckDlgButton ( hDlg, EMULATION_VT100, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_GO51, 0 ) ;
					break ;

				case EMULATION_VT100 :

					CheckDlgButton ( hDlg, EMULATION_NONE, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_VT100, 1 ) ;
					CheckDlgButton ( hDlg, EMULATION_GO51, 0 ) ;
					break ;

				case EMULATION_GO51 :

					CheckDlgButton ( hDlg, EMULATION_NONE, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_VT100, 0 ) ;
					CheckDlgButton ( hDlg, EMULATION_GO51, 1 ) ;
					break ;

				case IDOK :

					// load up return information from dialog box
					if ( IsDlgButtonChecked ( hDlg, EMULATION_NONE ) )
						display_options->emulation = NO_EMULATION ;
					if ( IsDlgButtonChecked ( hDlg, EMULATION_VT100 ) )
						display_options->emulation = VT100_EMUL ;
					if ( IsDlgButtonChecked ( hDlg, EMULATION_GO51 ) )
						display_options->emulation = GO51_EMUL ;

					if ( IsDlgButtonChecked ( hDlg, DISPLAY_FIX ) )
					{
						display_options->fixed_size = TRUE ;
						display_options->max_x = GetDlgItemInt ( hDlg, X_CHARS, &tx, FALSE ) ;
						display_options->max_y = GetDlgItemInt ( hDlg, Y_CHARS, &tx, FALSE ) ;
					}
					else
						display_options->fixed_size = FALSE ;

					if ( IsDlgButtonChecked ( hDlg, TEXT_WRAP ) )
						display_options->wrap_flag = TRUE ;
					else
						display_options->wrap_flag = FALSE ;

					EndDialog ( hDlg, 0 ) ;
					break ;
			}
			return 0 ;
	}
	return 0 ;
}

LRESULT CALLBACK EditProc (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
window_access *user_window ;
PAINTSTRUCT ps;

	switch ( Message )
	{
		case WM_INITDIALOG :

			return 1 ;

		case WM_COMMAND :

			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case AUTO_REPAINT :

					// special message caused within paint operations
					// to force paint message
					InvalidateRect ( hDlg, NULL, FALSE ) ;
					break ;
			}
			return 0 ;

		case WM_LBUTTONDOWN :

			// if mouse button is pressed, notify the system a copy region is
			// about to be selected
			user_window = find_owner_class ( GetParent ( hDlg ) ) ;
			user_window->start_copy_region ( LOWORD ( lParam ), HIWORD ( lParam ) ) ;
			return 0 ;

		case WM_LBUTTONUP :

			// if button released, set update state flags
			user_window = find_owner_class ( GetParent ( hDlg ) ) ;
			if ( user_window->copying_active )
				user_window->end_copy_region() ;
			return 0 ;

		case WM_RBUTTONDOWN :

			// new for v4.0 - right click inserts contents of selected
			// region into paste buffer - 'quick paste'
			find_owner_class ( GetParent ( hDlg ) )->copy_selected_region( PASTEBFR ) ;
			return 0 ;

		case WM_MOUSEMOVE :

			// if mouse is moved whilst button is depressed, highlight the selected
			// region
			user_window = find_owner_class ( GetParent ( hDlg ) ) ;
			if ( user_window->copying_active && ( ( wParam & MK_LBUTTON ) == MK_LBUTTON ) )
				user_window->mark_copy_region ( LOWORD ( lParam ), HIWORD ( lParam ) ) ;
			return 0 ;

		case WM_PAINT :

			// call the window class paint handler
			BeginPaint ( hDlg, &ps ) ;

			user_window = find_owner_class( GetParent ( hDlg ) ) ;
			if ( user_window != NULL )
				user_window->paint_window () ;
			EndPaint ( hDlg, &ps ) ;
			return 0 ;

		case WM_ERASEBKGND :
		{
			HBRUSH back_brush ;
			RECT rc ;
			HDC hdc = (HDC)wParam ;

			// erase background to selected window's colour
			// or system default if it is not yet active
			user_window = find_owner_class( GetParent ( hDlg ) ) ;
			if ( user_window != NULL )
				back_brush = CreateSolidBrush ( user_window->back_colour ) ;
			else
				back_brush = CreateSolidBrush ( system_config.back_colour ) ;
			GetClientRect ( hDlg, &rc ) ;
			FillRect ( hdc, &rc, back_brush ) ;
			DeleteObject ( back_brush ) ;
			return 1 ;
		}

		case WM_SIZE :

			// on a re-size, get the new co-ordinates of the window
			// and indicate the arriving paint message has been caused by a re-size
			user_window = find_owner_class( GetParent ( hDlg ) ) ;
			if ( ( user_window != NULL ) && ( wParam != SIZE_MINIMIZED ) )
				user_window->resize_window ( LOWORD ( lParam ), HIWORD ( lParam ) ) ;
			return 0 ;

		case WM_VSCROLL:

			find_owner_class (
				GetParent(hDlg) )->process_scroll_request ( wParam, lParam ) ;
			break ;

		default:
				return DefWindowProc(hDlg, Message, wParam, lParam);
	}
	return 0 ;
}

BOOL CALLBACK _export EditToolBar (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
window_access *user_window ;
HMENU hMenu ;
extern char helpfilename[] ;
int cmd_ident ;
char cmd[CMD_BUF_SIZE] ;

	switch ( Message )
	{
		case WM_INITDIALOG :
			return 1 ;

		case WM_COMMAND :

			switch ( GET_WM_COMMAND_ID ( wParam, lParam ) )
			{
				case TOOL_FILEIMPORT :

					user_window = find_owner_class ( GetParent ( hDlg ) ) ;
					if ( user_window->import_file != NULL )
						fclose ( user_window->import_file ) ;
					user_window->import_file = OpenLoadFile ( hDlg, "import", "txt", "rb" ) ;
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_FILEEXPORT :

					// copy selected region to file
					find_owner_class ( GetParent(hDlg) )->copy_selected_region( DISK ) ;
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_FILECAPTURE :

					user_window = find_owner_class ( GetParent(hDlg) ) ;
					hMenu = GetMenu ( GetParent(hDlg) ) ;
					if ( user_window->capture_file != NULL )
					{
						fclose ( user_window->capture_file ) ;
						user_window->capture_file = NULL ;
						CheckMenuItem ( hMenu, CM_FILECAPTURE, MF_UNCHECKED ) ;
					}
					else
					{
						user_window->capture_file = OpenSaveFile ( hDlg, "capture", "txt", "wt" ) ;
						if ( user_window->capture_file != NULL )
							CheckMenuItem ( hMenu, CM_FILECAPTURE, MF_CHECKED ) ;
					}
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_COPY :

					// if copy region selected, copy it to the clipboard
					user_window = find_owner_class ( GetParent(hDlg) ) ;
					if ( user_window->copying_selected )
						user_window->copy_selected_region( CLIPBOARD ) ;
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_PASTE :

					// transfer clipboard into local paste buffer
					find_owner_class ( GetParent(hDlg) )->copy_clipboard_to_paste_buffer() ;
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_FONT :

					// get new font and update current display
					find_owner_class ( GetParent(hDlg) )->new_display_font() ;
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_WINDOW :

					user_window = find_owner_class ( GetParent( hDlg ) ) ;
					// activate dialog box to get new display options
					DialogBoxParam ( hInst, MAKEINTRESOURCE ( DISPLAY_DIALOG ), hDlg,
						(DLGPROC)DisplayOptions, (LPARAM)&user_window->display_options ) ;
					user_window->set_display_options() ;
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_HELP :

					// start help
					WinHelp(GetParent(hDlg),helpfilename,HELP_INDEX,0L);
					SetFocus ( GetParent ( hDlg ) ) ;
					break ;

				case TOOL_HISTORY :

					// messages sent to cmd history combo box
					switch ( GET_WM_COMMAND_CMD ( wParam, lParam ) )
					{

						case CBN_SELENDOK :
						case CBN_SELCHANGE :

							 // double click, so get selected command ident
							 cmd_ident = SendDlgItemMessage ( hDlg, TOOL_HISTORY,
								CB_GETCURSEL, 0, 0 ) ;
							 // then get text associated with it
							 SendDlgItemMessage ( hDlg, TOOL_HISTORY, CB_GETLBTEXT,
								cmd_ident, (LPARAM)cmd ) ;
							 // store text in paste buffer ready for later
							 // retrieval
							 find_owner_class (
								GetParent ( hDlg ) )->transfer_string_to_paste_bfr
									( cmd ) ;
							 SetFocus ( GetParent ( hDlg ) ) ;
							 break ;
					}
					break ;
			}
			return 0 ;

	}

	return 0 ;
}




















































