// absolute display extents
#define MAX_POS_X 140
#define MAX_POS_Y 80

// scrollable extents
#define MAX_SCROLL_Y MAX_POS_Y * 3
#define MAX_SCROLL_X MAX_POS_X    // X is not scrollable in this release
#define MAX_SCROLL_Y_POS MAX_SCROLL_Y-MAX_POS_Y

#define CMD_BUF_SIZE 255

// externally declared classes
class common_emulator ;
class graphics_support ;

struct display_options_type
	{	unsigned char emulation ;
		BOOL fixed_size ;
		int max_x ;
		int max_y ;
		BOOL wrap_flag ; } ;

typedef enum { CLIPBOARD, DISK, PASTEBFR } SelectOpType ;

// defines class for handling a PCShare display window
class window_access {
private:
	// data & variables
	char keyb_buffer[256] ;
	int keyb_write ;
	char *paste_buffer, *paste_ptr ;
	unsigned short paste_bfr_size ;
	struct display_options_type display_options ;
	int x_char, y_char, max_x, max_y ;
	int cursor_x, cursor_y ;
	char **text_region ;
	unsigned char **attr_region ;
	int window_base_y, scroll_base_y ;
	char **text_map ; // [MAX_POS_Y][MAX_POS_X] ;
	unsigned char **attr_map ; // [MAX_POS_Y][MAX_POS_X] ;
	HFONT holdsfont, hnewsfont, attr_font ;
	LOGFONT font_definition ;
	BOOL cursor_state, graphic_event ;
	BOOL copying_active, copying_selected ;
	int csx, csy, cex, cey ;
	int sx,sy,ex,ey ;
	common_emulator *screen_emulator ;
	graphics_support *graphic_win ;
	unsigned char display_attr, saved_attr ;
	int attr_save_x, attr_save_y ;
	COLORREF fore_colour, back_colour ;
	struct keymap_type keymap ;
	unsigned char keystate ;
	BOOL keypress_discard ;
	HDC hDC ;
	int window_id, scroll_pos ;
	FILE *import_file, *capture_file ;
	char import_bfr[255] ;
	// new in V4.0
	char cmd_line[CMD_BUF_SIZE] ;
	HWND CommHwnd, ToolWnd ;
	int cmd_pos ;
  BOOL GScroll ;
	// operations
	void paint_window( void ) ;
	void resize_window ( int x_size, int y_size ) ;
	void draw_cursor ( BOOL draw_flag ) ;
	void redraw_display ( void ) ;
	void paint_copy_region ( void ) ;
	void scroll_down() ;
	void scroll_up() ;
	void start_copy_region( int x_pos, int y_pos ) ;
	void mark_copy_region ( int x_pos, int y_pos ) ;
	void clear_copy_region ( void ) ;
	void end_copy_region ( void ) ;
	void copy_selected_region ( SelectOpType SelectOp ) ;
	void draw_attr_text ( int x, int y, int length ) ;
	void process_emulation_code (  unsigned char op_code ) ;
	void set_display_attr ( unsigned char attr_code ) ;
	void set_display_options ( void ) ;
	void copy_clipboard_to_paste_buffer ( void ) ;
	void copy_keypress_to_paste_buffer ( unsigned char keypress, BOOL special ) ;
	void new_display_font( void ) ;
	void remap_window( int top_row ) ;
	void process_scroll_request ( WPARAM wParam, LPARAM lParam ) ;
	// new in V4.00
	void transfer_string_to_paste_bfr ( char *string ) ;
   void write_line_buf ( void ) ;
public:
	// data and variables
	HWND MainWnd ;
	// operations
	window_access( char *window_name, int ident ) ;
	~window_access() ;
	void output_text ( char textcr ) ;
	void select_video_emulation ( unsigned char emulation_code ) ;
	void set_text_attr ( int fore_col, int back_col ) ;
	void graphics_set_mode ( unsigned x_pixels, unsigned y_pixels, float x_scale, float y_scale ) ;
	void graphics_cursor_update ( unsigned x_coord, unsigned y_coord ) ;
	void graphics_set_point ( unsigned x_coord, unsigned y_coord, unsigned char colour ) ;
	void graphics_draw_line ( unsigned x1_coord,
                            unsigned y1_coord,
                            unsigned x2_coord,
                            unsigned y2_coord,
                            unsigned char colour,
                            unsigned char width,
                            unsigned char style ) ;
	void graphics_draw_circle ( unsigned x_coord,
                              unsigned y_coord,
                              unsigned radius,
                              unsigned char colour,
                              unsigned char width,
                              unsigned char style ) ;
	unsigned char graphics_read_pixel ( unsigned x_coord, unsigned y_coord ) ;
	void graphics_colour_change ( unsigned char colour ) ;
	void graphics_clear_screen ( unsigned char colour ) ;
	// new in v4.0 - define palette entry
	void graphics_define_palette_entry ( int pal_entry, unsigned char red,
					unsigned char green, unsigned char blue ) ;
	char *get_paste_buffer ( unsigned short *buffer_size ) ;
	void flash_cursor ( void ) ;
	friend LRESULT CALLBACK EditProc (HWND, UINT, WPARAM, LPARAM);
	friend BOOL CALLBACK _export EditToolBar (HWND, UINT, WPARAM, LPARAM);
	friend LRESULT CALLBACK MainWndProc (HWND, UINT, WPARAM, LPARAM ) ;
} ;

#define MAX_WINDOWS 10
extern window_access *active_windows[MAX_WINDOWS] ;

extern HWND active_hwnd ;
