#include <windows.h>
#include <windowsx.h>
#include "pcshare.h"
#include "io_ports.h"
#include "pcdragon_rc.h"

static BOOL Pending = FALSE ;
static HANDLE InWHandle, InRHandle, InDupWHandle ;
static HANDLE OutWHandle, OutRHandle ;
static unsigned char PChar ;
static HINSTANCE hInst ;
static char Path[MAX_PATH] ;
static char Opts[80] ;

STARTUPINFO StartUpInfo = {
  sizeof(STARTUPINFO),
  NULL,
  NULL,
  NULL,  // LPTSTR  lpTitle;
  0,  // DWORD   dwX;
  0,  // DWORD   dwY;
  0,  // DWORD   dwXSize;
  0,  // DWORD   dwYSize;
  0,  // DWORD   dwXCountChars;
  0,  // DWORD   dwYCountChars;
  0,  // DWORD   dwFillAttribute;
  STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES, // DWORD   dwFlags;
  SW_SHOWDEFAULT,    // WORD    wShowWindow;
  0,  // WORD    cbReserved2;
  NULL,  // LPBYTE  lpReserved2;
  NULL,  // HANDLE  hStdInput;
  NULL,  // HANDLE  hStdOutput;
  NULL,  // HANDLE  hStdError;
  };

/* config dialog box */
BOOL CALLBACK _export ConfigProc (HWND, UINT, WPARAM, LPARAM);

BOOL CALLBACK _export ConfigProc (HWND hDlg, UINT Message,
								  WPARAM wParam, LPARAM lParam)
{
  DWORD Spare  ;
  static HKEY hKey ;
  DWORD Size ;
  const char *EmuPath = "PCDragon Path" ;
  const char *EmuOpts = "PCDragon Opts" ;
  const char *DefaultOpts = "dragon.exe /64 -nb dosv1.dgn" ;
  const char *DefaultPath = "c:\\PCDragon" ;

 	switch ( Message )
	{
		case WM_INITDIALOG :

      if ( RegCreateKeyEx ( HKEY_LOCAL_MACHINE,
                            REGPATH,
                            0,
                            NULL,
                            REG_OPTION_NON_VOLATILE,
                            KEY_QUERY_VALUE | KEY_SET_VALUE,
                            NULL,
                            &hKey,
                            &Spare ) == ERROR_SUCCESS )
      {
        Size = sizeof(Path) ;
        if ( RegQueryValueEx ( hKey,
                              (LPCTSTR)EmuPath,
                               NULL,
                               &Spare,
                              (LPBYTE)Path,
                              &Size ) !=  ERROR_SUCCESS )
        {
          strcpy ( Path, DefaultPath ) ;
        }

        Size = sizeof(Opts) ;
        if ( RegQueryValueEx ( hKey,
                               (LPCTSTR)EmuOpts,
                               NULL,
                              &Spare,
                              (LPBYTE)Opts,
                              &Size ) != ERROR_SUCCESS )
        {
          strcpy ( Opts, DefaultOpts ) ;
        }
        SetDlgItemText ( hDlg, PCDRAGON_PATH, Path ) ;
        SetDlgItemText ( hDlg, PCDRAGON_OPTS, Opts ) ;
      }
      return 1 ;

 		case WM_COMMAND :

			switch (GET_WM_COMMAND_ID ( wParam, lParam ))
			{
        case IDOK :

          GetDlgItemText ( hDlg, PCDRAGON_PATH, Path, sizeof(Path) ) ;
          GetDlgItemText ( hDlg, PCDRAGON_OPTS, Opts, sizeof(Opts) ) ;
          RegSetValueEx ( hKey,
                          EmuPath,
                          0,
                          REG_SZ,
                          (CONST BYTE*)Path,
                          strlen(Path)+1 ) ;
          RegSetValueEx ( hKey,
                          EmuOpts,
                          0,
                          REG_SZ,
                          (CONST BYTE*)Opts,
                          strlen(Opts)+1 ) ;
          RegCloseKey ( hKey ) ;
          EndDialog ( hDlg, 0 ) ;
          return 0 ;
      }
      break ;
  }
  return 0 ;
}

#pragma argsused
BOOL WINAPI DllEntryPoint(HINSTANCE hDLLInst, DWORD  fdwReason,
					LPVOID lpvReserved)
{
	switch (fdwReason)
	{
		/*
		 * The DLL is attaching to a process due to process
		 * initialization or a call to LoadLibrary.
		 */
		 case DLL_PROCESS_ATTACH:

        hInst = hDLLInst ;
        break ;
  }
	return TRUE;
}

void far __pascal _export read_data ( void *buf, int count )
{
  unsigned char *Ptr = buf ;
  DWORD Cbr ;

  // handle pending char from prior in_port_status call
  if ( Pending )
  {
    *Ptr++ = PChar ;
    count-- ;
    Pending = FALSE ;
  }

  // read next section
  while ( count )
  {
    ReadFile ( OutRHandle, Ptr, count, &Cbr, NULL ) ;
    Ptr+=Cbr ;
    count-=Cbr ;
  }
}

int far __pascal _export in_port_status ( void )
{
  DWORD Cbr ;

  if ( Pending )
    return TRUE ;
  // is blocking so always true
  ReadFile ( OutRHandle, &PChar, 1, &Cbr, NULL ) ;
  Pending = TRUE ;
  return TRUE ;
}

void far __pascal _export write_data ( void *buf, int count )
{
  DWORD cbw ;

  WriteFile ( InDupWHandle, buf, count, &cbw, NULL ) ;
}

int far __pascal _export out_port_status ( void )
{
  return TRUE ;
}

#pragma argsused
void far __pascal _export set_port_base ( int port_base, int irq, int dma )
{
  PROCESS_INFORMATION ProcessInformation;
  SECURITY_ATTRIBUTES saAttr;
  char ExecPath[MAX_PATH] ;

  // config info
  DialogBox ( hInst, MAKEINTRESOURCE ( PCDRAGON_BOOT ), NULL, ConfigProc ) ;

  /* Set the bInheritHandle flag so pipe handles are inherited.
    - only relevent for nt mind u. */
  saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
  saAttr.bInheritHandle = TRUE;
  saAttr.lpSecurityDescriptor = NULL;

  /* create two pipes - one for emu's stdin, one for emu's stdout */
  CreatePipe ( &InRHandle, &InWHandle, &saAttr, 32768 ) ;
  CreatePipe ( &OutRHandle, &OutWHandle, &saAttr, 32768 ) ;
  StartUpInfo.hStdOutput = OutWHandle ;
  StartUpInfo.hStdError = OutWHandle ;
  StartUpInfo.hStdInput = InRHandle ;
  /* stdin w-handle is non-inheritable */
  DuplicateHandle ( GetCurrentProcess(),
                    InWHandle,
                    GetCurrentProcess(),
                    &InDupWHandle,
                    0,
                    FALSE,
                    DUPLICATE_SAME_ACCESS);

  /* start pc-dragon emulator */
  strcpy ( ExecPath, Path ) ;
  strcat ( ExecPath, "\\" ) ;
  strcat ( ExecPath, Opts ) ;
  if ( !CreateProcess(NULL,ExecPath,
          NULL, // pointer to process security attributes
          NULL, // pointer to thread security attributes
          TRUE, // handle inheritance flag
          CREATE_NEW_CONSOLE,  // creation flags
          NULL, // pointer to new environment block
          Path,   // pointer to current directory name
          &StartUpInfo, // pointer to STARTUPINFO
         &ProcessInformation  // pointer to PROCESS_INFORMATION
          ) )
  {
    MessageBox ( NULL, "PCDragon failed to start",
                       "CreateProcess failed", MB_OK | MB_ICONSTOP ) ;
    PostQuitMessage(0);
  }

  // close non-inher handle
  CloseHandle ( InWHandle ) ;
}

void far __pascal _export close_io_port ( void )
{
  // close all pipe handles - dont't touch the spawned process
  CloseHandle ( InDupWHandle ) ;
  CloseHandle ( InRHandle ) ;
  CloseHandle ( OutRHandle ) ;
  CloseHandle ( OutWHandle ) ;
}
